#include "hazy_message.h"
#include <string>
#include <iostream>
#include <sstream>
#include "vectors.h"
#include "globals.h"

#ifdef __cplusplus
extern "C" {
#endif
  //void free_message_data(message_data msg) { free(msg); } 
  
  /**
   * Copies given source string to the destination string
   * \param[out] dest destination string
   * \param[in] src source string
   * \param[in] nMaxSize maximum size of destination string
   * \return success
   */
  int
  inline_copy(char *dest, const char *src, unsigned int nMaxSize) {
    if (strlen(src) < nMaxSize) {
      strcpy(dest, src);
      return 1;
    } else { 
      return -1;
    }
  }
  
  
  /**
   * Builds answer msg for read single entity queries
   * \param[in] entity_label class of the entity queried
   * \param[out] _out msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
int
build_read_se_answer(int entity_label, message_data _out, unsigned int nMaxSize)
{
  std::ostringstream msg;
  msg.precision(sizeof(VFloat)==4 ? 7 : 16);
  msg << entity_label << szDelim;
  return inline_copy((char *) _out, msg.str().c_str(), nMaxSize);
}
  
  /**
   * Builds answer msg for read count queries
   * \param[in] count count of entities in class
   * \param[out] _out msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_read_count_answer(int count, message_data _out, unsigned int nMaxSize)
  {
    std::ostringstream msg;
    msg.precision(sizeof(VFloat)==4 ? 7 : 16);
    msg << count << szDelim;
    assert(msg.str().length() < nMaxSize);
    return inline_copy((char *) _out, msg.str().c_str(), nMaxSize);
  }
  
  /**
   * Builds answer msg for update queries
   * \param[out] _out msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_update_answer(message_data _out, unsigned int nMaxSize) {
    return inline_copy((char *) _out,done_msg, nMaxSize);
  }

  int
  build_delete_answer(message_data _out, unsigned int nMaxSize) {
    return inline_copy((char *) _out, done_msg, nMaxSize);
  }
  
  /**
   * Builds answer msg for querying server to see whether server is up and running
   * \param[in] id server id
   * \param[out] _out msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_query_server_answer(int id, message_data _out, unsigned int nMaxSize) {
    std::ostringstream msg;
    msg.precision(sizeof(VFloat)==4 ? 7 : 16);
    msg << id << szDelim;
    return inline_copy((char *) _out, msg.str().c_str(), nMaxSize);
  }
  
  /**
   * Builds answer msg for closing the server
   * \param[in] close_conn success of closing the connection
   * \param[out] _out msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_close_connection_answer(int close_conn, message_data _out, unsigned int nMaxSize) {
    std::ostringstream msg;
    msg.precision(sizeof(VFloat)==4 ? 7 : 16);
    msg << close_conn << szDelim;
    return inline_copy((char *) _out, msg.str().c_str(), nMaxSize);
  }
  
  /**
   * Builds answer msg for no operation with server (for experimental purposes)
   * \param[in] acknowledge ack message
   * \param[out] _out msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int build_noop_server_answer(int acknowledge, message_data _out, unsigned int nMaxSize) {
    std::ostringstream msg;
    msg.precision(sizeof(VFloat)==4 ? 7 : 16);
    msg << acknowledge << szDelim;
    return inline_copy((char *) _out, msg.str().c_str(), nMaxSize);
  }
  
  int build_create_view_answer(message_data _out, unsigned int nMaxSize) {
    return inline_copy((char *) _out,done_msg, nMaxSize);
  }

  /**
   * Builds querying msg for single entity eager
   * \param[in] view_id id of the view
   * \param[in] entity_id id of the entity
   * \param[out] _out msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_se_eager_msg(const int view_id, const int entity_id, message_data _out, unsigned int nMaxSize) {
    std::ostringstream msg;
    msg.precision(sizeof(VFloat)==4 ? 7 : 16);
    msg << READ_SE_EAGER << szDelim << view_id << szDelim << entity_id << szDelim;
    return inline_copy((char *) _out, msg.str().c_str(), nMaxSize);
  }
  
  /**
   * Builds querying msg for single entity lazy naive
   * \param[in] view_id id of the view
   * \param[in] vector_str feature vector of the entity
   * \param[out] _out msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_se_lazy_naive_msg(const int view_id, const char* vector_str, message_data _out, unsigned int nMaxSize) {
    std::ostringstream msg;
    msg.precision(sizeof(VFloat)==4 ? 7 : 16);
    msg << READ_SE_LAZY_NAIVE << szDelim << view_id << szDelim << vector_str << szDelim;
    return inline_copy((char *) _out, msg.str().c_str(), nMaxSize);
  } 
  
  /**
   * Builds querying msg for single entity lazy naive that sends binary data
   * \param[in] view_id id of the view
   * \param[in] vector_ptr pointer to the feature vector of the entity
   * \param[in] vec_size size of the vector pointer
   * \param[out] _out msg being prepared
   * \param[in] out_size size of the _out msg 
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_se_lazy_naive_msg_bin(const int view_id, const void* vector_ptr, size_t vec_size, 
			      void* _out, size_t *out_size, unsigned int nMaxSize) {
    *out_size = vec_size + sizeof(struct se_lazy_naive_msg_bin_header);
    assert(nMaxSize > *out_size);
    
    struct se_lazy_naive_msg_bin_header *msg = (struct se_lazy_naive_msg_bin_header *) _out;
    msg->t            = READ_SE_LAZY_NAIVE_BIN;
    msg->view_id      = view_id;
    msg->payload_size = vec_size;
    void* data = (void*) (&msg[1]);
    memcpy(data, vector_ptr, vec_size);
    return 0;
  } 

  int
  build_delete_msg_bin(const int view_id, const int entity_id, const void *db_name_ptr, size_t db_name_ptr_size,
		       const void *table_name_ptr, size_t table_name_ptr_size,
		       void *_out, size_t *out_size, unsigned int nMaxSize) {

    *out_size = db_name_ptr_size + table_name_ptr_size + sizeof(struct delete_bin_header);
    assert(nMaxSize > *out_size);
    
    struct delete_bin_header *msg = (struct delete_bin_header *) _out;
    msg->t               = DELETE_BIN;
    msg->view_id         = view_id;
    msg->deleted_id      = entity_id;
    msg->db_name_size    = db_name_ptr_size;
    msg->table_name_size = table_name_ptr_size;

    void *data = (void *) (&msg[1]);
    memcpy(data, db_name_ptr, db_name_ptr_size);
    void *data2 = ((char *) data) + db_name_ptr_size;
    memcpy(data2, table_name_ptr, table_name_ptr_size);

    return 0;
  }
  
  
  /**
   * Builds querying msg for single entity lazy hazy
   * \param[in] view_id id of the view
   * \param[in] vector_str feature vector of the entity
   * \param[in] eps eps value of the entity
   * \param[out] _out msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_se_lazy_hazy_msg(const int view_id, const char *vector_str, const double eps, message_data _out, unsigned int nMaxSize) {
    std::ostringstream msg;
    msg.precision(sizeof(VFloat)==4 ? 7 : 16);
    msg << READ_SE_LAZY_HAZY << szDelim << view_id << szDelim << vector_str << szDelim << eps << szDelim;
    return inline_copy((char *) _out, msg.str().c_str(), nMaxSize);
  }
  
  
  /**
   * Builds querying msg for single entity lazy hazy that sends binary data
   * \param[in] view_id id of the view
   * \param[in] eps epsilon value of the entity being queries
   * \param[in] vector_ptr pointer to the feature vector of the entity
   * \param[in] vec_size size of the vector pointer
   * \param[out] _out msg being prepared
   * \param[in] out_size size of the _out msg
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_se_lazy_hazy_msg_bin(const int view_id, const double eps, const void* vector_ptr, size_t vec_size,
			     void* _out, size_t *out_size, unsigned int nMaxSize) {
    *out_size = vec_size + sizeof(struct se_lazy_hazy_msg_bin_header);
    assert(nMaxSize > *out_size);
    
    struct se_lazy_hazy_msg_bin_header *msg = (struct se_lazy_hazy_msg_bin_header *) _out;
    msg->t            = READ_SE_LAZY_HAZY_BIN;
    msg->view_id      = view_id;
    msg->eps          = eps;
    msg->payload_size = vec_size;
    void* data = (void*) (&msg[1]);
    memcpy(data, vector_ptr, vec_size);
    return 0;
  }
  
  /**
   * Builds querying msg for count
   * \param[in] view_id id of the view
   * \param[out] _out msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_count_msg(const int view_id, message_data _out, unsigned int nMaxSize) {
    std::ostringstream msg;
    msg.precision(sizeof(VFloat)==4 ? 7 : 16);
    msg << READ_COUNT << szDelim << view_id << szDelim;
    return inline_copy((char *) _out, msg.str().c_str(), nMaxSize);
  }
  
  /**
   * Builds querying msg for update
   * \param[in] view_id id of the view
   * \param[in] update_example new update example
   * \param[out] _out msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_update_msg(const int view_id, const char *update_example, message_data _out, unsigned int nMaxSize) {
    std::ostringstream msg;
    msg.precision(sizeof(VFloat)==4 ? 7 : 16);
    msg << UPDATE << szDelim << view_id << szDelim << update_example << szDelim;
    int ret_val = inline_copy((char *) _out, msg.str().c_str(), nMaxSize);
    return ret_val;
  }
  
  /**
   * Builds querying msg for update that sends binary data
   * \param[in] view_id id of the view
   * \param[in] label label of the entity
   * \param[in] vector pointer to the feature vector of the entity
   * \param[in] vec_size size of the vector pointer
   * \param[out] _out msg being prepared
   * \param[in] out_size size of the _out msg
   * \return success of building the message
   */
  int
  build_update_msg_bin(const int view_id, const int label, 
		       const void* vector, size_t vec_size,
		       void* _out, size_t *out_size) {
    *out_size = vec_size + sizeof(struct update_bin_header);
    
    struct update_bin_header *msg = (struct update_bin_header *) _out;
    msg->t            = UPDATE_BIN;
    msg->view_id      = view_id;
    msg->label        = label;
    msg->payload_size = vec_size;
    void* data = (void*) (&msg[1]);
    memcpy(data, vector, vec_size);
    return 0;
  }
  
  
  /**
   * Builds querying msg for server id
   * \param[out] _out msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_query_server_id_msg(message_data _out, unsigned int nMaxSize) {
    std::ostringstream msg;
    msg << QUERY_SERVER_ID;
    return inline_copy((char *) _out, msg.str().c_str(), nMaxSize);
  }
  
  /**
   * Builds querying msg for closing the connection
   * \param[out] _out msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_close_connection_msg(message_data _out, unsigned int nMaxSize) {
    std::ostringstream msg;
    msg << CLOSE_CONNECTION;
    return inline_copy((char *) _out, msg.str().c_str(), nMaxSize);
  }
  
  /**
   * Builds querying msg for no operation with server
   * \param[out] _out msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_noop_server_msg(message_data _out, unsigned int nMaxSize) {
    std::ostringstream msg;
    msg << SERVER_NOOP;
    return inline_copy((char *) _out, msg.str().c_str(), nMaxSize);
  }
  
  ///get message type
  /**
   * Parses given message and extracts message type
   * \param[in] msg message to be parsed
   * \return message_type message type of given msg
   */
  message_type get_message_type(const message_data msg) {
    int int_ret_val = int(((char *) msg)[0]);
    if(int_ret_val >= CONVERT_CHAR_TO_INT)
      int_ret_val -= CONVERT_CHAR_TO_INT;
    return (message_type) int_ret_val;
  }
#ifdef __cplusplus
}
#endif

int build_create_view_msg(struct create_view_bin_str *cv_struct, message_data _out, unsigned int nMaxSize) {
  std::ostringstream msg;
  msg.precision(sizeof(VFloat)==4 ? 7 : 16);
  msg << CREATE_VIEW << szDelim << cv_struct->view_id << szDelim << cv_struct->sm << szDelim << 
    cv_struct->dense << szDelim << cv_struct->useepsmap << szDelim << cv_struct->hazy_strategy << 
    szDelim << cv_struct->reservoir_size << szDelim << cv_struct->resort_time << szDelim << 
    cv_struct->tao_val << szDelim << cv_struct->alpha_ratio << szDelim << cv_struct->buffer_size <<
    szDelim << cv_struct->view_name << szDelim << cv_struct->view_key << szDelim << 
    cv_struct->entity_table_name << szDelim << cv_struct->entity_table_key << szDelim << 
    cv_struct->training_table_name << szDelim << cv_struct->dim << szDelim << cv_struct->db_name <<
    szDelim << cv_struct->m_factor << szDelim;

  int ret_val = inline_copy((char *) _out, msg.str().c_str(), nMaxSize);
  return ret_val;
}

// Server side only methods.

///parse msg methods
/**
 * Parses single entity eager query messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 * \param[out] entity_id id of the entity
 */
void parse_se_eager_msg(message_data msg, int &view_id, int &entity_id) {
  std::stringstream parser;	
  parser << (char *) msg;
  std::string parsed_word;
  
  //gets message type
  std::getline(parser, parsed_word, szDelim);	
  //gets view_id
  std::getline(parser, parsed_word, szDelim);	
  view_id = atoi(parsed_word.c_str());
  //gets entity_id
  std::getline(parser, parsed_word, szDelim);	
  entity_id = atoi(parsed_word.c_str());
}

/**
 * Parses view id in single entity eager query messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 * \param[out] entity_id id of the entity
 */
void parse_view_id_in_se_eager_msg(message_data msg, int &view_id) {
  std::stringstream parser;	
  parser << (char *) msg;
  std::string parsed_word;
  
  //gets message type
  std::getline(parser, parsed_word, szDelim);	
  //gets view_id
  std::getline(parser, parsed_word, szDelim);	
  view_id = atoi(parsed_word.c_str());
}

/**
 * Parses count query messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */
void parse_count_msg(message_data msg, int &view_id) {
  std::stringstream parser;
  parser << (char *) msg;
  std::string parsed_word;
  
  //gets message type
  std::getline(parser, parsed_word, szDelim);
  //gets view_id
  std::getline(parser, parsed_word, szDelim);
  view_id = atoi(parsed_word.c_str());
}

void parse_delete_msg_bin(message_data msg, int &view_id, int &entity_id, std::string &db_name, std::string &training_table_name) {
  struct delete_bin_header *msg_header = (struct delete_bin_header*) msg;
  assert(msg_header->t == DELETE_BIN);
  view_id              = msg_header->view_id;
  entity_id            = msg_header->deleted_id;
  int db_name_size     = msg_header->db_name_size;
  int table_name_size  = msg_header->table_name_size; 

  char *copy_db_name = (char *) malloc(sizeof(char) * (db_name_size + 1));
  char *copy_training_table_name = (char *) malloc(sizeof(char) * table_name_size + 1);
  
  void *data = (void *) (&msg_header[1]);
  std::string adb_name((char *) data, db_name_size);
  memcpy(copy_db_name, data, db_name_size);
  copy_db_name[db_name_size] = 0;
  db_name = copy_db_name;
  void *data2 = ((char *) data) + db_name_size;
  memcpy(copy_training_table_name, data2, table_name_size);
  copy_training_table_name[table_name_size] = 0;
  training_table_name = copy_training_table_name;

  free(copy_db_name);
  free(copy_training_table_name);
}


void parse_view_id_in_delete_msg_bin(message_data msg, int &view_id) {
  struct delete_bin_header *msg_header = (struct delete_bin_header*) msg;
  assert(msg_header->t == DELETE_BIN);
  view_id              = msg_header->view_id;
}

/**
 * Parses view id in single entity lazy naive messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */
void parse_view_id_in_se_lazy_naive_msg(message_data msg, int &view_id) {
  std::stringstream parser;
  parser << (char *) msg;
  std::string parsed_word;
  
  //gets message type
  std::getline(parser, parsed_word, szDelim);
  //gets view_id
  std::getline(parser, parsed_word, szDelim);
  view_id = atoi(parsed_word.c_str());
}


/**
 * Parses view id in single entity lazy naive messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */
void parse_view_id_in_se_lazy_naive_msg_bin(void* msg, int &view_id) {
  struct se_lazy_naive_msg_bin_header* msg_header = (struct se_lazy_naive_msg_bin_header*)msg;
  assert(msg_header->t == READ_SE_LAZY_NAIVE_BIN);
  view_id          = msg_header->view_id;
}


/**
 * Parses view id in single entity lazy hazy messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */ 
void parse_view_id_in_se_lazy_hazy_msg(message_data msg, int &view_id) {
  std::stringstream parser;
  parser << (char *) msg;
  std::string parsed_word;
  
  //gets message type
  std::getline(parser, parsed_word, szDelim);
  //gets view_id
  std::getline(parser, parsed_word, szDelim);
  view_id = atoi(parsed_word.c_str());
}


/**
 * Parses single entity lazy hazy messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */
void parse_view_id_in_se_lazy_hazy_msg_bin(void* msg, int &view_id) {
  struct se_lazy_hazy_msg_bin_header* msg_header = (struct se_lazy_hazy_msg_bin_header*)msg;
  assert(msg_header->t == READ_SE_LAZY_HAZY_BIN);
  view_id          = msg_header->view_id;
}

/**
 * Parses update messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */
void parse_view_id_in_update_msg(message_data msg, int &view_id) {
  std::stringstream parser;
  parser << (char *) msg;
  std::string parsed_word;
  
  //gets message type
  std::getline(parser, parsed_word, szDelim);
  //gets view_id
  std::getline(parser, parsed_word, szDelim);
  view_id = atoi(parsed_word.c_str());
}


/**
 * Parses view id in binary update messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */
void parse_view_id_in_update_msg_bin(message_data msg, int &view_id) {
  struct update_bin_header *msg_header = (struct update_bin_header*) msg;
  assert(msg_header->t == UPDATE_BIN);
  view_id              = msg_header->view_id;
}

void parse_create_view_msg(message_data msg, struct create_view_bin_str *cv_struct) {
  std::stringstream parser;	
  parser << (char *) msg;
  std::string parsed_word;
  
  //gets message type
  std::getline(parser, parsed_word, szDelim);	
  //gets view_id
  std::getline(parser, parsed_word, szDelim);	
  cv_struct->view_id = atoi(parsed_word.c_str());
  //gets sm
  std::getline(parser, parsed_word, szDelim);
  cv_struct->sm = (enum storage_manager_type) atoi(parsed_word.c_str());
  //gets dense
  std::getline(parser, parsed_word, szDelim);
  cv_struct->dense = atoi(parsed_word.c_str());
  //gets useepsmap
  std::getline(parser, parsed_word, szDelim);
  cv_struct->useepsmap = atoi(parsed_word.c_str());
  //gets hazy_strategy
  std::getline(parser, parsed_word, szDelim);
  cv_struct->hazy_strategy = (hazy_model::strategy) atoi(parsed_word.c_str());
  //gets reservoir_size
  std::getline(parser, parsed_word, szDelim);
  cv_struct->reservoir_size = atoi(parsed_word.c_str());
  //gets resort_time
  std::getline(parser, parsed_word, szDelim);
  cv_struct->resort_time = atof(parsed_word.c_str());
  //gets tao_val
  std::getline(parser, parsed_word, szDelim);
  cv_struct->tao_val = atof(parsed_word.c_str());
  //gets alpha_ratio
  std::getline(parser, parsed_word, szDelim);
  cv_struct->alpha_ratio = atof(parsed_word.c_str());
  //gets buffer_size
  std::getline(parser, parsed_word, szDelim);
  cv_struct->buffer_size = atoi(parsed_word.c_str());
  //gets view_name
  std::getline(parser, parsed_word, szDelim);
  cv_struct->view_name = parsed_word;
  //gets view_key
  std::getline(parser, parsed_word, szDelim);
  cv_struct->view_key = parsed_word;
  //gets entity_table_name
  std::getline(parser, parsed_word, szDelim);
  cv_struct->entity_table_name = parsed_word;
  //gets entity_table_key
  std::getline(parser, parsed_word, szDelim);
  cv_struct->entity_table_key = parsed_word;
  //gets training_table_name
  std::getline(parser, parsed_word, szDelim);
  cv_struct->training_table_name = parsed_word;
  //gets dim
  std::getline(parser, parsed_word, szDelim);
  cv_struct->dim = atoi(parsed_word.c_str());
  //gets db_name
  std::getline(parser, parsed_word, szDelim);
  cv_struct->db_name = parsed_word;
  //gets m_factor
  std::getline(parser, parsed_word, szDelim);
  cv_struct->m_factor = atof(parsed_word.c_str());
}
