/**
 * \file hazy_message.h
 */

#ifndef HAZY_MESSAGE_H
#define HAZY_MESSAGE_H
#include <string.h>
#include "globals.h"

/**
 * This class prepares and parses communication messages between clients and the server
 */

///message types
enum message_type {READ_SE_EAGER, READ_SE_LAZY_NAIVE, READ_SE_LAZY_HAZY, READ_COUNT, UPDATE, QUERY_SERVER_ID, CLOSE_CONNECTION, SERVER_NOOP, CREATE_VIEW, READ_SE_LAZY_NAIVE_BIN, READ_SE_LAZY_HAZY_BIN, UPDATE_BIN, DELETE_BIN};

///Structure for messages used in single entity lazy naive queries
struct se_lazy_naive_msg_bin_header {
  enum message_type t; 
  int view_id;
  int payload_size;
};

///Structure for messages used in update queries
struct update_bin_header {
  enum message_type t; 
  int view_id;
  int label;
  int payload_size;
};

///Structure for messages used in single entity lazy hazy queries
struct se_lazy_hazy_msg_bin_header {
  enum message_type t;
  int view_id;
  double eps;
  int payload_size;
};

struct delete_bin_header {
  enum message_type t;
  int view_id;
  int deleted_id;
  int db_name_size;
  int table_name_size;
};


///type definition for message_data which is void pointer
typedef void* message_data;

///delimiter between messages
const char szDelim = '|';
///constant to convert character to its integer values
const int CONVERT_CHAR_TO_INT = 48;
///done msg used for updates
const char done_msg[6] = "done|";

///build msg methods for server side
#ifdef __cplusplus
extern "C"  {
#endif

  /// the interface allows write in place.
  /// return 0 on success
  /// -1 on on failure
  /**
   * Builds answer msg for read single entity queries
   * \param[in] entity_label class of the entity queried
   * \param[out] msg msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int build_read_se_answer(int entity_label, message_data msg, unsigned int nMaxSize);
  /**
   * Builds answer msg for read count queries
   * \param[in] count count of entities in class
   * \param[out] msg msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int build_read_count_answer(int count, message_data msg, unsigned int nMaxSize);
  /**
   * Builds answer msg for update queries
   * \param[out] msg msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int build_update_answer(message_data msg, unsigned int nMaxSize);
  /**
   * Builds answer msg for querying server to see whether server is up and running
   * \param[in] id server id
   * \param[out] msg msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */

  int build_delete_answer(message_data msg, unsigned int nMaxSize);

  int build_query_server_answer(int id, message_data msg, unsigned int nMaxSize);
  /**
   * Builds answer msg for closing the server
   * \param[in] close_conn success of closing the connection
   * \param[out] msg msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int build_close_connection_answer(int close_conn, message_data msg, unsigned int nMaxSize);
  
  /**
   * Builds noop server msg for no operation (for experimental purposes)
   * \param[in] acknowledge ack for noop
   * \param[out] msg msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int build_noop_server_answer(int acknowledge, message_data msg, unsigned int nMaxSize);

  int build_create_view_answer(message_data msg, unsigned int nMaxSize);
  
  ///build msg methods for triggers & functions
  /**
   * Builds querying msg for single entity eager
   * \param[in] view_id id of the view
   * \param[in] entity_id id of the entity
   * \param[out] msg msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int build_se_eager_msg(const int view_id, const int entity_id, message_data msg, unsigned int nMaxSize);
  /**
   * Builds querying msg for single entity lazy naive
   * \param[in] view_id id of the view
   * \param[in] vector_str feature vector of the entity
   * \param[out] msg msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int build_se_lazy_naive_msg(const int view_id, const char *vector_str, message_data msg, unsigned int nMaxSize);
  /**
   * Builds querying msg for single entity lazy hazy
   * \param[in] view_id id of the view
   * \param[in] vector_str feature vector of the entity
   * \param[in] eps eps value of the entity
   * \param[out] msg msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int build_se_lazy_hazy_msg(const int view_id, const char *vector_str, const double eps, message_data msg, unsigned int nMaxSize);
  /**
   * Builds querying msg for count
   * \param[in] view_id id of the view
   * \param[out] msg msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int build_count_msg(const int view_id, message_data msg, unsigned int nMaxSize);
  /**
   * Builds querying msg for update
   * \param[in] view_id id of the view
   * \param[in] update_example new update example
   * \param[out] msg msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int build_update_msg(const int view_id, const char *update_example, message_data msg, unsigned int nMaxSize);
  /**
   * Builds querying msg for server id
   * \param[out] msg msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int build_query_server_id_msg(message_data msg, unsigned int nMaxSize);
  /**
   * Builds querying msg for closing the connection
   * \param[out] msg msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int build_close_connection_msg(message_data msg, unsigned int nMaxSize);
  
  /**
   * Builds querying msg for no operation with server (for experimental purposes)
   * \param[out] msg msg being prepared
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int build_noop_server_msg(message_data msg, unsigned int nMaxSize);
  
  ///get message type
  /**
   * Parses given message and extracts message type
   * \param[in] msg message to be parsed
   * \return message_type message type of given msg
   */
  enum message_type get_message_type(const message_data msg);  
  
  /**
   * Builds querying msg for single entity lazy naive that sends binary data
   * \param[in] view_id id of the view
   * \param[in] vector_ptr pointer to the feature vector of the entity
   * \param[in] vec_size size of the vector pointer
   * \param[out] _out msg being prepared
   * \param[in] out_size size of the _out msg 
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_se_lazy_naive_msg_bin(const int view_id, const void* vector_ptr, size_t vec_size, 
			      void* _out, size_t *out_size, unsigned int nMaxSize);
  
  /**
   * Builds querying msg for single entity lazy hazy that sends binary data
   * \param[in] view_id id of the view
   * \param[in] eps epsilon value of the entity being queries
   * \param[in] vector_ptr pointer to the feature vector of the entity
   * \param[in] vec_size size of the vector pointer
   * \param[out] _out msg being prepared
   * \param[in] out_size size of the _out msg
   * \param[in] nMaxSize max size of msg
   * \return success of building the message
   */
  int
  build_se_lazy_hazy_msg_bin(const int view_id, const double eps, const void* vector_ptr, size_t vec_size,
			     void* _out, size_t *out_size, unsigned int nMaxSize);

  int
  build_delete_msg_bin(const int view_id, const int entity_id, const void *db_name_ptr, size_t db_name_ptr_size,
		       const void *table_name_ptr, size_t table_name_ptr_size,
		       void *_out, size_t *out_size, unsigned int nMaxSize);
  
#ifdef __cplusplus
}
#endif

#ifdef __cplusplus
#include <string>
#include <iostream>
#include <sstream>
#include "vectors.h"
#include "hazy_model.h"

struct create_view_bin_str {
  enum message_type t;
  int view_id;
  enum storage_manager_type sm;
  int dense;
  int useepsmap;
  hazy_model::strategy hazy_strategy;
  int reservoir_size;
  double resort_time;
  double tao_val;
  double alpha_ratio;
  int buffer_size;
  std::string view_name;
  std::string view_key;
  std::string entity_table_name;
  std::string entity_table_key;
  std::string training_table_name;
  int dim;
  std::string db_name;
  double m_factor;
};

int
build_create_view_msg(struct create_view_bin_str *s, message_data _out, unsigned int nMaxSize);

///parse msg methods (THESE ARE IN C++ and so we use those features)
/**
 * Parses single entity eager query messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 * \param[out] entity_id id of the entity
 */
void parse_se_eager_msg(const message_data msg, int &view_id, int &entity_id);	

/**
 * Parses view id in single entity eager query messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */
void parse_view_id_in_se_eager_msg(const message_data msg, int &view_id);

/**
 * Parses count query messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */
void parse_count_msg(const message_data msg, int &view_id); 

void parse_delete_msg_bin(message_data msg, int &view_id, int &entity_id, std::string &db_name, std::string &training_table_name);

void parse_view_id_in_delete_msg_bin(message_data msg, int &view_id);

// Templates
template<class T>
/**
 * Parses single entity lazy naive messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 * \param[out] vec feature vector of the entity
 */
void parse_se_lazy_naive_msg(message_data msg, int &view_id, T &vec) {
  std::stringstream parser;
  parser << (char *) msg;
  std::string parsed_word;
  
  //gets message type
  std::getline(parser, parsed_word, szDelim);
  //gets view_id
  std::getline(parser, parsed_word, szDelim);
  view_id = atoi(parsed_word.c_str());
  //gets vector_id
  std::getline(parser, parsed_word, szDelim);
  vec.fromPSQL(parsed_word);
}

/**
 * Parses view id in single entity lazy naive messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */
void parse_view_id_in_se_lazy_naive_msg(message_data msg, int &view_id);

// Templates
template<class T>
/**
 * Parses single entity lazy naive messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 * \param[out] vec feature vector of the entity
 */
void parse_se_lazy_naive_msg_bin(void* msg, int &view_id, T &vec) {
  struct se_lazy_naive_msg_bin_header* msg_header = (struct se_lazy_naive_msg_bin_header*)msg;
  assert(msg_header->t == READ_SE_LAZY_NAIVE_BIN);
  view_id          = msg_header->view_id;
  int payload_size = msg_header->payload_size;
  vec.read_bin((void*) (&msg_header[1]), payload_size);
}


/**
 * Parses view id in single entity lazy naive messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */
void parse_view_id_in_se_lazy_naive_msg_bin(void* msg, int &view_id);

/**
 * Parses single entity lazy hazy messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 * \param[out] vec feature vector of the entity
 * \param[out] eps epsilon of the entity
 */ 
template<class T>
void parse_se_lazy_hazy_msg(message_data msg, int &view_id, T &vec, double &eps) {
  std::stringstream parser;
  parser << (char *) msg;
  std::string parsed_word;
  
  //gets message type
  std::getline(parser, parsed_word, szDelim);
  //gets view_id
  std::getline(parser, parsed_word, szDelim);
  view_id = atoi(parsed_word.c_str());
  //gets vector_id
  std::getline(parser, parsed_word, szDelim);
  vec.fromPSQL(parsed_word);
  //gets eps
  std::getline(parser, parsed_word, szDelim);
  eps = atof(parsed_word.c_str());
}


/**
 * Parses view id in single entity lazy hazy messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */ 
void parse_view_id_in_se_lazy_hazy_msg(message_data msg, int &view_id);

// Templates
template<class T>
/**
 * Parses single entity lazy hazy messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 * \param[out] vec feature vector of the entity
 * \param[out] eps epsilon of the entity
 */
void parse_se_lazy_hazy_msg_bin(void* msg, int &view_id, T &vec, double &eps) {
  struct se_lazy_hazy_msg_bin_header* msg_header = (struct se_lazy_hazy_msg_bin_header*)msg;
  assert(msg_header->t == READ_SE_LAZY_HAZY_BIN);
  view_id          = msg_header->view_id;
  eps              = msg_header->eps;
  int payload_size = msg_header->payload_size;
  vec.read_bin((void*) (&msg_header[1]), payload_size);
}

/**
 * Parses single entity lazy hazy messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */
void parse_view_id_in_se_lazy_hazy_msg_bin(void* msg, int &view_id);

/**
 * Parses update messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 * \param[out] vec feature vector of the new example
 * \param[out] classOfExample class of the new example
 */
template<class T>
void parse_update_msg(message_data msg, int &view_id, T &vec, int &classOfExample) {
  std::stringstream parser;
  parser << (char *) msg;
  std::string parsed_word;
  
  //gets message type
  std::getline(parser, parsed_word, szDelim);
  //gets view_id
  std::getline(parser, parsed_word, szDelim);
  view_id = atoi(parsed_word.c_str());
  //gets vector_id
  std::getline(parser, parsed_word, szDelim);
  parse_labeled_vector<T>(parsed_word, classOfExample, vec);
}

/**
 * Parses update messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */
void parse_view_id_in_update_msg(message_data msg, int &view_id);

/**
 * Parses binary update messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 * \param[out] vec feature vector of the new example
 * \param[out] classOfExample class of the new example
 */
template<class T>
void parse_update_msg_bin(message_data msg, int &view_id, T &vec, int &classOfExample) {
  struct update_bin_header *msg_header = (struct update_bin_header*) msg;
  assert(msg_header->t == UPDATE_BIN);
  view_id              = msg_header->view_id;
  classOfExample       = msg_header->label;
  int payload_size     = msg_header->payload_size;
  vec.read_bin((void*) (&msg_header[1]), payload_size);
}

/**
 * Parses view id in binary update messages
 * \param[in] msg message to be parsed
 * \param[out] view_id id of the view
 */
void parse_view_id_in_update_msg_bin(message_data msg, int &view_id);

void parse_create_view_msg(message_data msg, struct create_view_bin_str *cv_struct);

#endif // _cpluspc


#endif
