#include <errno.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include "low_level_net.h"
#include <string.h>

#ifdef __cplusplus
extern "C"  {
#endif
  
  ///initial size
  int current_size = -1;
  
  ///wire message structure that contains length of the message which is sent before the msg itself
  struct wire_message {
    //DEBUG_ONLY(char debug_header[4] = "msg";);
    int length;
  };
  
  /**
   * Initializes buffer with the default size
   * \return buffer_handle buffer
   */
  struct buffer_handle
  initialize_buffer() {
    struct buffer_handle bh;
    bh.current_size = 4096 << 12; 
    bh.pData        = (void*) malloc(bh.current_size); 
    assert(bh.pData != NULL);
    return bh;
  }
  
  /**
   * Frees the buffer
   * \param[in] b buffer
   */
  void 
  destroy_handle(struct buffer_handle b) { 
    free(b.pData);
  }
  
  /**
   * Updates the buffer size with the max of given size and current size
   * \param[in] bh buffer
   * \param[in] new_size new size of the buffer
   */
  void
  update_buffer_size(struct buffer_handle *bh, int new_size) {
    if(new_size < bh->current_size) { return; }
    else {
      // we want it page aligned.
      while( bh->current_size > 0 && bh->current_size < new_size ) { bh->current_size = bh->current_size << 1; }
      void *p = realloc(bh->pData, bh->current_size);
      
      assert(p != NULL);
      bh->pData = p;
      printf("\t Buffer grows: %d\n", bh->current_size);
    }
  }
  
  /**
   * Retrieves length of wire message
   * \param[in] ret return value comes from socket communication
   * \param[in] wm wire message comes from socket communication
   * \return message length in success or negative value in failure
   */
  int get_message_length(int ret, struct wire_message *wm) {
    if(ret < 0) {
      // .. crash...
      printf("get message length crashed");
      assert(0);
      return ret;
    }
    else
      return wm->length;
  }
  
  // this does not malloc memory, but you should free the pointer that's returned 
  // except by destroying the handle.
  /**
   * Reads the msg from socket
   * \param[in] buff buffer
   * \param[in] fd socket id
   */
  void *
  read_wire_message(struct buffer_handle *buff, int fd) {
    // first item read better be a message
    int bytes_read            = recv(fd, buff->pData, buff->current_size, 0);
    if(bytes_read < 0) { 
      assert (0);
    }
    int message_payload_length = ((struct wire_message*) buff->pData)->length;
    assert(message_payload_length > 0);
    update_buffer_size(buff, message_payload_length);
    void *data_ret = ((char *) buff->pData) + sizeof(struct wire_message);
    
    int total_bytes_read             = bytes_read - sizeof(struct wire_message);
    while(total_bytes_read < message_payload_length) {
      int bytes_read = recv(fd, ((char *) data_ret + total_bytes_read), message_payload_length - total_bytes_read, 0);
      if(bytes_read == -1) {
	printf("Error at recv() : %s\n", strerror(errno));      
	assert(0);
      }
      total_bytes_read += bytes_read;
    }
    assert(total_bytes_read <= message_payload_length);

    ((char *) data_ret)[message_payload_length] = 0;
    return data_ret;
  }
  
  /**
   * Writes the msg to the socket
   * \param[in] fd socket id
   * \param[in] buf buffer
   * \param[in] nLength length of the msg
   * \return success of write
   */
  int
  write_wire_message(int fd, const void* buf, int nLength) {
    // break the buffer buffer into messages.
    struct wire_message m;
    m.length = nLength;
    assert(nLength > 0);
    
    // send the intial message
    int ret = send(fd, &m, sizeof(m), 0);  
    if(ret == -1 ) {
      return ret;
    }
    int current_index = 0;
    while(current_index < nLength) {
      int ret = send(fd, &(((char *) buf)[current_index]), nLength - current_index, 0);
      if(ret == -1) {
	return ret;
      }
      current_index += ret;
    }
    return nLength;
  }
  
#ifdef __cplusplus
}
#endif
