/**
 * \file hazy_sgd.h
 */

#ifndef HAZY_SGD_H
#define HAZY_SGD_H

//#include "sgd_model.h"
#include "skiing.h"
#include "reservoir.h"
#include "incremental_sgd.h"
#include "storage_manager.h"
#include "ondisk_storage_manager.h"
#include "hybrid_storage_manager.h"
#include "main_memory_storage_manager.h"
#include <pthread.h>

#include <string>

///Lambda value used for experiments and default for sgd code
#define LAMBDA 0.0001

/**
 * This class handles all queries sent by triggers
 * It uses locks for public functions to handle parallelism
 */
template <class T>
class Hazy_Sgd {
 protected:
  int _dim;
  
  ///Skiing object that maintains low & high water, acc cost and resort cost
  Skiing ski;
  
  ///Sgd model which is maintained
  IncrementalSGD<T> *sgd;  
  /**
   * Keeps track of the hazy model
   * Kept up-to-date by SGD code (so needs to be initialized after)
   * Also contains the low and high water values.
   */
  struct hazy_model hm;
  
  ///Storage manager object that updates related storages for entities
  Storage_Manager<T> *st_man;
  
  ///Fine-grained read-write lock
  pthread_rwlock_t _rwlock;
  
  ///The model that the entities are clustered on
  struct model _last_sort_model;

  typedef std::map<std::string, Hazy_Database*> _connection_map;
  _connection_map connections;
  
  /**
   * According to the strategy, updates storage manager
   */
  void updateModel();
  
  /**
   * Calculates differences between models
   * \param[out] delta_w the difference between feature vector of the model from last sort and feature vector of the current model
   * \param[out] delta_b the difference between bias of the model from last sort and bias of the current model
   */
  void holder_difference_models(double &delta_w, double &delta_b); 
  
  /**
   * LAZY_HAZY flags
   * if (_high_low_water_valid) then the high and the low water in the model are valid
   * the model will become invalidated on a lazy_hazy update.
   * if the model is invalid on a lazy_hazy read, then the high and low water need to be recomputed.
   */
  bool _high_low_water_valid; 
  /**
   * Calculates low & high water values
   */
  void  update_low_high_water() { // Implicitly uses the current model and the last sort model.
    double delta_bias = 0.0, delta_w = 0.0;
    holder_difference_models(delta_w, delta_bias);
    
    // Update the high and low water
    double cond_plus = -(delta_bias - delta_w);
    double cond_neg = -(delta_bias + delta_w);
    hm.low_water  = std::min(hm.low_water, cond_neg);
    hm.high_water = std::max(hm.high_water, cond_plus);
    LOGGING_ONLY(std::cout << "low & high water: " << hm.low_water << ", " << hm.high_water << std::endl;);
  }
 public:
  /**
   * Constructor for Hazy that uses default model and requires Storage Manager
   * \param[in] sm storage manager used in hazy sgd
   * \param[in] sgd sgd model
   * \param[in] ski skiing object that contains information about skiing strategy
   * \param[in] s strategy which is either LAZY_NAIVE, LAZY_HAZY, EAGER_NAIVE, EAGER_HAZY
   */
  Hazy_Sgd(Storage_Manager<T> *sm, IncrementalSGD<T> *sgd, Skiing &ski, hazy_model::strategy s);    
  /**
   * Updates model with given example and updates external results accordingly
   * \param[in] vec feature vector of given example
   * \param[in] label class of given example
   */
  void updateModel(T vec, sClass label);
  
  void deleteModel(std::string db_name, std::string table_name);

  /**
   * Gets entity id and gives class of entity
   * \param[in] entity_id id of the entity
   * \param[out] c class of the entity
   */
  void readEntityClass(key entity_id, sClass &c);
  /**
   * Gets feature vector of the entity and gives class of the entity. Used for lazy naive reads on on-disk
   * \param[in] vec feature vector of the entity
   * \param[out] label class of the entity
   */
  void readEntityClass(T vec, sClass &label);
  /**
   * Gets feature vector and eps of the entity and gives class of the entity. Used for lazy hazy reads on on-disk
   * \param[in] vec feature vector of the entity
   * \param[in] eps epsilon value of the entity
   * \param[out] label class of the entity
   */
  void readEntityClass(T vec, double eps, sClass &label);
  
  /**
   * Gives the # of entities that belong to the given class c
   * \param[in] c class that we are looking for
   * \param[out] nClass number of entities that belong to class c
   */
  void readNumInClass(sClass c, int &nClass);
};
#include "hazy_sgd.hxx"

/**
 * This class extends Hazy_Sgd and is specialized for Read experiments. It accepts low & high water as parameters and update them with the given values to restore a snapshot.
 */
template<class T>
class Read_Specialized_Hazy_Sgd : public Hazy_Sgd<T> {
 public:
  /**
   * Constructor for Hazy that uses default model and requires Storage Manager
   * \param[in] sm storage manager used in hazy sgd
   * \param[in] sgd sgd model
   * \param[in] ski skiing object that contains information about skiing strategy
   * \param[in] s strategy which is either LAZY_NAIVE, LAZY_HAZY, EAGER_NAIVE, EAGER_HAZY
   * \param[in] _low_water low water value
   * \param[in] _high_water high water value
   */
 Read_Specialized_Hazy_Sgd(Storage_Manager<T> *sm, IncrementalSGD<T> *sgd, Skiing &_ski, hazy_model::strategy s, 
			   double _low_water, double _high_water) : 
  Hazy_Sgd<T>(sm, sgd, _ski, s) {this->hm.low_water = _low_water; this->hm.high_water = _high_water;}
};

/**
 * This class extends Hazy_Sgd and is specialized for Multiclass experiments. It gives opportunity to explicitly resort.
 */
template<class T>
class Multiclass_Specialized_Hazy_Sgd : public Hazy_Sgd<T> {
 public:
  /**
   * Constructor for Hazy that uses default model and requires Storage Manager
   * \param[in] sm storage manager used in hazy sgd
   * \param[in] sgd sgd model
   * \param[in] ski skiing object that contains information about skiing strategy
   * \param[in] s strategy which is either LAZY_NAIVE, LAZY_HAZY, EAGER_NAIVE, EAGER_HAZY
   */
 Multiclass_Specialized_Hazy_Sgd(Storage_Manager<T> *sm, IncrementalSGD<T> *sgd, Skiing ski, hazy_model::strategy s) :
  Hazy_Sgd<T>(sm, sgd, ski, s) {}
  
  /**
   * Resorts the entities by calling resort method in storage manager and then adjusts ski strategy accordingly.
   */
  void resort() {
    Timer t(true);
    this->st_man->resort(this->hm);
    this->_last_sort_model = model(this->hm._model);
    
    this->ski.doResort(t.stop());
    this->hm.low_water = 0.0, this->hm.high_water = 0.0;
  }
};

#endif
