#include "postgres_util.h"
#include "storage_manager.h"
#include "ondisk_storage_manager.h"
#include "hybrid_storage_manager.h"
#include "main_memory_storage_manager.h"
#include "incremental_sgd.h"
#include "globals.h"
#include "frontend_util.h"
#include <errno.h>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>

using namespace std;

double lambda = 0.0001;
int rSize = 1000;

/**
 * This class is implemented to get low & high water values for read experiments
 * We first train a model with k examples
 * Then search on the entities to find low & high water values that produce # of tuples between low & high water which is p percent of all entities
 */

template <class T>
void
produce_models(Hazy_Database *db_conn, struct ondisk_storage_spec *od_spec, int numOfEntityBetweenLowHigh, char *training_file_name, char *output_file_name, int dim) {
  IncrementalSGD<T> *sgd = (IncrementalSGD<T>*) new ReservoirSGD<T>(dim, lambda, rSize);
  
  struct hazy_model hm(sgd->getModel(), hazy_model::EAGER_HAZY);
  hm.low_water = 0;
  hm.high_water = 0;
  
  Ondisk_Storage_Manager_Factory<T>      x;
  Storage_Manager<T>* sm          = (Storage_Manager<T>*) x.build_storage_manager(db_conn, od_spec, hm);
  
  
  ifstream our_file(training_file_name);
  
  if(our_file.is_open()) {
    std::string line;
    while(! our_file.eof() && getline(our_file, line)){
      int label;
      T vec;
      parse_labeled_vector<T>(line, label, vec);
      sgd->addExample(label, vec);
    }
    our_file.close();
  } else {
    cout << "Failed to open " << training_file_name << endl;
    exit(-1);
  }
  
  sm->resort(hm);
  cerr << "end of resort" << endl;
  
  int halfNumOfTuples = numOfEntityBetweenLowHigh / 2;
  
  ostringstream low_query;
  low_query << "SELECT eps FROM (SELECT eps FROM " << od_spec->intermediate_table_name << " WHERE eps < 0 ORDER BY eps DESC LIMIT " << halfNumOfTuples << ") as temp1 ORDER BY eps LIMIT 1;";
  ostringstream high_query;
  high_query << "SELECT eps FROM (SELECT eps FROM " << od_spec->intermediate_table_name << " WHERE eps > 0 ORDER BY eps LIMIT " << halfNumOfTuples << ") as temp1 ORDER BY eps DESC LIMIT 1;";
  
  double low_water = 0.0;
  double high_water = 0.0;
  
  db_conn->execute_query_msg_double(low_query.str().c_str(), "[produce_models] query low eps", low_water);
  db_conn->execute_query_msg_double(high_query.str().c_str(), "[produce_models] query high eps", high_water);
  cout << "low water: " << low_water << ", high water: " << high_water << endl;
  ofstream f(output_file_name);
  model m = sgd->getModel();
  // SCALE THE MODEL.   
  f << m;
}


int main(int argc, char **argv) {
  if(argc != 9) {
    cout << "usage: <dbname> <data_file> <training_file_name> <output_model> <dim> <dense> <entity_size> <percent>" << endl;
    exit(0);
  }
  
  char *db_name = argv[1];
  char *data_file = argv[2];
  char *training_file_name = argv[3];
  char *output_model_file_name = argv[4];
  int dim = atoi(argv[5]);
  int dense = atoi(argv[6]);
  int entity_size = atoi(argv[7]);
  int percent = atoi(argv[8]);	
  
  struct ondisk_storage_spec *od_spec = new ondisk_storage_spec();
  od_spec->unique_id               = 0;
  od_spec->entity_table_name       = std::string("sm_test_entity_table_name");
  od_spec->external_table_name     = std::string("sm_test_external_table");
  od_spec->intermediate_table_name = std::string("sm_test_intermediate_table");
  od_spec->bRebuild                = true;
  od_spec->strategy = hazy_model::EAGER_HAZY;
  
  Hazy_Database *db_conn = new Hazy_Database (db_name);
  
  int between_low_high = (entity_size * percent) / 100;
  
  if(dense) {
    string load_query = "DROP TABLE IF EXISTS " + od_spec->entity_table_name + " CASCADE;";
    load_query += "CREATE TABLE " + od_spec->entity_table_name + "(id int, feature_vector FLOAT8[]);";
    load_query += "COPY " + od_spec->entity_table_name + " FROM '" + data_file + "' DELIMITER ';'";
    db_conn->execute_statement_msg(load_query.c_str(), "[get_model_snapshot.cxx] in main: load query");
    produce_models<FVector>(db_conn, od_spec, between_low_high, training_file_name, output_model_file_name, dim);
  }
  else {
    string load_query = "DROP TABLE IF EXISTS " + od_spec->entity_table_name + " CASCADE;";
    load_query += "CREATE TABLE " + od_spec->entity_table_name + "(id int, feature_vector pair[]);";
    load_query += "COPY " + od_spec->entity_table_name + " FROM '" + data_file + "' DELIMITER ';'";
    db_conn->execute_statement_msg(load_query.c_str(), "[get_model_snapshot.cxx] in main: load query");
    produce_models<SVector>(db_conn, od_spec, between_low_high, training_file_name, output_model_file_name, dim);
  }
}
