#include "hazy_message.h"
#include "hazy_sgd_wrapper.h"
#include "postgres_util.h"
#include "server_conn.h"
#include "hashing.h"
#include "hazy_sgd.h"
#include "globals.h"
#include "frontend_util.h"
#include "parse_conf_file.h"
#include <errno.h>
#include <string>
#include <pthread.h>

#define HAZY_HYBRID "hazy_hybrid"
#define HAZY_ON_DISK "hazy_on_disk"
#define HAZY_MM "hazy_mm"

#define LAMBDA 0.0001

const string EXTERNAL_TABLE_BASE_NAME = "sm_test_external_table";
const string INTERNAL_TABLE_BASE_NAME = "internal_table";

typedef std::map<int, Hazy_Sgd_Wrapper*> _t_hazy_sgd_map;
_t_hazy_sgd_map hazy_sgd_map;
static pthread_mutex_t hazy_sgd_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * This class represents Hazy system. A user can create classification view and maintain it.
 */

Hazy_Database *db_conn;

void usage() {
  cerr << "usage: " << endl;
  return;
}

using namespace std;

///Hazy Server
Hazy_Server *serv;

void startServer(int server_id, string dbname);

int main(int argc, char **argv) {
  string dbname = "";	//req
  int server_id = 0;	//req

  string hazy_conf_file_path = retrieve_hazy_conf_file_path();

  int ret_val = parse_dbname_from_conf_file(hazy_conf_file_path, dbname);

  if(dbname.length() == 0) {
    cerr << "database name could not be parsed from hazy.conf" << endl;
    cerr << "exiting..." << endl;
    exit(1);
  }

  if(ret_val != SUCCESSFUL_PARSE) {
    handle_success_code(ret_val);
    exit(1);
  }

  hazy_sgd_map.clear();
  startServer(server_id, dbname);
  
  return 0;
}

struct worker_thread_info {
  Hazy_Worker *hw;
  int server_id;
  
  worker_thread_info(Hazy_Worker *_hw, int _server_id) {
    hw = _hw;
    server_id = _server_id;
  }
};

template <class T>
void process_client_requests(Hazy_Sgd<T> *param_sgd, void *data, int server_id, Hazy_Worker *hw, bool &close_conn) {
  Hazy_Sgd<T> *sgd = param_sgd;
  int nSize = 2048;
  message_data answer = (char *) malloc(nSize * sizeof(char));

  if(data == NULL) {
    return;
  }
    
  switch(get_message_type(data)) {
  case READ_SE_LAZY_NAIVE:
    {
      int view_id = 0;
      T vec;
      parse_se_lazy_naive_msg(data, view_id, vec);
      int entity_label = -1;
      sgd->readEntityClass(vec, entity_label);
      LOGGING_ONLY(cout << "label: " << entity_label << endl;);
	
      if(build_read_se_answer(entity_label, answer, nSize) < 0) {
	cerr << "problem in build msg" << endl;
	return;
      }
    }
    break;
  case READ_SE_LAZY_NAIVE_BIN:
    {
      int view_id = 0;
      T vec;
      parse_se_lazy_naive_msg_bin(data, view_id, vec);
      int entity_label = -1;
      sgd->readEntityClass(vec, entity_label);
      LOGGING_ONLY(cout << "label: " << entity_label << endl;);
	
      if(build_read_se_answer(entity_label, answer, nSize) < 0) {
	cerr << "problem in build msg" << endl;
	return;
      }
    }
    break;
  case READ_SE_LAZY_HAZY:
    {
      int view_id = 0;
      T vec;
      double eps = 0;
      parse_se_lazy_hazy_msg(data, view_id, vec, eps);
      int entity_label = -1;
      sgd->readEntityClass(vec, eps, entity_label);
      LOGGING_ONLY(cout << "label: " << entity_label << endl;);
	
      if(build_read_se_answer(entity_label, answer, nSize) < 0) {
	cerr << "problem in build msg";
	return;
      }
    }
    break;
  case READ_SE_LAZY_HAZY_BIN:
    { 
      int view_id = 0;
      T vec;
      double eps = 0;
      parse_se_lazy_hazy_msg_bin(data, view_id, vec, eps);
      int entity_label = -1;
      sgd->readEntityClass(vec, eps, entity_label);
      LOGGING_ONLY(cout << "label: " << entity_label << endl;);
	
      if(build_read_se_answer(entity_label, answer, nSize) < 0) {
	cerr << "problem in build msg";
	return;
      }
    } 
    break;
  case READ_SE_EAGER:
    {				
      LOGGING_ONLY(Timer rse(true););
      int view_id = 0;
      int entity_id = 0;
      parse_se_eager_msg(data, view_id, entity_id);
      int entity_label = -1;
      LOGGING_ONLY(Timer rec(true););
      sgd->readEntityClass(entity_id, entity_label);
      LOGGING_ONLY(cout << "[READ MESSAGE] label: " << entity_label << " in " << rec.stop() << endl;);
	
      if(build_read_se_answer(entity_label, answer, nSize) < 0) {
	cerr << "problem in build msg";
	return;
      }
      LOGGING_ONLY(cout << "[HAZY:MAIN]" << rse.stop() << endl;);
    }
    break;
  case READ_COUNT:
    {
      int view_id = 0;
      parse_count_msg(data, view_id);
      int numInClass = 0;
      sgd->readNumInClass(1, numInClass);
      LOGGING_ONLY(cout << "count: " << numInClass << endl;);
	
      if(build_read_count_answer(numInClass, answer, nSize) < 0) {
	cerr << "problem in build msg";
	return;
      }
    }
    break;
  case UPDATE:
    {
      int view_id = 0;
      T vec;
      int classOfExample = 0;
      parse_update_msg(data, view_id, vec, classOfExample);
      sgd->updateModel(vec, classOfExample);
	
      if(build_update_answer(answer, nSize) < 0) {
	cerr << "problem in build msg";
	return;
      }	
    }
    break;
  case UPDATE_BIN:
    {
      int view_id = 0;
      T vec;
      int classOfExample = 0;
      parse_update_msg_bin(data, view_id, vec, classOfExample);
      sgd->updateModel(vec, classOfExample);
	
      if(build_update_answer(answer, nSize) < 0) {
	cerr << "problem in build msg";
	return;
      }
    }   
    break;
  case DELETE_BIN:
    {
      int view_id = 0;
      int deleted_id = 0;
      string db_name = "";
      string rel_name = "";
      parse_delete_msg_bin(data, view_id, deleted_id, db_name, rel_name);
      sgd->deleteModel(db_name, rel_name);

      if(build_update_answer(answer, nSize) < 0) {
	cerr << "problem in build msg";
	return;
      }
    }
    break;
  case QUERY_SERVER_ID:
    {
      if(build_query_server_answer(server_id, answer, nSize) < 0) {
	cerr << "problem in build msg";
	return;
      }
    }
    break;
  case CLOSE_CONNECTION:
    {
      close_conn = true;
      if(build_close_connection_answer((int) close_conn, answer, nSize) < 0) {
	cerr << "problem in build msg";
	return;
      }
    }
    break;
  case SERVER_NOOP:
    {
      int ack = 1;
      if(build_noop_server_answer(ack, answer, nSize) < 0) {
	cerr << "problem in build msg";
	return;
      }
    }
    break;
  default: 
    cerr << "there is no such function code (" << get_message_type(data) << ")" << endl;
  }
    
  int success = hw->writeData(answer, strlen((char *) answer), false);
  if(success < 0)
    cerr << "error when writing to the client: " << strerror(errno);
  if(close_conn)
    return;

  while (1) {
    void *data = hw->readData();
    if(data == NULL) {
      break;
    }
    
    switch(get_message_type(data)) {
    case READ_SE_LAZY_NAIVE:
      {
  	int view_id = 0;
  	T vec;
  	parse_se_lazy_naive_msg(data, view_id, vec);
  	int entity_label = -1;
  	sgd->readEntityClass(vec, entity_label);
  	LOGGING_ONLY(cout << "label: " << entity_label << endl;);
	
  	if(build_read_se_answer(entity_label, answer, nSize) < 0) {
  	  cerr << "problem in build msg" << endl;
  	  return;
  	}
      }
      break;
    case READ_SE_LAZY_NAIVE_BIN:
      {
  	int view_id = 0;
        T vec;
        parse_se_lazy_naive_msg_bin(data, view_id, vec);
        int entity_label = -1;
        sgd->readEntityClass(vec, entity_label);
        LOGGING_ONLY(cout << "label: " << entity_label << endl;);
	
        if(build_read_se_answer(entity_label, answer, nSize) < 0) {
          cerr << "problem in build msg" << endl;
          return;
        }
      }
      break;
    case READ_SE_LAZY_HAZY:
      {
  	int view_id = 0;
  	T vec;
  	double eps = 0;
  	parse_se_lazy_hazy_msg(data, view_id, vec, eps);
  	int entity_label = -1;
  	sgd->readEntityClass(vec, eps, entity_label);
  	LOGGING_ONLY(cout << "label: " << entity_label << endl;);
	
  	if(build_read_se_answer(entity_label, answer, nSize) < 0) {
  	  cerr << "problem in build msg";
  	  return;
        }
      }
      break;
    case READ_SE_LAZY_HAZY_BIN:
      { 
        int view_id = 0;
        T vec;
        double eps = 0;
        parse_se_lazy_hazy_msg_bin(data, view_id, vec, eps);
        int entity_label = -1;
        sgd->readEntityClass(vec, eps, entity_label);
        LOGGING_ONLY(cout << "label: " << entity_label << endl;);
	
        if(build_read_se_answer(entity_label, answer, nSize) < 0) {
          cerr << "problem in build msg";
          return;
        }
      } 
      break;
    case READ_SE_EAGER:
      {				
  	LOGGING_ONLY(Timer rse(true););
  	int view_id = 0;
  	int entity_id = 0;
  	parse_se_eager_msg(data, view_id, entity_id);
  	int entity_label = -1;
  	LOGGING_ONLY(Timer rec(true););
  	sgd->readEntityClass(entity_id, entity_label);
  	LOGGING_ONLY(cout << "[READ MESSAGE] label: " << entity_label << " in " << rec.stop() << endl;);
	
  	if(build_read_se_answer(entity_label, answer, nSize) < 0) {
  	  cerr << "problem in build msg";
  	  return;
  	}
  	LOGGING_ONLY(cout << "[HAZY:MAIN]" << rse.stop() << endl;);
      }
      break;
    case READ_COUNT:
      {
  	int view_id = 0;
  	parse_count_msg(data, view_id);
  	int numInClass = 0;
  	sgd->readNumInClass(1, numInClass);
  	LOGGING_ONLY(cout << "count: " << numInClass << endl;);
	
      	if(build_read_count_answer(numInClass, answer, nSize) < 0) {
  	  cerr << "problem in build msg";
          return;
      	}
      }
      break;
    case UPDATE:
      {
  	int view_id = 0;
  	T vec;
  	int classOfExample = 0;
  	parse_update_msg(data, view_id, vec, classOfExample);
  	sgd->updateModel(vec, classOfExample);
	
  	if(build_update_answer(answer, nSize) < 0) {
  	  cerr << "problem in build msg";
          return;
      	}	
      }
      break;
    case UPDATE_BIN:
      {
        int view_id = 0;
        T vec;
        int classOfExample = 0;

        parse_update_msg_bin(data, view_id, vec, classOfExample);
        sgd->updateModel(vec, classOfExample);
	
        if(build_update_answer(answer, nSize) < 0) {
          cerr << "problem in build msg";
          return;
        }
      }   
      break;
    case DELETE_BIN:
      {
  	int view_id = 0;
  	int deleted_id = 0;
  	string db_name = "";
  	string rel_name = "";
  	parse_delete_msg_bin(data, view_id, deleted_id, db_name, rel_name);
  	sgd->deleteModel(db_name, rel_name);

        if(build_update_answer(answer, nSize) < 0) {
          cerr << "problem in build msg";
          return;
        }
      }
      break;
    case QUERY_SERVER_ID:
      {
  	if(build_query_server_answer(server_id, answer, nSize) < 0) {
  	  cerr << "problem in build msg";
          return;
  	}
      }
      break;
    case CLOSE_CONNECTION:
      {
  	close_conn = true;
  	if(build_close_connection_answer((int) close_conn, answer, nSize) < 0) {
  	  cerr << "problem in build msg";
  	  return;
  	}
      }
      break;
    case SERVER_NOOP:
      {
  	int ack = 1;
  	if(build_noop_server_answer(ack, answer, nSize) < 0) {
  	  cerr << "problem in build msg";
          return;
  	}
      }
      break;
    default: 
      cerr << "there is no such function code (" << get_message_type(data) << ")" << endl;
    }
    
    int success = hw->writeData(answer, strlen((char *) answer), false);
    if(success < 0)
      cerr << "error when writing to the client: " << strerror(errno);
    if(close_conn)
      break;
  }
}

void retrieve_view_id(int server_id, Hazy_Worker *hw, bool &close_conn) {
  int nSize = 2048;
  message_data answer = (char *) malloc(nSize * sizeof(char));

  int view_id = -1;
  void *data;
  
  while(view_id == -1) {
    data = hw->readData();
    
    if(data == NULL) {
      break;
    }
    
    switch(get_message_type(data)) {
      //If create view, then create view and put lock on general sgds table and insert it
    case READ_SE_LAZY_NAIVE:
      parse_view_id_in_se_lazy_naive_msg(data, view_id);
      break;
    case READ_SE_LAZY_NAIVE_BIN:
      parse_view_id_in_se_lazy_naive_msg_bin(data, view_id);
      break;
    case READ_SE_LAZY_HAZY:
      parse_view_id_in_se_lazy_hazy_msg(data, view_id);
      break;
    case READ_SE_LAZY_HAZY_BIN:
      parse_view_id_in_se_lazy_hazy_msg_bin(data, view_id);
      break;
    case READ_SE_EAGER:
      parse_view_id_in_se_eager_msg(data, view_id);
      break;
    case READ_COUNT:
      parse_count_msg(data, view_id);
      break;
    case UPDATE:
      parse_view_id_in_update_msg(data, view_id);
      break;
    case UPDATE_BIN: {
      parse_view_id_in_update_msg_bin(data, view_id);
    }
      break;
    case DELETE_BIN:
      parse_view_id_in_delete_msg_bin(data, view_id);
      break;
    case CREATE_VIEW:
      {
	struct create_view_bin_str *parsed_cv_struct = new struct create_view_bin_str();
	parse_create_view_msg(data, parsed_cv_struct);

	storage_spec *spec = NULL;
	if(parsed_cv_struct->sm == MAIN_MEMORY_SM)
	  spec = ((storage_spec*) new main_memory_storage_spec());
	else
	  spec = ((storage_spec*) new ondisk_storage_spec());
	
	spec->unique_id = parsed_cv_struct->view_id;
	spec->entity_table_name = parsed_cv_struct->entity_table_name;
	spec->bRebuild = true;
	spec->strategy = parsed_cv_struct->hazy_strategy;

	if(parsed_cv_struct->sm != MAIN_MEMORY_SM) {
	  ((ondisk_storage_spec*) spec)->external_table_name = parsed_cv_struct->view_name;
	  ostringstream internal_table_name_str;
	  internal_table_name_str << INTERNAL_TABLE_BASE_NAME << parsed_cv_struct->view_id;
	  ((ondisk_storage_spec*) spec)->intermediate_table_name = internal_table_name_str.str();
	}
	 
	if(parsed_cv_struct->dense) {
	  Hazy_Sgd<FVector> *sgd;

	  IncrementalSGD<FVector> *inc_sgd = (parsed_cv_struct->reservoir_size > 0) ? 
	    (IncrementalSGD<FVector> *) new ReservoirSGD<FVector>(parsed_cv_struct->dim, LAMBDA, parsed_cv_struct->reservoir_size) :
	    (IncrementalSGD<FVector> *) new TrueSGD<FVector>(parsed_cv_struct->dim, LAMBDA);

	  //train with the examples
	  std::vector< _id_label_vector_entry<FVector> > _tuples;
	  //we randomly choose training examples
	  std::string retrieve_query = "SELECT id, label, feature_vector FROM " + parsed_cv_struct->training_table_name
	    + " ORDER BY random();";
	  db_conn->postgresTupleParser<_id_label_vector_entry <FVector> >(retrieve_query.c_str(), _id_label_vector_entry<FVector>::parse_tuple,
									  _tuples);

	  bool model_changed = false;
	  for(unsigned int i = 0; i < _tuples.size(); i ++) {
	    int classOfExample = _tuples[i].label;
	    FVector featureArray = _tuples[i].x;
	    model_changed = inc_sgd->addExample(classOfExample, featureArray);
	  }
	  
	  hazy_model hm(inc_sgd->getModel(), spec->strategy);
	  if(model_changed)
	    hm.invalidate_db_model();

	  //storage manager
	  Storage_Manager<FVector> *st_manager;

	  if(parsed_cv_struct->sm == HYBRID_SM) {
	    Ondisk_Storage_Manager_Factory<FVector> x;
	    Ondisk_Storage_Manager<FVector> *disk = x.build_storage_manager(db_conn, (ondisk_storage_spec*) spec, hm);
	    st_manager = (Storage_Manager<FVector> *) new Hybrid_Storage_Manager<FVector>(db_conn, (ondisk_storage_spec*) spec, disk, 
											  parsed_cv_struct->buffer_size, parsed_cv_struct->useepsmap == 1);
	  }
	  else {
	    Ondisk_Storage_Manager_Factory<FVector> x;
	    st_manager = (parsed_cv_struct->sm == MAIN_MEMORY_SM) ? 
	      ((Storage_Manager<FVector> *) new STLMainMemory_Storage_Manager<FVector>(db_conn, (main_memory_storage_spec*) spec, hm)) : 
	      ((Storage_Manager<FVector> *) x.build_storage_manager(db_conn, (ondisk_storage_spec*) spec, hm));
	  }
	  //skiing spec
	  skiing_spec ski_spec;
	  ski_spec.initial_resort = parsed_cv_struct->resort_time;
	  ski_spec.taoVal = parsed_cv_struct->tao_val;
	  ski_spec.m_factor = parsed_cv_struct->m_factor;
	  ski_spec.ratio = parsed_cv_struct->alpha_ratio;
	  Skiing ski_obj(ski_spec);
	  
	  sgd = new Hazy_Sgd<FVector>(st_manager, inc_sgd, ski_obj, spec->strategy);
	  Hazy_Sgd_Wrapper *wrapper = new Hazy_Sgd_FVector(sgd);
	  //put lock and create the view
	  pthread_mutex_lock(&hazy_sgd_mutex);
	  hazy_sgd_map[parsed_cv_struct->view_id] = wrapper;
	  pthread_mutex_unlock(&hazy_sgd_mutex);
	} else {
	  Hazy_Sgd<SVector> *sgd;

	  IncrementalSGD<SVector> *inc_sgd = (parsed_cv_struct->reservoir_size > 0) ? 
	    (IncrementalSGD<SVector> *) new ReservoirSGD<SVector>(parsed_cv_struct->dim, LAMBDA, parsed_cv_struct->reservoir_size) :
	    (IncrementalSGD<SVector> *) new TrueSGD<SVector>(parsed_cv_struct->dim, LAMBDA);

	  //train with the examples
	  std::vector< _id_label_vector_entry<SVector> > _tuples;
	  std::string retrieve_query = "SELECT id, label, feature_vector FROM " + parsed_cv_struct->training_table_name + " ORDER BY random();";
	  db_conn->postgresTupleParser<_id_label_vector_entry <SVector> >(retrieve_query.c_str(), _id_label_vector_entry<SVector>::parse_tuple,
									  _tuples);

	  bool model_changed = false;

	  for(unsigned int i = 0; i < _tuples.size(); i ++) {
	    int classOfExample = _tuples[i].label;
	    SVector featureArray = _tuples[i].x;
	    model_changed = inc_sgd->addExample(classOfExample, featureArray);
	  }
	  	  
	  hazy_model hm(inc_sgd->getModel(), spec->strategy);
	  if(model_changed)
	    hm.invalidate_db_model();

	  //storage manager
	  Storage_Manager<SVector> *st_manager;

	  if(parsed_cv_struct->sm == HYBRID_SM) {
	    Ondisk_Storage_Manager_Factory<SVector> x;
	    Ondisk_Storage_Manager<SVector> *disk = x.build_storage_manager(db_conn, (ondisk_storage_spec*) spec, hm);
	    st_manager = (Storage_Manager<SVector> *) new Hybrid_Storage_Manager<SVector>(db_conn, (ondisk_storage_spec*) spec, disk, 
											  parsed_cv_struct->buffer_size, parsed_cv_struct->useepsmap == 1);
	  }
	  else {
	    Ondisk_Storage_Manager_Factory<SVector> x;
	    st_manager = (parsed_cv_struct->sm == MAIN_MEMORY_SM) ? 
	      ((Storage_Manager<SVector> *) new STLMainMemory_Storage_Manager<SVector>(db_conn, (main_memory_storage_spec*) spec, hm)) : 
	      ((Storage_Manager<SVector> *) x.build_storage_manager(db_conn, (ondisk_storage_spec*) spec, hm));
	  }
	  //skiing spec
	  skiing_spec ski_spec;
	  ski_spec.initial_resort = parsed_cv_struct->resort_time;
	  ski_spec.taoVal = parsed_cv_struct->tao_val;
	  ski_spec.m_factor = parsed_cv_struct->m_factor;
	  ski_spec.ratio = parsed_cv_struct->alpha_ratio;
	  Skiing ski_obj(ski_spec);
	  
	  sgd = new Hazy_Sgd<SVector>(st_manager, inc_sgd, ski_obj, spec->strategy);
	  Hazy_Sgd_Wrapper *wrapper = new Hazy_Sgd_SVector(sgd);
	  //put lock and create the view
	  pthread_mutex_lock(&hazy_sgd_mutex);
	  hazy_sgd_map[parsed_cv_struct->view_id] = wrapper;
	  pthread_mutex_unlock(&hazy_sgd_mutex);
	}

        if(build_create_view_answer(answer, nSize) < 0) {
          cerr << "problem in build msg";
          return;
        }

	int success = hw->writeData(answer, strlen((char *) answer), false);
	if(success < 0)
	  cerr << "error when writing to the client: " << strerror(errno);
	
	//delete parsed_cv_struct
	delete parsed_cv_struct;
      }
      break;
    case QUERY_SERVER_ID:
      {
  	if(build_query_server_answer(server_id, answer, nSize) < 0) {
  	  cerr << "problem in build msg";
          return;
  	}
      }
      break;
    case CLOSE_CONNECTION:
      {
  	close_conn = true;
  	if(build_close_connection_answer((int) close_conn, answer, nSize) < 0) {
  	  cerr << "problem in build msg";
  	  return;
  	}
      }
      break;
    case SERVER_NOOP:
      {
  	int ack = 1;
  	if(build_noop_server_answer(ack, answer, nSize) < 0) {
  	  cerr << "problem in build msg";
          return;
  	}
      }
      break;
    default:
      cerr << "there is no such function code (" << get_message_type(data) << ")" << endl;
    }
  }

  if(view_id >= 0) {
    //put lock and create the view
    pthread_mutex_lock(&hazy_sgd_mutex);
    _t_hazy_sgd_map::iterator i = hazy_sgd_map.find(view_id);
    if(i == hazy_sgd_map.end()) {
      cerr << "Not created view is queried. Problem!!!" << endl;
      pthread_mutex_unlock(&hazy_sgd_mutex);
    }
    else {
      Hazy_Sgd_Wrapper *hazy_sgd_wrapper = i->second;
      pthread_mutex_unlock(&hazy_sgd_mutex);
      //Hazy Sgd type is FVector
      if(hazy_sgd_wrapper->is_sgd_dense()) {
	Hazy_Sgd<FVector> *sgd = ((Hazy_Sgd_FVector *) hazy_sgd_wrapper)->get_Hazy_Sgd();
	process_client_requests(sgd, data, server_id, hw, close_conn);
      }
      //Hazy Sgd type is SVector
      else {
	Hazy_Sgd<SVector> *sgd = ((Hazy_Sgd_SVector *) hazy_sgd_wrapper)->get_Hazy_Sgd();
	process_client_requests(sgd, data, server_id, hw, close_conn);
      }
    }
  }
  else {
    cerr << "Invalid view id!!!" << endl;
  }
}

void*
worker_thread(void *data) {
  worker_thread_info *wit = (worker_thread_info *) data;
  Hazy_Worker *hw = wit->hw;

  int server_id = wit->server_id;
  bool close_conn = false;

  retrieve_view_id(server_id, hw, close_conn);
  
  ///Server's client connection is closed
  hw->closeClientConnection();
  delete hw;
  if(close_conn) {
    ///Server connection is closed
    cout << "server is being closed" << endl;
    serv->closeConnection();
    exit(0);
  }
  return NULL;
}

void startServer(int server_id, string dbname) {
  ///database connector
  db_conn = new Hazy_Database(dbname);
  vector<pthread_t> workers;
  
  ///Server is initialized
  serv = new Hazy_Server();
  cout << "server started" << endl;
  
  while(1) {
    Hazy_Worker* hw = serv->acceptConnection();
    if (hw == NULL) {
      printf("Error at accept() : %s\n", strerror(errno));
      continue;
    }
    
    pthread_t t;
    worker_thread_info *wit = new worker_thread_info(hw, server_id);
    int iret = pthread_create(&t, NULL, worker_thread, (void *) wit);
    if(iret != 0 ) {
      cerr << "pthread create failed " << iret << endl;
      break;
    }
    workers.push_back(t);
  }
  
  ///Server connection is closed
  serv->closeConnection();
  ///Database connection is closed
  db_conn->closeConnection();
}
