#include "hazy_message.h"
#include "vectors.h"
#include "postgres_util.h"
#include <iostream>
#include <string>
#include <math.h>
using namespace std;

/**
 * This class implements regression tests for hazy message class
 * For each query in hazy message, a test first fills both SVector and FVector lists to be used
 * Then, it creates a msg for that query and check that msg is correctly created.
 * Then, tests that we correctly retrieve message code
 * Finally, tests whether we correctly parse elements or not
 */
const int numSVectors = 50;
int dim = 10000;
int max_svector_dim = 10000;

const int test_num = numSVectors;
const int range = 100;

FVector fVectorList[numSVectors];
SVector sVectorList[numSVectors];

/**
 * Converts FVector to dbstring without using cast as float8[]
 * \param[in] v FVector
 * \param[out] feature_values dbstring which is created
 */
void
fvector_to_dbstring_without_cast(const FVector &v, std::string &feature_values) {
  std::ostringstream values;
  
  std::ios::fmtflags oldflags = values.flags();
  std::streamsize oldprec = values.precision();
  values << std::scientific << std::setprecision(sizeof(VFloat)==4 ? 7 : 16);
  
  values << "{";
  for(int i = 0; i < v.size(); i ++)
    values << v.get(i) << ((i < v.size() - 1) ? "," : "");
  
  values << "}";
  
  feature_values = std::string(values.str());
  
  values.precision(oldprec);
  values.flags(oldflags);
  
}

/**
 * Converts SVector to dbstring without using cast as pair[]
 * \param[in] s SVector
 * \param[out] feature_vector_str dbstring which is created
 */
void
svector_to_dbstring_without_cast(const SVector &s, std::string& feature_vector_str) {
  std::ostringstream feature_vector;
  
  std::ios::fmtflags oldflags = feature_vector.flags();
  std::streamsize oldprec = feature_vector.precision();
  feature_vector << std::scientific << std::setprecision(sizeof(VFloat)==4 ? 7 : 16);
  
  int n = s.npairs();
  feature_vector << "{";
  for(const SVector::Pair *p = s; p->i >= 0; p ++) {
    feature_vector << "\"(" << p->i << "," << p->v << ")\"" << ((n > 1) ? "," : "");
    n--;
  }
  feature_vector << "}";
  
  feature_vector_str = std::string(feature_vector.str());
  
  feature_vector.precision(oldprec);
  feature_vector.flags(oldflags);
}

/**
 * Randomly fill fvector and svector lists
 */
void fillVectors() {
  for(int i = 0; i < numSVectors; i ++) {
    int svector_dim = int(drand48() * max_svector_dim);
    while(svector_dim == 0)
      svector_dim = int(drand48() * max_svector_dim);
    
    for(int j = 0; j < svector_dim; j ++) {
      sVectorList[i].set(int(drand48() * dim), ((VFloat) drand48() * 2 -1));
    }
    for(int j = 0; j < dim; j ++)
      fVectorList[i].set(j, ((VFloat) drand48() * 2 - 1));
  }
  
}

/**
 * Tests count hazy message method
 * \param[out] success
 */
void test_count(bool &success) {
  success = true;
  int nSize        = 10000000;
  message_data msg = (char*) malloc(nSize * sizeof(char));
  for(int i = 0; i < test_num; i ++) {
    int view_id = (int) (drand48() * range);
    
    int res = build_count_msg(view_id,msg, nSize);
    if(res < 0) {
      cout << "Message building fail" << endl;
      success = false;
      break;
    }
    
    if(get_message_type(msg) != READ_COUNT) {
      cout << "fail: msg type is: " << get_message_type(msg) << endl;
      success = false;
      break;
    }
    
    int obtained_view_id = 0;
    parse_count_msg(msg, obtained_view_id);
    if(obtained_view_id != view_id) {
      cout << "fail in view id" << endl;
      cout << "view id: " << view_id << ", obtained view id: " << obtained_view_id << endl;
      success = false;
      break;
    }
    
    int count_num = (int) (drand48() * range);
    ostringstream test_count;
    test_count << count_num << "|";
    
    res = build_read_count_answer(count_num, msg, nSize);
    if(res < 0) {
      cout << "Message building fail" << endl;
      success = false;
      break;
    }
    
    
    if(!test_count.str().compare((char *) msg) == 0) {
      cout << "fail in prepared answer comparison" << endl;
      cout << "msg: -" << test_count.str() << "-, obtained msg: -" << msg << "-" << endl;
      success = false;
      break;
    }		
  }
  free(msg);
}

/**
 * Tests that update message is working correctly
 * \param[in] vec feature vector of the entity
 * \param[out] success
 */
template <class T>
void single_test_update(T vec, bool &success) {
  int view_id = (int) (drand48() * range);
  int label = (int) (drand48() * 2);
  ostringstream ostr;
  ostr << label << " " << vec;
  
  int nSize        = 10000000;
  message_data msg = (char*) malloc(nSize * sizeof(char));
  
  string vec_str = ostr.str().substr(0, ostr.str().size() - 1);	//remove new line
  
  if(build_update_msg(view_id, vec_str.c_str(), msg, nSize) < 0) {
    cout << "Building update message failed" << endl;
    success = false;
    return;
  }
  
  if(get_message_type(msg) != UPDATE) {
    cout << "msg type in update is not correct" << endl;
    success = false;
    return;
  }
  
  int obtained_view_id = 0;
  T obtained_vec;
  int obtained_label = 0;
  
  parse_update_msg(msg, obtained_view_id, obtained_vec, obtained_label);
  if(view_id != obtained_view_id || vec.size() != obtained_vec.size() || label != obtained_label) {
    cout << "view id, vec size or label of vec is not correct";
    cout << "view id: " << view_id << ", obtained view id: " << obtained_view_id << endl;
    cout << "vec size: " << vec.size() << ", obtained vec size: " << obtained_vec.size() << endl;
    cout << "label: " << label << ", obtained label: " << obtained_label << endl;
    success = false;
    return;
  }
  
  for(int i = 0; i < vec.size(); i ++) {
    if(vec.get(i) != obtained_vec.get(i)) {
      cout << "vector content differ: " << vec.get(i) << ", " << obtained_vec.get(i) << endl;
      success = false;
      return;
    }
  }
  if(build_update_answer(msg, nSize) < 0) {
    success = false;
    cout << "Building update answer message failed" << endl;
    return;
  }
  
  string ret_msg = "done|";
  if(ret_msg.compare((char *) msg) != 0) {
    cout << "fail in prepared answer comparison" << endl;
    cout << ret_msg << ", " << msg << endl;
    success = false;
    return;
  }
  
  free(msg);
}


/**
 * Tests that single entity read lazy naive message is working correctly
 * \param[in] vec feature vector of the entity
 * \param[out] success
 * \param[in] dense whether vector is sparse or dense
 */
template <class T>
void single_test_se_lazy_naive(T vec, bool &success, bool dense) {
  int view_id = (int) (drand48() * range);
  int nSize        = 10000000;
  message_data msg = (char*) malloc(nSize * sizeof(char));
  
  
  string vec_str = "";
  if(dense)
    fvector_to_dbstring_without_cast(vec, vec_str);
  else
    svector_to_dbstring_without_cast(vec, vec_str);
  
  if(build_se_lazy_naive_msg(view_id, vec_str.c_str(), msg, nSize) < 0) {
    cout << "build failed" << endl;
    success = false;
    return;
  }
  if(get_message_type(msg) != READ_SE_LAZY_NAIVE) {
    cout << "fail: msg type is: " << get_message_type(msg) << endl;
    success = false;
    return;
  }
  
  int obtained_view_id = 0;
  T obtained_vec;
  
  parse_se_lazy_naive_msg(msg, obtained_view_id, obtained_vec);
  
  if(view_id != obtained_view_id || vec.size() != obtained_vec.size()) {
    cout << "fail in view id or vec size: " << endl;
    cout << "view id: " << view_id << ", obtained view id: " << obtained_view_id << endl;
    cout << "vec size: " << vec.size() << ", obtained vec size: " << obtained_vec.size() << endl;
    success = false;
    return;
  }
  for(int i = 0; i < vec.size(); i ++) {
    if(vec.get(i) != obtained_vec.get(i)) {
      cout << "vector content differ: " << vec.get(i) << ", " << obtained_vec.get(i) << endl;
      success = false;
      return;
    }
  }
  
  int entity_label = (int) (drand48() * 2);
  ostringstream test_label;
  test_label << entity_label << "|";
  
  if(build_read_se_answer(entity_label, msg, nSize) < 0) {
    cout << "build failed" << endl;
    success = false;
    return;
  }
  
  if(test_label.str().compare((char *) msg) != 0) {
    cout << "fail in prepared answer comparison" << endl;
    cout << "msg: -" << test_label.str() << "-, obtained msg: -" << msg << "-" << endl;
    success = false;
    return;
  }
  
  free(msg);
}

/**
 * Tests that single entity read lazy hazy is working correctly
 * \param[in] vec feature vector of the entity
 * \param[out] success
 * \param[in] dense whether feature vector is dense or sparse
 */
template <class T>
void single_test_se_lazy_hazy(T vec, bool &success, bool dense) {
  int view_id = (int) (drand48() * range);
  double eps = drand48() * 2 - 1;
  int nSize        = 10000000;
  message_data msg = (char*) malloc(nSize * sizeof(char));
  
  string vec_str = "";
  if(dense)
    fvector_to_dbstring_without_cast(vec, vec_str);
  else
    svector_to_dbstring_without_cast(vec, vec_str);
  
  if(build_se_lazy_hazy_msg(view_id, vec_str.c_str(), eps, msg, nSize) < 0) {
    cout << "msg failed" << endl;
    success = false;
    return;
  }
  if(get_message_type(msg) != READ_SE_LAZY_HAZY) {
    cout << "fail: msg type is: " << get_message_type(msg) << endl;
    success = false;
    return;
  }
  
  int obtained_view_id = 0;
  T obtained_vec;
  double obtained_eps = 0;
  
  parse_se_lazy_hazy_msg(msg, obtained_view_id, obtained_vec, obtained_eps);
  
  if(view_id != obtained_view_id || fabs(eps - obtained_eps) > 1.0e-16 || vec.size() != obtained_vec.size()) {
    cout << "fail in view id or eps: " << endl;
    cout << "view id: " << view_id << ", obtained view id: " << obtained_view_id << endl;
    cout << "eps: " << eps << ", obtained eps: " << obtained_eps << endl;
    cout << "vec size: " << vec.size() << ", obtained vec size: " << obtained_vec.size() << endl;
    success = false;
    return;
  }
  
  for(int i = 0; i < vec.size(); i ++) {
    if(vec.get(i) != obtained_vec.get(i)) {
      cout << "vector content differ: " << vec.get(i) << ", " << obtained_vec.get(i) << endl;
      success = false;
      return;
    }
  }
  
  int entity_label = (int) (drand48() * 2);
  ostringstream test_label;
  test_label << entity_label << "|";
  
  if(build_read_se_answer(entity_label, msg, nSize) < 0) {
    cout << "build failed" << endl;
    success = false;
    return;
  }
  
  if(test_label.str().compare((char *) msg) != 0) {
    cout << "fail in prepared answer comparison" << endl;
    cout << "msg: -" << test_label.str() << "-, obtained msg: -" << msg << "-" << endl;
    success = false;
    return;
  }
  
  free(msg);
}

/**
 * Tests that update method is working correctly
 * \param[out] success
 */
void test_update(bool &success) {
  success = true;
  fillVectors();
  
  for(int i = 0; i < test_num; i ++) {
    single_test_update(fVectorList[i], success);
    if(!success)
      break;
    single_test_update(sVectorList[i], success);
    if(!success)
      break;
  }
}


/**
 * Tests that single entity read lazy hazy is working correctly
 * \param[out] success
 */
void test_se_lazy_hazy(bool &success) {
  success = true;
  fillVectors();
  
  for(int i = 0; i < test_num; i ++) {
    single_test_se_lazy_hazy(fVectorList[i], success, true);
    if(!success)
      break;
    single_test_se_lazy_hazy(sVectorList[i], success, false);
    if(!success)
      break;
  }
}

/**
 * Tests that single entity read lazy naive is working correctly
 * \param[out] success
 */
void test_se_lazy_naive(bool &success) {
  success = true;
  fillVectors();
  
  for(int i = 0; i < test_num; i ++) {
    single_test_se_lazy_naive(fVectorList[i], success, true);
    if(!success)             
      break;                 
    single_test_se_lazy_naive(sVectorList[i], success, false);
    if(!success)             
      break;                 
  }   
}

/**
 * Tests that query server id is working correctly
 * \param[out] success
 */
void test_query_id(bool &success) {
  success = true;
  int nSize = 10000000;
  message_data msg = (char *) malloc(nSize * sizeof(char));

  for(int i = 0; i < test_num; i ++) {
    if(build_query_server_id_msg(msg, nSize) < 0) {
      cout << "build failed" << endl;
      success = false;
      break;
    }
    
    if(get_message_type(msg) != QUERY_SERVER_ID) {
      cout << "fail: msg type is: " << get_message_type(msg) << endl;
      cout << "msg: " << msg << endl;
      success = false;
      break;
    }
    
    int id = time ( NULL );
    ostringstream test_str;
    test_str << id << "|";
    
    if(build_query_server_answer(id, msg, nSize) < 0) {
      cout << "build failed" << endl;
      success = false;
      break;
    }
    
    if(test_str.str().compare((char *) msg) != 0) {
      cout << "fail in prepared answer comparison" << endl;
      cout << "msg: -" << test_str.str() << "-, obtained msg: -" << msg << "-" << endl;
      success = false;
      break;
    }
  }
}


/**
 * Tests that close connection query is working correctly
 * \param[out] success
 */
void test_close_connection(bool &success) {
  success = true;
  int nSize = 10000000;
  message_data msg = (char *) malloc(nSize * sizeof(char));
  
  for(int i = 0; i < test_num; i ++) {
    if(build_close_connection_msg(msg, nSize) < 0) {
      cout << "build failed" << endl;
      success = false;
      break;
    }
    
    if(get_message_type(msg) != CLOSE_CONNECTION) {
      cout << "fail: msg type is: " << get_message_type(msg) << endl;
      cout << "msg: " << msg << endl;
      success = false;
      break;
    }
    
    int close_conn = 1;
    ostringstream test_str;
    test_str << close_conn << "|";
    
    if(build_close_connection_answer(close_conn, msg, nSize) < 0) {
      cout << "build failed" << endl;
      success = false;
      break;
    }
    
    if(test_str.str().compare((char *) msg) != 0) {
      cout << "fail in prepared answer comparison" << endl;
      cout << "msg: -" << test_str.str() << "-, obtained msg: -" << msg << "-" << endl;
      success = false;
      break;
    }
  }
}

void create_random_string(char *s, size_t len) {
  static const char alphanum[] =
    "0123456789"
    "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    "abcdefghijklmnopqrstuvwxyz";
  
  for(unsigned int i = 0; i < len - 1; i ++)
    s[i] = alphanum[rand() % (sizeof(alphanum) - 1)];

  s[len - 1] = 0;
}

void test_create_view(bool &success) {
  success = true;
  for(int i = 0; i < test_num; i ++) {
    int nSize = 50;
    message_data answer = (char *) malloc(nSize * sizeof(char));
    struct create_view_bin_str *cv_struct = new struct create_view_bin_str();
    //set view id
    cv_struct->view_id = rand() % range;
    //set storage manager
    int rand_num = rand() % 3;
    if(rand_num == 0)
      cv_struct->sm = MAIN_MEMORY_SM;
    else if(rand_num == 1)
      cv_struct->sm = ON_DISK_SM;
    else
      cv_struct->sm = HYBRID_SM;
    //set dense or not
    rand_num = rand() % 2;
    cv_struct->dense = rand_num;
    //set useepsmap
    if(cv_struct->sm == HYBRID_SM) {
      rand_num = rand() % 2;
      cv_struct->useepsmap = rand_num;
    }
    else
      cv_struct->useepsmap = 0;
    //set hazy_strategy
    rand_num = rand() % 4;
    switch(rand_num) {
    case 0: cv_struct->hazy_strategy = hazy_model::LAZY_NAIVE;
      break;
    case 1: cv_struct->hazy_strategy = hazy_model::LAZY_HAZY;
      break;
    case 2: cv_struct->hazy_strategy = hazy_model::EAGER_NAIVE;
      break;
    case 3: cv_struct->hazy_strategy = hazy_model::EAGER_HAZY;
      break;
    default: cv_struct->hazy_strategy = hazy_model::EAGER_HAZY;
    }
    //set reservoir_size
    cv_struct->reservoir_size = rand() % range;
    //set resort_time
    cv_struct->resort_time = drand48() * 10.0;
    //set tao_val
    cv_struct->tao_val = drand48() * 0.01;
    //set alpha_ratio
    cv_struct->alpha_ratio = drand48() * 1.0;
    //set buffer_size
    if(cv_struct->sm == HYBRID_SM) {
      cv_struct->buffer_size = rand() % range;
    }
    else
      cv_struct->buffer_size = 0;
    //set view_name
    size_t view_name_size = rand() % 1000;
    char *view_name = (char *) malloc(sizeof(char) * view_name_size);
    create_random_string(view_name, view_name_size);
    cv_struct->view_name = view_name;
    //set view_key
    size_t view_key_size = rand() % 1000;
    char *view_key = (char *) malloc(sizeof(char) * view_key_size);
    create_random_string(view_key, view_key_size);
    cv_struct->view_key = view_key;
    //set entity_table_name
    size_t entity_table_name_size = rand() % 1000;
    char *entity_table_name = (char *) malloc(sizeof(char) * entity_table_name_size);
    create_random_string(entity_table_name, entity_table_name_size);
    cv_struct->entity_table_name = entity_table_name;
    //set entity_table_key
    size_t entity_table_key_size = rand() % 1000;
    char *entity_table_key = (char *) malloc(sizeof(char) * entity_table_key_size);
    create_random_string(entity_table_key, entity_table_key_size);
    cv_struct->entity_table_key = entity_table_key;
    //set training_table_name
    size_t training_table_name_size = rand() % 1000;
    char *training_table_name = (char *) malloc(sizeof(char) * training_table_name_size);
    create_random_string(training_table_name, training_table_name_size);
    cv_struct->training_table_name = training_table_name;
    //set dim
    cv_struct->dim = rand() % range;
    //set db_name
    size_t db_name_size = rand() % 1000;
    char *db_name = (char *) malloc(sizeof(char) * db_name_size);
    create_random_string(db_name, db_name_size);
    cv_struct->db_name = db_name;
    //set m_factor
    double m_factor = 1.0;
    cv_struct->m_factor = m_factor;

    int msgSize        = sizeof(struct create_view_bin_str) + cv_struct->view_name.size() + cv_struct->view_key.size() + 
      cv_struct->entity_table_name.size() + cv_struct->entity_table_key.size() + cv_struct->training_table_name.size() +
      cv_struct->db_name.size() + 1000;
    message_data msg = (char*) malloc(msgSize * sizeof(char));
    
    if(build_create_view_msg(cv_struct, msg, msgSize) < 0) {
      cout << "Building create view message failed" << endl;
      success = false;
      return;
    }
    
    if(get_message_type(msg) != CREATE_VIEW) {
      cout << "msg type in update is not correct" << endl;
      success = false;
      return;
    }
    
    struct create_view_bin_str *parsed_cv_struct = new struct create_view_bin_str();
    parse_create_view_msg(msg, parsed_cv_struct);

    if(!(cv_struct->view_id == parsed_cv_struct->view_id && cv_struct->sm == parsed_cv_struct->sm &&
	 cv_struct->dense == parsed_cv_struct->dense && cv_struct->useepsmap == parsed_cv_struct->useepsmap &&
	 cv_struct->hazy_strategy == parsed_cv_struct->hazy_strategy && cv_struct->reservoir_size == parsed_cv_struct->reservoir_size &&
	 cv_struct->resort_time == parsed_cv_struct->resort_time && (fabs(cv_struct->tao_val == parsed_cv_struct->tao_val) < 1.0e-05) &&
	 cv_struct->alpha_ratio == parsed_cv_struct->alpha_ratio && cv_struct->buffer_size == parsed_cv_struct->buffer_size &&
	 (cv_struct->view_name.compare(parsed_cv_struct->view_name) == 0) && 
	 (cv_struct->view_key.compare(parsed_cv_struct->view_key) == 0) &&
	 (cv_struct->entity_table_name.compare(parsed_cv_struct->entity_table_name) == 0) &&
	 (cv_struct->entity_table_key.compare(parsed_cv_struct->entity_table_key) == 0) &&
	 (cv_struct->training_table_name.compare(parsed_cv_struct->training_table_name) == 0) &&
	 cv_struct->dim == parsed_cv_struct->dim && (cv_struct->db_name.compare(parsed_cv_struct->db_name) == 0) &&
	 cv_struct->m_factor == parsed_cv_struct->m_factor)) {      
      bool t = (cv_struct->view_id == parsed_cv_struct->view_id && cv_struct->sm == parsed_cv_struct->sm &&
	 cv_struct->dense == parsed_cv_struct->dense && cv_struct->useepsmap == parsed_cv_struct->useepsmap &&
	 cv_struct->hazy_strategy == parsed_cv_struct->hazy_strategy && cv_struct->reservoir_size == parsed_cv_struct->reservoir_size &&
	 cv_struct->resort_time == parsed_cv_struct->resort_time && (fabs(cv_struct->tao_val == parsed_cv_struct->tao_val) < 1.0e-05) &&
	 cv_struct->alpha_ratio == parsed_cv_struct->alpha_ratio && cv_struct->buffer_size == parsed_cv_struct->buffer_size &&
	 (cv_struct->view_name.compare(parsed_cv_struct->view_name) == 0) && 
	 (cv_struct->view_key.compare(parsed_cv_struct->view_key) == 0) &&
	 (cv_struct->entity_table_name.compare(parsed_cv_struct->entity_table_name) == 0) &&
	 (cv_struct->entity_table_key.compare(parsed_cv_struct->entity_table_key) == 0) &&
	 (cv_struct->training_table_name.compare(parsed_cv_struct->training_table_name) == 0) &&
	 cv_struct->dim == parsed_cv_struct->dim && (cv_struct->db_name.compare(parsed_cv_struct->db_name) == 0) &&
		cv_struct->m_factor == parsed_cv_struct->m_factor);
      cout << "t is: " << t << endl;
      cout << "fabs is: " << (fabs(cv_struct->tao_val == parsed_cv_struct->tao_val) < 1.0e-05) << endl;

      cout << "[test_create_view] fail in create_view_bin struct comparison" << endl;
      cout << "[test_create_view] view_id: " << cv_struct->view_id << ", parsed view_id: " << parsed_cv_struct->view_id << endl;
      cout << "[test_create_view] sm: " << cv_struct->sm << ", parsed sm: " << parsed_cv_struct->sm << endl;
      cout << "[test_create_view] dense: " << cv_struct->dense << ", parsed dense: " << parsed_cv_struct->dense << endl;
      cout << "[test_create_view] useepsmap: " << cv_struct->useepsmap << ", parsed useepsmap: " << parsed_cv_struct->useepsmap << endl;
      cout << "[test_create_view] hazy_strategy: " << cv_struct->hazy_strategy << ", parsed hazy_strategy: " << parsed_cv_struct->hazy_strategy << endl;
      cout << "[test_create_view] reservoir_size: " << cv_struct->reservoir_size << ", parsed reservoir_size: " << parsed_cv_struct->reservoir_size << endl;
      cout << "[test_create_view] resort_time: " << cv_struct->resort_time << ", parsed resort_time: " << parsed_cv_struct->resort_time << endl;
      cout << "[test_create_view] tao_val: " << cv_struct->tao_val << ", parsed tao_val: " << parsed_cv_struct->tao_val << endl;
      cout << "[test_create_view] alpha_ratio: " << cv_struct->alpha_ratio << ", parsed alpha_ratio: " << parsed_cv_struct->alpha_ratio << endl;
      cout << "[test_create_view] buffer_size: " << cv_struct->buffer_size << ", parsed buffer_size: " << parsed_cv_struct->buffer_size << endl;
      cout << "[test_create_view] view_name: " << cv_struct->view_name << ", parsed view_name: " << parsed_cv_struct->view_name << endl;
      cout << "[test_create_view] view_key: " << cv_struct->view_key << ", parsed view_key: " << parsed_cv_struct->view_key << endl;
      cout << "[test_create_view] entity_table_name: " << cv_struct->entity_table_name << ", parsed entity_table_name: " << parsed_cv_struct->entity_table_name << endl;
      cout << "[test_create_view] entity_table_key: " << cv_struct->entity_table_key << ", parsed entity_table_key: " << parsed_cv_struct->entity_table_key << endl;
      cout << "[test_create_view] training_table_name: " << cv_struct->training_table_name << ", parsed training_table_name: " << parsed_cv_struct->training_table_name << endl;
      cout << "[test_create_view] dim: " << cv_struct->dim << ", parsed dim: " << parsed_cv_struct->dim << endl;
      cout << "[test_create_view] db_name: " << cv_struct->db_name << ", parsed db_name: " << parsed_cv_struct->db_name << endl;
      cout << "[test_create_view] m_factor: " << cv_struct->m_factor << ", parsed m_factor: " << parsed_cv_struct->m_factor << endl;
      success = false;
      break;
    }

    if(build_create_view_answer(answer, nSize) < 0) {
      success = false;
      cout << "Building create view answer message failed" << endl;
      return;
    }
    
    string ret_msg = "done|";
    if(ret_msg.compare((char *) answer) != 0) {
      cout << "fail in prepared answer comparison" << endl;
      cout << ret_msg << ", " << answer << endl;
      success = false;
      return;
    }
  }
}

void test_delete_bin(bool &success) {
  success = true;
  
  for(int i = 0; i < test_num; i ++) {    
    int nSize = 50;
    message_data answer = (char *) malloc(nSize * sizeof(char));
    size_t data_out_size = sizeof(int) * 10000000;
    void *data_out = malloc(data_out_size);

    size_t db_name_size = rand() % 1000;
    size_t table_name_size = rand() % 1000;

    char *db_name = (char *) malloc(sizeof(char) * db_name_size);
    char *table_name = (char *) malloc(sizeof(char) * table_name_size);

    create_random_string(db_name, db_name_size);
    create_random_string(table_name, table_name_size);

    int view_id = rand() % range;
    int deleted_id = rand() % range;

    size_t out_size;
    int data_size = db_name_size + table_name_size + 1000; //1000 is default

    if(build_delete_msg_bin(view_id, deleted_id, db_name, db_name_size,
			    table_name, table_name_size, data_out, &out_size, data_size) < 0) {
      cout << "build failed" << endl;
      success = false;
      break;
    }
    if(get_message_type(data_out) != DELETE_BIN) {
      cout << "fail: msg type is: " << get_message_type(data_out) << endl;
      success = false;
      break;
    }
    
    string obtained_db_name;
    string obtained_table_name;
    int obtained_view_id = 0;
    int obtained_deleted_id = 0;

    parse_delete_msg_bin(data_out, obtained_view_id, obtained_deleted_id, obtained_db_name, obtained_table_name);

    if(view_id != obtained_view_id || deleted_id != obtained_deleted_id || 
       strcmp(db_name, obtained_db_name.c_str()) != 0 || 
       strcmp(table_name, obtained_table_name.c_str()) != 0) {
      cout << "[delete_bin] fail in view_id, db_name or table_name: " << endl;
      cout << "[delete_bin] view_id: " << view_id << ", obtained_view_id: " << obtained_view_id << endl;
      cout << "[delete_bin} deleted_id: " << deleted_id << ", obtained_deleted_id: " << obtained_deleted_id << endl;
      cout << "[delete_bin] db_name: " << db_name << ", obtained_db_name: " << obtained_db_name << endl;
      cout << "[delete_bin] table_name: " << table_name << ", obtained_table_name: " << obtained_table_name << endl;
      success = false;
      break;
    }

    if(build_delete_answer(answer, nSize) < 0) {
      success = false;
      cout << "Building update answer message failed" << endl;
      return;
    }
    
    string ret_msg = "done|";
    if(ret_msg.compare((char *) answer) != 0) {
      cout << "fail in prepared answer comparison" << endl;
      cout << ret_msg << ", " << answer << endl;
      success = false;
      return;
    }
    
    free(answer);
    free(data_out);
    free(db_name);
    free(table_name);
  }
}

/**
 * Tests that single entity read eager is working correctly
 * \param[out] success
 */
void test_se_eager(bool &success) {
  success = true;
  int nSize        = 10000000;
  message_data msg = (char*) malloc(nSize * sizeof(char));
  
  for(int i = 0; i < test_num; i ++) {
    int view_id = (int) (drand48() * range);
    int entity_id = (int) (drand48() * range);
    
    if(build_se_eager_msg(view_id, entity_id, msg, nSize) < 0) {
      cout << "build failed" << endl;
      success = false;
      break;
    }
    if(get_message_type(msg) != READ_SE_EAGER) {
      cout << "fail: msg type is: " << get_message_type(msg) << endl;
      cout << "msg: " << msg << endl;
      success = false;
      break;
    }
    
    int obtained_view_id = 0;
    int obtained_entity_id = 0;
    parse_se_eager_msg(msg, obtained_view_id, obtained_entity_id);
    if(obtained_view_id != view_id || obtained_entity_id != entity_id) {
      cout << "fail in view id or entity id: " << endl;
      cout << "view id: " << view_id << ", obtained view id: " << obtained_view_id << endl;
      cout << "entity id: " << entity_id << ", obtained entity id: " << obtained_entity_id << endl;
      success = false;
      break;
    }
    
    int entity_label = (int) (drand48() * 2);
    ostringstream test_label;
    test_label << entity_label << "|";
    
    if(build_read_se_answer(entity_label, msg, nSize) < 0) {
      cout << "build failed" << endl;
      success = false;
      break;
    }
    
    if(!test_label.str().compare((char *) msg) == 0) {
      cout << "fail in prepared answer comparison" << endl;
      cout << "msg: -" << test_label.str() << "-, obtained msg: -" << msg << "-" << endl;
      success = false;
      break;
    }
    
  }
  
  free(msg);
}

/**
 * Tests that single entity read lazy naive that sends binary data is working correctly
 * \param[in] vec feature vector of the entity
 * \param[out] success
 */
template <class T>
void single_test_se_lazy_naive_bin(T vec, bool &success) {
  success = true;
  fillVectors();
  size_t data_out_size = sizeof(int) * 10000000;
  void *data_out = malloc(data_out_size);
  
  size_t data_size, out_size;
  int view_id_in = 8, view_id_out = 0;
  
  void *vec_data = vec.write_bin(data_size);
  build_se_lazy_naive_msg_bin(view_id_in, vec_data, data_size, data_out, &out_size, data_out_size);
  
  
  T new_vec;
  parse_se_lazy_naive_msg_bin<T>(data_out, view_id_out, new_vec);
  if(view_id_in != view_id_out) { 
    cout << "View id differs " << view_id_in << ", " << view_id_out << endl; 
    success = false; 
    return;   
  }
  free(vec_data);
  
  // now compare the vectors.
  if(vec.size() != new_vec.size()) {
    cout << "Vector sizes differ " << vec.size() << ", " << new_vec.size() << endl; 
    success = false;     
    return; 
  }
  
  for(int j = 0; j < vec.size(); j ++) {
    if(new_vec.get(j) != vec.get(j)) {
      cout << "vector content differ at j=" << j << " -- " << new_vec.get(j) << ", " << vec.get(j) << endl;
      success = false;
      return;
    }
  }
}


/**
 * Tests that single entity read lazy naive that sends binary data is working correctly
 * \param[out] success
 */
void test_se_lazy_naive_bin(bool &success) {
  success = true;
  fillVectors();
  
  for(int i = 0; i < test_num; i ++) {
    single_test_se_lazy_naive_bin(fVectorList[i], success);
    if(!success)
      break;
    single_test_se_lazy_naive_bin(sVectorList[i], success);
    if(!success)
      break;
  }
} 

/**
 * Tests that single entity read lazy hazy that sends binary data is working correctly
 * \param[in] vec feature vector of the entity
 * \param[out] success
 */
template <class T>
void single_test_se_lazy_hazy_bin(T vec, bool &success) {
  success = true;
  fillVectors();
  size_t data_out_size = sizeof(int) * 10000000;
  void *data_out = malloc(data_out_size);
  
  size_t data_size, out_size;
  int view_id_in = 8, view_id_out = 0;
  double eps_in = 1.2913, eps_out = 0;
  
  void *vec_data = vec.write_bin(data_size);
  build_se_lazy_hazy_msg_bin(view_id_in, eps_in, vec_data, data_size, data_out, &out_size, data_out_size);
  
  
  T new_vec;
  parse_se_lazy_hazy_msg_bin(data_out, view_id_out, new_vec, eps_out);
  if(view_id_in != view_id_out) {
    cout << "View id differs " << view_id_in << ", " << view_id_out << endl;
    success = false;
    return;  
  }
  free(vec_data);
  
  if(eps_in != eps_out) {
    cout << "eps differs " << eps_in << ", " << eps_out << endl;
    success = false;
    return;
  }
  
  // now compare the vectors.
  if(vec.size() != new_vec.size()) {
    cout << "Vector sizes differ " << vec.size() << ", " << new_vec.size() << endl;
    success = false;
    return;
  }
  
  for(int j = 0; j < vec.size(); j ++) {
    if(new_vec.get(j) != vec.get(j)) {
      cout << "vector content differ at j=" << j << " -- " << new_vec.get(j) << ", " << vec.get(j) << endl;
      success = false;
      return;
    }
  }
}


/**
 * Tests that single entity read lazy hazy that sends binary data is working correctly
 * \param[out] success
 */
void test_se_lazy_hazy_bin(bool &success) {
  success = true;
  fillVectors();
  
  for(int i = 0; i < test_num; i ++) {
    single_test_se_lazy_hazy_bin(fVectorList[i], success);
    if(!success)
      break;
    single_test_se_lazy_hazy_bin(sVectorList[i], success);
    if(!success)
      break;
  }
} 


int main () {
  bool success = false;
  
  test_se_eager(success);
  cout << (success ? "successfull" : "fail") << endl;
  test_count(success);
  cout << (success ? "successfull" : "fail") << endl;
  test_se_lazy_hazy(success);
  cout << (success ? "successfull" : "fail") << endl;
  test_se_lazy_naive(success);
  cout << (success ? "successfull" : "fail") << endl;
  test_update(success);
  cout << (success ? "successfull" : "fail") << endl;
  test_query_id(success);
  cout << (success ? "successfull" : "fail") << endl;
  test_close_connection(success);
  cout << (success ? "successfull" : "fail") << endl;
  test_se_lazy_naive_bin(success);
  cout << (success ? "successfull" : "fail") << endl;
  test_se_lazy_hazy_bin(success);
  cout << (success ? "successfull" : "fail") << endl;
  test_delete_bin(success);
  cout << (success ? "successfull" : "fail") << endl;
  test_create_view(success);
  cout << (success ? "successfull" : "fail") << endl;

  /*  message_data msg = build_read_se_answer(10);
      cout << msg << endl;   
      free_message_data(msg);
      
      msg = build_se_lazy_naive_msg(10, "{1,2,3,4}");
      cout << msg << endl;
      int viewid; FVector f;
      parse_se_lazy_naive_msg<FVector>(msg, viewid, f);
      cout << "id=" << viewid << " fvector=" << f << endl;
      free_message_data(msg);
      
      msg = build_update_answer();
      cout << msg << endl;  
      free_message_data(msg);*/
  return 0;
}
