#include "hazy_message.h"
#include "vectors.h"
#include "low_level_net.h"
#include "postgres_util.h"
#include "frontend_util.h"
#include "server_conn.h"
#include <iostream>
#include <string>
#include <math.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>

using namespace std;
	
/**
 * This class implements regression tests for hazy message class with triggers
 * For each query in hazy message, a test first fills both SVector and FVector lists to be used
 * Then, it creates a msg for that query and sends it to the thread that plays server role.
 * Then, server parses trigger msg and sends the result back to the trigger.
 * Then, the msg sent by the server is retrieved and compared with the one that is expected to see it is correct.
 * Also, we test whether the server parsed the msg correctly or not.
 */

int bDense = 0;

const int numSVectors = 50;
int dim = 100;
int max_svector_dim = 100;

const int test_num = numSVectors;
const int range = 100;

FVector fVectorList[numSVectors];
SVector sVectorList[numSVectors];

int parsed_view_id = -1;
double parsed_eps = -1;
int parsed_entity_id = -1;
int parsed_label = -1;
int entity_label_answer = -1;
int parsed_server_id=0;
int parsed_close_conn = 0;
int count_answer = -1;
Hazy_Database *db_conn;
string db_name = "";

int parsed_deleted_id = -1;
string parsed_db_name = "";
string parsed_rel_name = "";

struct create_view_bin_str *parsed_cv_struct;


template<class T>
struct _id_label_vector_entry {
  int id;
  T x;
  int label;
  static int parse_tuple(PGresult *res, int index, _id_label_vector_entry &c) {
    c.id = atoi(PQgetvalue(res, index, 0));
    c.label = atoi(PQgetvalue(res, index, 1));
    std::string f_vec = PQgetvalue(res, index, 2);
    int r = c.x.fromPSQL(f_vec);
    return r;
  }
  static bool _entity_compare(const _id_label_vector_entry &x, const _id_label_vector_entry &y) {
    return x.eps < y.eps;
  }
};

/**
 * Writes the given msg to the server and retrieves integer from the received msg
 * \param[in] msg msg to be written to the server
 * \param[in] length size of the msg
 */
int write_func_create_view(message_data msg, size_t length) {
  struct addrinfo *res = NULL;
  int sockfd = -1;
  pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_lock( &mutex1 );
  
  int buffer_initialized = 0;
  struct buffer_handle update_trig_buf;
  if (buffer_initialized == 0) { update_trig_buf = initialize_buffer(); buffer_initialized = 1; } 
  
  struct addrinfo hints;
  
  if(sockfd < 0) {
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags  = AI_NUMERICSERV;
    
    if( getaddrinfo("localhost", "10003", &hints, &res) != 0 ) {
      cerr << "[hazy_msg_test_with_triggers.cxx] Getaddrinfo fails\n" << endl;
      res = NULL;
      destroy_handle(update_trig_buf);
      pthread_mutex_unlock( &mutex1 );
      return -1;
    }
    
    sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    
    if(sockfd < 0) {
      cerr << "[hazy_msg_test_with_triggers.cxx] msg error: %s \n" << endl;
      sockfd = -1;
      destroy_handle(update_trig_buf);
      pthread_mutex_unlock( &mutex1 );
      
      cerr << "[hazy_msg_test_with_triggers.cxx] Connect failed!" << endl;
      return -1;
    }
    
    int flag = 1;
    if(setsockopt( sockfd, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag) ) == -1) {
      cerr << "Couldn't setsockopt(TCP_NODELAY)\n" << endl;
      destroy_handle(update_trig_buf);
      pthread_mutex_unlock( &mutex1 );
      cerr << "EXIT_FAILURE" << endl;
    }
    
    if(connect(sockfd, res->ai_addr, res->ai_addrlen) < 0) {
      cerr << "[hazy_msg_test_with_triggers.cxx] msg error: %s \n" << endl;
      sockfd = -1;
      pthread_mutex_unlock( &mutex1 );
      
      cerr << "[hazy_msg_test_with_triggers.cxx] Connect failed\n" << endl; 
      return -1;
    }
  }

  if(write_wire_message(sockfd, msg, length) != (int) length) {
    cerr << "msg length: " << length << endl;
    cerr << "msg error: " << strerror(errno) << endl;
    sockfd = -1;
    destroy_handle(update_trig_buf);
    pthread_mutex_unlock( &mutex1 );
    
    cerr << "[hazy_msg_test_with_triggers.cxx] Write failed!" << endl;
    return -1;
  }	
  
  char *pData = (char *) read_wire_message(&update_trig_buf, sockfd);
  
  if(pData == NULL) {
    cerr << "msg error: " << strerror(errno) << endl;
    sockfd = -1;
    destroy_handle(update_trig_buf);
    pthread_mutex_unlock( &mutex1 );
    
    cerr << "[hazy_msg_test_with_triggers.cxx] Read Failed!" << endl;
    return -1;
  }
  
  pthread_mutex_unlock( &mutex1 );
  
  return 0;
}

/**
 * Converts FVector to dbstring without using cast as float8[]
 * \param[in] v FVector
 * \param[out] feature_values dbstring which is created
 */
void
fvector_to_dbstring_without_cast(const FVector &v, std::string &feature_values) {
  std::ostringstream values;
  
  std::ios::fmtflags oldflags = values.flags();
  std::streamsize oldprec = values.precision();
  values << std::scientific << std::setprecision(sizeof(VFloat)==4 ? 7 : 16);
  
  values << "{";
  for(int i = 0; i < v.size(); i ++)
    values << v.get(i) << ((i < v.size() - 1) ? "," : "");
  
  values << "}";
  
  feature_values = std::string(values.str());
  
  values.precision(oldprec);
  values.flags(oldflags);  
}


/**
 * Converts SVector to dbstring without using cast as pair[]
 * \param[in] s SVector
 * \param[out] feature_vector_str dbstring which is created
 */
void
svector_to_dbstring_without_cast(const SVector &s, std::string& feature_vector_str) {
  std::ostringstream feature_vector;
  
  std::ios::fmtflags oldflags = feature_vector.flags();
  std::streamsize oldprec = feature_vector.precision();
  feature_vector << std::scientific << std::setprecision(sizeof(VFloat)==4 ? 7 : 16);
  
  int n = s.npairs();
  feature_vector << "{";
  for(const SVector::Pair *p = s; p->i >= 0; p ++) {
    feature_vector << "\"(" << p->i << "," << p->v << ")\"" << ((n > 1) ? "," : "");
    n--;
  }
  feature_vector << "}";
  
  feature_vector_str = std::string(feature_vector.str());
  
  feature_vector.precision(oldprec);
  feature_vector.flags(oldflags);
}

/**
 * Randomly fill fvector and svector lists
 */
void fillVectors() {
  for(int i = 0; i < numSVectors; i ++) {
    int svector_dim = int(drand48() * max_svector_dim);
    while(svector_dim == 0)
      svector_dim = int(drand48() * max_svector_dim);
    
    for(int j = 0; j < svector_dim; j ++) {
      sVectorList[i].set(int(drand48() * dim), ((VFloat) drand48() * 2 -1));
    }
    for(int j = 0; j < dim; j ++)
      fVectorList[i].set(j, ((VFloat) drand48() * 2 - 1));
  }
}

/**
 * Tests that query server id is working correctly
 * \param[out] success
 */
void test_query_server_id(bool &success) {
  success = true;
  
  for(int i = 0; i < test_num; i ++) {
    ostringstream query;
    query << "SELECT getServerId();";
    int server_id_from_query = -1;
    db_conn->execute_query_msg_int(query.str().c_str(), "error in query server id", server_id_from_query);
    
    //LOGGING_ONLY(cout << "server id come: " << server_id_from_query);
    if(parsed_server_id != server_id_from_query) {
      cout << "fail in server id" << endl;
      cout << server_id_from_query << ", " << parsed_server_id << endl;
      success = false;
      break;
    }
  }
}

/**
 * Tests that close connection query is working correctly
 * \param[out] success
 */
void test_close_server_conn(bool &success) {
  success = true;
  
  for(int i = 0; i < test_num; i ++) {
    ostringstream query;
    query << "SELECT closeServerConn();";
    int close_conn_from_query = -1;
    db_conn->execute_query_msg_int(query.str().c_str(), "error in query close connection", close_conn_from_query);
    
    if(close_conn_from_query != parsed_close_conn) {
      cout << "fail in close conn" << endl;
      cout << close_conn_from_query << ", " << parsed_close_conn << endl;
      success = false;
      break;
    }
  }
}

/**
 * Tests count hazy message method
 * \param[out] success
 */
void test_count(bool &success) {
  success = true;
  
  for(int i = 0; i < test_num; i ++) {
    int view_id = 0;
    
    ostringstream query;
    query << "SELECT getView();";
    int entity_count_from_query = -1;
    db_conn->execute_query_msg_int(query.str().c_str(), "error in count query", entity_count_from_query);
    
    if(parsed_view_id != view_id || entity_count_from_query != count_answer) {
      cout << "fail in view id or count answer" << endl;
      cout << view_id << ", " << parsed_view_id << endl;
      cout << entity_count_from_query << ", " << count_answer << endl;
      success = false;
      break;
    }
  }
}

/**
 * Tests that update message is working correctly
 * \param[in] vec feature vector of the entity
 * \param[out] success
 * \param[in] parsed_vec parsed vector by the server
 */
template <class T>
void single_test_update(T vec, bool &success, T *parsed_vec) {
  int view_id = 0;
  int label = (int) (drand48() * 2);
  ostringstream ostr;
  ostr << label << " " << vec;
  
  string vec_str = ostr.str().substr(0, ostr.str().size() - 1);	//remove new line
  
  ostringstream query;
  query << "INSERT INTO large_update_pos VALUES(0, '" << vec_str << "');";
  db_conn->execute_statement_msg(query.str().c_str(), "single test update");
  
  if(view_id != parsed_view_id || parsed_label != label || vec.size() != parsed_vec->size()) {
    cout << "problem in view id, label or vec size" << endl;
    cout << "view_id=" << view_id << ", " << parsed_view_id << endl;
    cout << "label=" << label << ", " << parsed_label << endl;
    cout << "vecsize=" << vec.size() << ", " << parsed_vec->size() << endl;
    cout << "vec        : " << vec << endl;
    cout << "parsed_vec : " << *(parsed_vec) << endl;
    success = false;
    return;
  }
  
  
  for(int i = 0; i < vec.size(); i ++) {
    if(vec.get(i) != parsed_vec->get(i)) {
      cout << "vector content differs: " << vec.get(i) << ", " << parsed_vec->get(i) << endl;
      success = false;
      return;
    }
  }
}

/**
 * Tests that update message that sends binary data is working correctly
 * \param[in] vec feature vector of the entity
 * \param[in] dense whether the feature vector is dense or sparse
 * \param[out] success
 * \param[in] parsed_vec the vector which is parsed by the server
 */
template <class T>
void single_test_update_bin(T vec, bool dense, bool &success, T *parsed_vec) {
  int view_id = 0;
  int label = (int) (drand48() * 2);
  string vec_str = ""; 
  
  ostringstream query;
  if(dense) {
    fvector_to_dbstring(vec, vec_str);
    query << "INSERT INTO large_update_type_test VALUES(0, " << label << ", " << vec_str << ");";
  }
  else {
    svector_to_dbstring(vec, vec_str);
    query << "INSERT INTO large_update_type_test3 VALUES(0, " << label << ", " << vec_str << ");";
  }
  
  db_conn->execute_statement_msg(query.str().c_str(), "single test update");
  
  if(view_id != parsed_view_id || parsed_label != label || vec.size() != parsed_vec->size()) {
    cout << "problem in view id, label or vec size" << endl;
    cout << "view_id=" << view_id << ", " << parsed_view_id << endl;
    cout << "label=" << label << ", " << parsed_label << endl;
    cout << "vecsize=" << vec.size() << ", " << parsed_vec->size() << endl;
    cout << "vec        : " << vec << endl;
    cout << "parsed_vec : " << *(parsed_vec) << endl;
    success = false;
    return;
  }
  
  
  for(int i = 0; i < vec.size(); i ++) {
    if(vec.get(i) != parsed_vec->get(i)) {
      cout << "vector content differs: " << vec.get(i) << ", " << parsed_vec->get(i) << endl;
      success = false;
      return;
    }
  }
}



/**
 * Tests that single entity read lazy naive message is working correctly
 * \param[in] vec feature vector of the entity
 * \param[out] success
 * \param[in] dense whether vector is sparse or dense
 * \param[in] parsed_vec the vector that is parsed by the server
 */
template <class T>
void single_test_se_lazy_naive(T vec, bool &success, bool dense, T *parsed_vec) {
  int view_id = 0;
  
  string vec_str = "";
  if(dense)
    fvector_to_dbstring_without_cast(vec, vec_str);
  else
    svector_to_dbstring_without_cast(vec, vec_str);
  
  ostringstream query;
  query.precision(16);
  query << "SELECT getEntityLabelWithSQLNoEps('" << vec_str << "'::text);";
  int entity_label_from_query = 0;
  db_conn->execute_query_msg_int(query.str().c_str(), "error in se lazy naive", entity_label_from_query);
  
  if(view_id != parsed_view_id || vec.size() != parsed_vec->size()) {
    cout.precision(16);
    cout << "fail in view id or eps: " << endl;
    cout << "view id: " << view_id << ", obtained view id: " << parsed_view_id << endl;
    cout << "vec size: " << vec.size() << ", obtained vec size: " << parsed_vec->size() << endl;
    success = false;
    return;
  }
  
  for(int i = 0; i < vec.size(); i ++) {
    if(vec.get(i) != parsed_vec->get(i)) {
      cout << "vector content differ: " << vec.get(i) << ", " << parsed_vec->get(i) << endl;
      success = false;
      return;
    }
  }
  
  if(entity_label_from_query != entity_label_answer) {
    cout << "entity labels did not match" << endl;
    cout << entity_label_from_query << ", " << entity_label_answer << endl;
    success = false;
    return;
  }
}

/**
 * Tests that single entity read lazy hazy is working correctly
 * \param[in] vec feature vector of the entity
 * \param[out] success
 * \param[in] dense whether feature vector is dense or sparse
 * \param[in] parsed_vec the vector that is parsed by the server
 */
template <class T>
void single_test_se_lazy_hazy(T vec, bool &success, bool dense, T *parsed_vec) {
  int view_id = 0;
  double eps = drand48() * 2 - 1;
  
  string vec_str = "";
  if(dense)
    fvector_to_dbstring_without_cast(vec, vec_str);
  else
    svector_to_dbstring_without_cast(vec, vec_str);
  
  ostringstream query;
  query.precision(16);
  query << "SELECT getEntityLabelWithSQL(" << eps << ", '" << vec_str << "'::text);";
  int entity_label_from_query = 0;
  db_conn->execute_query_msg_int(query.str().c_str(), "error in se lazy hazy", entity_label_from_query);
  
  if(view_id != parsed_view_id || fabs(eps - parsed_eps) > 1.0e-16 || vec.size() != parsed_vec->size()) {
    cout.precision(16);
    cout << "fail in view id or eps: " << endl;
    cout << "view id: " << view_id << ", obtained view id: " << parsed_view_id << endl;
    cout << "eps: " << eps << ", obtained eps: " << parsed_eps << endl;
    cout << "vec size: " << vec.size() << ", obtained vec size: " << parsed_vec->size() << endl;
    success = false;
    return;
  }
  
  for(int i = 0; i < vec.size(); i ++) {
    if(vec.get(i) != parsed_vec->get(i)) {
      cout << "vector content differ: " << vec.get(i) << ", " << parsed_vec->get(i) << endl;
      success = false;
      return;
    }
  }
  
  if(entity_label_from_query != entity_label_answer) {
    cout << "entity labels did not match" << endl;
    cout << entity_label_from_query << ", " << entity_label_answer << endl;
    success = false;
    return;
  }
}

/**
 * Tests that single entity read lazy hazy that sends binary data is working correctly
 * \param[in] vec feature vector of the entity
 * \param[out] success
 * \param[in] dense whether the feature vector is dense or not
 * \param[in] parsed_vec the vector that is parsed by the server
 */
template <class T>
void single_test_se_lazy_hazy_bin(T vec, bool &success, bool dense, T *parsed_vec) {
  int view_id = 0;
  double eps = drand48() * 2 - 1;
  
  string vec_str = "";
  if(dense)
    fvector_to_dbstring(vec, vec_str);
  else
    svector_to_dbstring(vec, vec_str);
  
  ostringstream query;
  query.precision(16);
  query << "SELECT getEntityLabelWithSQL(" << eps << ", " << vec_str << ");";
  int entity_label_from_query = 0;
  db_conn->execute_query_msg_int(query.str().c_str(), "error in se lazy hazy bin", entity_label_from_query);
  
  if(view_id != parsed_view_id || fabs(eps - parsed_eps) > 1.0e-16 || vec.size() != parsed_vec->size()) {
    cout.precision(16);
    cout << "fail in view id or eps: " << endl;
    cout << "view id: " << view_id << ", obtained view id: " << parsed_view_id << endl;
    cout << "eps: " << eps << ", obtained eps: " << parsed_eps << endl;
    cout << "vec size: " << vec.size() << ", obtained vec size: " << parsed_vec->size() << endl;
    success = false;
    return;
  }
  
  for(int i = 0; i < vec.size(); i ++) {
    if(vec.get(i) != parsed_vec->get(i)) {
      cout << "vector content differ: " << vec.get(i) << ", " << parsed_vec->get(i) << endl;
      success = false;
      return;
    }
  }
  
  if(entity_label_from_query != entity_label_answer) {
    cout << "entity labels did not match" << endl;
    cout << entity_label_from_query << ", " << entity_label_answer << endl;
    success = false;
    return;
  }
}

/**
 * Tests that single entity read lazy naive that sends binary data is working correctly
 * \param[in] vec feature vector of the entity
 * \param[out] success
 * \param[in] dense whether the feature vector is dense or not
 * \param[in] parsed_vec the vector that is parsed by the server
 */
template <class T>
void single_test_se_lazy_naive_bin(T vec, bool &success, bool dense, T *parsed_vec) {
  int view_id = 0;
  
  string vec_str = "";
  if(dense)
    fvector_to_dbstring(vec, vec_str);
  else
    svector_to_dbstring(vec, vec_str);
  
  ostringstream query;
  query.precision(16);
  query << "SELECT getEntityLabelWithSQLNoEps(" << vec_str << ");";
  int entity_label_from_query = 0;
  db_conn->execute_query_msg_int(query.str().c_str(), "error in se lazy naive bin", entity_label_from_query);
  
  if(view_id != parsed_view_id || vec.size() != parsed_vec->size()) {
    cout.precision(16);
    cout << "fail in view id or eps: " << endl;
    cout << "view id: " << view_id << ", obtained view id: " << parsed_view_id << endl;
    cout << "vec size: " << vec.size() << ", obtained vec size: " << parsed_vec->size() << endl;
    success = false;
    return;
  }
  
  for(int i = 0; i < vec.size(); i ++) {
    if(vec.get(i) != parsed_vec->get(i)) {
      cout << "vector content differ: " << vec.get(i) << ", " << parsed_vec->get(i) << endl;
      success = false;
      return;
    }
  }
  
  if(entity_label_from_query != entity_label_answer) {
    cout << "entity labels did not match" << endl;
    cout << entity_label_from_query << ", " << entity_label_answer << endl;
    success = false;
    return;
  }
}

/**
 * Tests that update message is working correctly
 * \param[out] success
 * \param[in] parsed_vec the vector parsed by the server
 */
template<class T>
void test_update(bool &success, T *parsed_vec) {
  success = true;
  
  for(int i = 0; i < test_num; i ++) {
    if(bDense) {
      single_test_update(fVectorList[i], success, (FVector *) parsed_vec);
      if(!success)
	break;
    }
    else {
      single_test_update(sVectorList[i], success, (SVector *) parsed_vec);
      if(!success)
	break;
    }
  }
}

/**
 * Tests that update message that sends binary data is working correctly
 * \param[out] success
 * \param[in] parsed_vec the vector parsed by the server
 */
template<class T>
void test_update_bin(bool &success, T *parsed_vec) {
  success = true;
  
  for(int i = 0; i < test_num; i ++) {
    if(bDense) {
      single_test_update_bin(fVectorList[i], bDense, success, (FVector *) parsed_vec);
      if(!success)
        break;
    }
    else {
      single_test_update_bin(sVectorList[i], bDense, success, (SVector *) parsed_vec);
      if(!success)
        break;
    }
  }
}

void create_random_string(char *s, size_t len) {
  static const char alphanum[] =
    "0123456789"
    "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    "abcdefghijklmnopqrstuvwxyz";
  
  for(unsigned int i = 0; i < len - 1; i ++)
    s[i] = alphanum[rand() % (sizeof(alphanum) - 1)];
  
  s[len - 1] = 0;
}

void single_test_create_view(bool &success) {
  buffer_handle bh;
  bh = initialize_buffer();
  struct create_view_bin_str *cv_struct = new struct create_view_bin_str();
  //set view id
  cv_struct->view_id = rand() % range;
  //set storage manager
  int rand_num = rand() % 3;
  if(rand_num == 0)
    cv_struct->sm = MAIN_MEMORY_SM;
  else if(rand_num == 1)
    cv_struct->sm = ON_DISK_SM;
  else
    cv_struct->sm = HYBRID_SM;
  //set dense or not
  rand_num = rand() % 2;
  cv_struct->dense = rand_num;
  //set useepsmap
  if(cv_struct->sm == HYBRID_SM) {
    rand_num = rand() % 2;
    cv_struct->useepsmap = rand_num;
  }
  else
    cv_struct->useepsmap = 0;
  //set hazy_strategy
  rand_num = rand() % 4;
  switch(rand_num) {
  case 0: cv_struct->hazy_strategy = hazy_model::LAZY_NAIVE;
    break;
  case 1: cv_struct->hazy_strategy = hazy_model::LAZY_HAZY;
    break;
  case 2: cv_struct->hazy_strategy = hazy_model::EAGER_NAIVE;
    break;
  case 3: cv_struct->hazy_strategy = hazy_model::EAGER_HAZY;
    break;
  default: cv_struct->hazy_strategy = hazy_model::EAGER_HAZY;
  }
  //set reservoir_size
  cv_struct->reservoir_size = rand() % range;
  //set resort_time
  cv_struct->resort_time = drand48() * 10.0;
  //set tao_val
  cv_struct->tao_val = drand48() * 0.01;
  //set alpha_ratio
  cv_struct->alpha_ratio = drand48() * 1.0;
  //set buffer_size
  if(cv_struct->sm == HYBRID_SM) {
    cv_struct->buffer_size = rand() % range;
  }
  else
    cv_struct->buffer_size = 0;
  //set view_name
  size_t view_name_size = rand() % 1000;
  while(view_name_size < 2)
    view_name_size = rand() % 1000;
  char *view_name = (char *) malloc(sizeof(char) * view_name_size);
  create_random_string(view_name, view_name_size);
  cv_struct->view_name = view_name;
  //set view_key
  size_t view_key_size = rand() % 1000;
  while(view_key_size < 2)
    view_key_size = rand() % 1000;
  char *view_key = (char *) malloc(sizeof(char) * view_key_size);
  create_random_string(view_key, view_key_size);
  cv_struct->view_key = view_key;
  //set entity_table_name
  size_t entity_table_name_size = rand() % 1000;
  while(entity_table_name_size < 2)
    entity_table_name_size = rand() % 1000;
  char *entity_table_name = (char *) malloc(sizeof(char) * entity_table_name_size);
  create_random_string(entity_table_name, entity_table_name_size);
  cv_struct->entity_table_name = entity_table_name;
  //set entity_table_key
  size_t entity_table_key_size = rand() % 1000;
  while(entity_table_key_size < 2)
    entity_table_key_size = rand() % 1000;
  char *entity_table_key = (char *) malloc(sizeof(char) * entity_table_key_size);
  create_random_string(entity_table_key, entity_table_key_size);
  cv_struct->entity_table_key = entity_table_key;
  //set training_table_name
  size_t training_table_name_size = rand() % 1000;
  while(training_table_name_size < 2)
    training_table_name_size = rand() % 1000;
  char *training_table_name = (char *) malloc(sizeof(char) * training_table_name_size);
  create_random_string(training_table_name, training_table_name_size);
  cv_struct->training_table_name = training_table_name;
  //set dim
  cv_struct->dim = rand() % range;
  //set db_name
  size_t db_name_size = rand() % 1000;
  while(db_name_size < 2)
    db_name_size = rand() % 1000;
  char *db_name = (char *) malloc(sizeof(char) * db_name_size);
  create_random_string(db_name, db_name_size);
  cv_struct->db_name = db_name;
  //set m_factor
  double m_factor = 1.0;
  cv_struct->m_factor = m_factor;

  int msgSize        = sizeof(struct create_view_bin_str) + cv_struct->view_name.size() + cv_struct->view_key.size() + 
    cv_struct->entity_table_name.size() + cv_struct->entity_table_key.size() + cv_struct->training_table_name.size() +
    cv_struct->db_name.size() + 1000;

  update_buffer_size(&bh, msgSize);
  
  if(build_create_view_msg(cv_struct, bh.pData, msgSize) < 0) {
    destroy_handle(bh);
    cout << "Building create view message failed" << endl;
    success = false;
    return;
  }

  //send this message to the server
  int retVal = write_func_create_view(bh.pData, strlen((char *) bh.pData));
  if(retVal < 0) {
    success = false;
    cerr << "There is a problem with sending the data" << endl;
    return;
  }
  
  destroy_handle(bh);
  //and then check
  if(!(cv_struct->view_id == parsed_cv_struct->view_id && cv_struct->sm == parsed_cv_struct->sm &&
      cv_struct->dense == parsed_cv_struct->dense && cv_struct->useepsmap == parsed_cv_struct->useepsmap &&
      cv_struct->hazy_strategy == parsed_cv_struct->hazy_strategy && cv_struct->reservoir_size == parsed_cv_struct->reservoir_size
      && cv_struct->resort_time == parsed_cv_struct->resort_time && cv_struct->tao_val == parsed_cv_struct->tao_val &&
      cv_struct->alpha_ratio == parsed_cv_struct->alpha_ratio && cv_struct->buffer_size == parsed_cv_struct->buffer_size &&
      cv_struct->view_name.compare(parsed_cv_struct->view_name) == 0 && 
      cv_struct->view_key.compare(parsed_cv_struct->view_key) == 0 &&
      cv_struct->entity_table_name.compare(parsed_cv_struct->entity_table_name) == 0 &&
      cv_struct->entity_table_key.compare(parsed_cv_struct->entity_table_key) == 0 &&
      cv_struct->training_table_name.compare(parsed_cv_struct->training_table_name) == 0 &&
      cv_struct->dim == parsed_cv_struct->dim && cv_struct->db_name.compare(parsed_cv_struct->db_name) == 0 &&
      cv_struct->m_factor == parsed_cv_struct->m_factor)) {
    cout << "[test_create_view] fail in create_view_bin struct comparison" << endl;
    cout << "[test_create_view] view_id: " << cv_struct->view_id << ", parsed view_id: " << parsed_cv_struct->view_id << endl;
    cout << "[test_create_view] sm: " << cv_struct->sm << ", parsed sm: " << parsed_cv_struct->sm << endl;
    cout << "[test_create_view] dense: " << cv_struct->dense << ", parsed dense: " << parsed_cv_struct->dense << endl;
    cout << "[test_create_view] useepsmap: " << cv_struct->useepsmap << ", parsed useepsmap: " << parsed_cv_struct->useepsmap << endl;
    cout << "[test_create_view] hazy_strategy: " << cv_struct->hazy_strategy << ", parsed hazy_strategy: " << parsed_cv_struct->hazy_strategy << endl;
    cout << "[test_create_view] reservoir_size: " << cv_struct->reservoir_size << ", parsed reservoir_size: " << parsed_cv_struct->reservoir_size << endl;
    cout << "[test_create_view] resort_time: " << cv_struct->resort_time << ", parsed resort_time: " << parsed_cv_struct->resort_time << endl;
    cout << "[test_create_view] tao_val: " << cv_struct->tao_val << ", parsed tao_val: " << parsed_cv_struct->tao_val << endl;
    cout << "[test_create_view] alpha_ratio: " << cv_struct->alpha_ratio << ", parsed alpha_ratio: " << parsed_cv_struct->alpha_ratio << endl;
    cout << "[test_create_view] buffer_size: " << cv_struct->buffer_size << ", parsed buffer_size: " << parsed_cv_struct->buffer_size << endl;
    cout << "[test_create_view] view_name: " << cv_struct->view_name << ", parsed view_name: " << parsed_cv_struct->view_name << endl;
    cout << "[test_create_view] view_key: " << cv_struct->view_key << ", parsed view_key: " << parsed_cv_struct->view_key << endl;
    cout << "[test_create_view] entity_table_name: " << cv_struct->entity_table_name << ", parsed entity_table_name: " << parsed_cv_struct->entity_table_name << endl;
    cout << "[test_create_view] entity_table_key: " << cv_struct->entity_table_key << ", parsed entity_table_key: " << parsed_cv_struct->entity_table_key << endl;
    cout << "[test_create_view] training_table_name: " << cv_struct->training_table_name << ", parsed training_table_name: " << parsed_cv_struct->training_table_name << endl;
    cout << "[test_create_view] dim: " << cv_struct->dim << ", parsed dim: " << parsed_cv_struct->dim << endl;
    cout << "[test_create_view] db_name: " << cv_struct->db_name << ", parsed db_name: " << parsed_cv_struct->db_name << endl;
    cout << "[test_create_view] m_factor: " << cv_struct->m_factor << ", parsed m_factor: " << parsed_cv_struct->m_factor << endl;
    success = false;
    return;
  }

  delete parsed_cv_struct;
}

void test_create_view(bool &success) {
  success = true;
  
  for(int i = 0; i < test_num; i ++)
    single_test_create_view(success);
}

template<class T>
void test_single_delete_bin(string table_name, int deleted_id, bool &success) {
  int view_id = 0;

  ostringstream delete_query;
  delete_query << "DELETE FROM " << table_name << " WHERE id = " << deleted_id << ";";
  
  db_conn->execute_statement_msg(delete_query.str().c_str(), "[test_delete_bin] create vector table query");

  if(view_id != parsed_view_id || db_name.compare(parsed_db_name) != 0 || parsed_rel_name.compare(table_name) != 0
     || deleted_id != parsed_deleted_id) {
    success = false;
    cout << "[test_single_delete_bin] fail in view_id, db_name or rel_name";
    cout << "[test_single_delete_bin] view_id: " << view_id << ", parsed_view_id: " << parsed_view_id << endl;
    cout << "[test_single_delete_bin] deleted_id: " << deleted_id << ", parsed_deleted_id: " << parsed_deleted_id << endl;
    cout << "[test_single_delete_bin] db_name: " << db_name << ", parsed_db_name: " << parsed_db_name << endl;
    cout << "[test_single_delete_bin] rel_name: " << table_name << ", parsed_rel_name: " << parsed_rel_name << endl;
    return;
  }

  //retrieve all entities and checked whether deleted one exists or not
  vector< _id_label_vector_entry<T> > _buffer;
  string retrieve_query = "SELECT id, label, feature_vector FROM " + table_name + ";";
  db_conn->postgresTupleParser<_id_label_vector_entry <T> >(retrieve_query.c_str(), _id_label_vector_entry<T>::parse_tuple, _buffer);

  for(unsigned int i = 0; i < _buffer.size(); i ++) {
    if(_buffer[i].id == deleted_id) {
      success = false;
      break;
    }
  }
}

void test_delete_bin(bool &success) {
  success = true;
  string table_name = "test_delete_fvector_bin";
  string trigger_name = "delete_ftrigger";
  string type_name = "float8[]";

  //test FVector
  string create_fvector_table_query = "DROP TABLE IF EXISTS " + table_name + "; CREATE TABLE " + table_name + "(id int, label int, feature_vector " + type_name + ");";
  string create_ftrigger_query = "DROP TRIGGER IF EXISTS " + trigger_name + " ON " + table_name + "; CREATE TRIGGER " + trigger_name + " AFTER DELETE ON " + table_name + " FOR EACH ROW EXECUTE PROCEDURE update_trig_bin();";
  db_conn->execute_statement_msg(create_fvector_table_query.c_str(), "[test_delete_bin] creaet fvector table query");
  db_conn->execute_statement_msg(create_ftrigger_query.c_str(), "[test_delete_bin] creaet ftrigger query");

  for(int i = 0; i < test_num; i ++) {
    ostringstream insert_query;
    string fvector;
    fvector_to_dbstring_without_cast(fVectorList[i], fvector);
    int label = 1;
    if(i % 3 == 0)
      label = -1;
    insert_query << "INSERT INTO test_delete_fvector_bin VALUES(" << i << ", " << label << ", '" << fvector << "');";
    db_conn->execute_statement_msg(insert_query.str().c_str(), "[test_delete_bin] insert fvector");
  }

  //didnt want to start deleting from the beginning or the end
  int start_id = 15;
  for(int i = 0; i < test_num; i ++) {
    int deleted_id = (start_id + i) % test_num;
    
    test_single_delete_bin<FVector>(table_name, deleted_id, success);
    if(!success)
      break;
  }
  
  if(!success)
    return;
  //test SVector
  table_name = "test_delete_svector_bin";
  trigger_name = "delete_strigger";
  type_name = "pair[]";

  //test FVector
  string create_svector_table_query = "DROP TABLE IF EXISTS " + table_name + "; CREATE TABLE " + table_name + "(id int, label int, feature_vector " + type_name + ");";
  string create_strigger_query = "DROP TRIGGER IF EXISTS " + trigger_name + " ON " + table_name + "; CREATE TRIGGER " + trigger_name + " AFTER DELETE ON " + table_name + " FOR EACH ROW EXECUTE PROCEDURE update_trig_bin();";
  db_conn->execute_statement_msg(create_svector_table_query.c_str(), "[test_delete_bin] creaet svector table query");
  db_conn->execute_statement_msg(create_strigger_query.c_str(), "[test_delete_bin] creaet strigger query");

  for(int i = 0; i < test_num; i ++) {
    ostringstream insert_query;
    string svector;
    svector_to_dbstring_without_cast(sVectorList[i], svector);
    int label = 1;
    if(i % 3 == 0)
      label = -1;
    insert_query << "INSERT INTO test_delete_svector_bin VALUES(" << i << ", " << label << ", '" << svector << "');";
    db_conn->execute_statement_msg(insert_query.str().c_str(), "[test_delete_bin] insert svector");
  }

  //didnt want to start deleting from the beginning or the end
  start_id = 15;
  for(int i = 0; i < test_num; i ++) {
    int deleted_id = (start_id + i) % test_num;
    
    test_single_delete_bin<SVector>(table_name, deleted_id, success);
    if(!success)
      break;
  }
  
  if(!success)
    return;
}

/**
 * Tests that single entity read lazy hazy message is working correctly
 * \param[out] success
 * \param[in] parsed_vec the vector parsed by the server
 */
template <class T>
void test_se_lazy_hazy(bool &success, T *parsed_vec) {
  success = true;
  
  for(int i = 0; i < test_num; i ++) {
    if(bDense) {
      single_test_se_lazy_hazy(fVectorList[i], success, true, (FVector *) parsed_vec);
      if(!success)
	break;
    }
    else {
      single_test_se_lazy_hazy(sVectorList[i], success, false, (SVector *) parsed_vec);
      if(!success)
	break;
    }
  }
}

/**
 * Tests that single entity read lazy hazy that sends binary message is working correctly
 * \param[out] success
 * \param[in] parsed_vec the vector parsed by the server
 */
template <class T>
void test_se_lazy_hazy_bin(bool &success, T *parsed_vec) {
  success = true;
  
  for(int i = 0; i < test_num; i ++) {
    if(bDense) {
      single_test_se_lazy_hazy_bin(fVectorList[i], success, true, (FVector *) parsed_vec);
      if(!success)
        break;
    }
    else {
      single_test_se_lazy_hazy_bin(sVectorList[i], success, false, (SVector *) parsed_vec);
      if(!success)
        break;
    }
  }
}


/**
 * Tests that single entity read lazy naive that sends binary message is working correctly
 * \param[out] success
 * \param[in] parsed_vec the vector parsed by the server
 */
template <class T>
void test_se_lazy_naive_bin(bool &success, T *parsed_vec) {
  success = true;
  
  for(int i = 0; i < test_num; i ++) {
    if(bDense) {
      single_test_se_lazy_naive_bin(fVectorList[i], success, true, (FVector *) parsed_vec);
      if(!success)
        break;
    }
    else {
      single_test_se_lazy_naive_bin(sVectorList[i], success, false, (SVector *) parsed_vec);
      if(!success)
        break;
    }
  }
}


/**
 * Tests that sends single entity read lazy naive message is working correctly
 * \param[out] success
 * \param[in] parsed_vec the vector parsed by the server
 */
template <class T>
void test_se_lazy_naive(bool &success, T *parsed_vec) {
  success = true;
  
  for(int i = 0; i < test_num; i ++) {
    if(bDense) {
      single_test_se_lazy_naive(fVectorList[i], success, true, (FVector *) parsed_vec);
      if(!success)             
      	break;    
    }
    else {             
      single_test_se_lazy_naive(sVectorList[i], success, false, (SVector *) parsed_vec);
      if(!success)             
      	break;                 
    }
  }   
}

/**
 * Tests that single entity read eager message is working correctly
 * \param[out] success
 */
void test_se_eager(bool &success) {
  success = true;
  
  for(int i = 0; i < test_num; i ++) {
    int view_id = 0;
    int entity_id = (int) (drand48() * range);
    
    ostringstream query;
    query << "SELECT getEntityLabel(" << entity_id << ");";
    int entity_label_from_query = 0;
    db_conn->execute_query_msg_int(query.str().c_str(), "error in se eager query", entity_label_from_query);
    
    if(parsed_view_id != view_id || parsed_entity_id != entity_id) {
      cout << "fail in view id or entity id: " << endl;
      cout << "view id: " << view_id << ", obtained view id: " << parsed_view_id << endl;
      cout << "entity id: " << entity_id << ", obtained entity id: " << parsed_entity_id << endl;
      success = false;
      break;
    }
    
    if(entity_label_from_query != entity_label_answer) {
      cout << "entity labels did not match" << endl;
      cout << entity_label_from_query << ", " << entity_label_answer << endl;
      success = false;
      break;
    }
    
  }
}


/**
 * structure that has Hazy_Worker and parsed_vec
 */
template<class T> 
struct worker_thread_info {
  Hazy_Worker *hw; // assume not NULL
  T& _parsed_vec;
  worker_thread_info(Hazy_Worker *_hw, T& pv) : _parsed_vec(pv) {
    hw = _hw;
  }
};


/**
 * The thread that simulates the server
 */
template<class T>
void* 
worker_thread(void *data) {
  cout << "[worker thread] STARTING" << endl;  
  worker_thread_info<T> *wit = (worker_thread_info<T> *) data;
  Hazy_Worker *hw = wit->hw;
  T& parsed_vec = wit->_parsed_vec;
  delete wit;
  
  
  int nSize        = 2048;
  message_data answer = (char*) malloc(nSize * sizeof(char));
  
  while (1) {
    void *data = hw->readData();
    
    if(data == NULL) {
      break;
    }
    switch(get_message_type(data)) {
    case READ_SE_LAZY_NAIVE:
      parse_se_lazy_naive_msg(data, parsed_view_id, parsed_vec);
      entity_label_answer = (int) (drand48() * 2);
      if(build_read_se_answer(entity_label_answer, answer, nSize) < 0) {
	cout << "[Worker] problem in build msg";
	return NULL;
      }
      break;
    case READ_SE_LAZY_HAZY:
      parse_se_lazy_hazy_msg(data, parsed_view_id, parsed_vec, parsed_eps);
      entity_label_answer = (int) (drand48() * 2);
      if(build_read_se_answer(entity_label_answer, answer, nSize) < 0) {
	cout << "[Worker] problem in build msg";
	return NULL;
      }
      break;
    case READ_SE_LAZY_HAZY_BIN:
      {
	parse_se_lazy_hazy_msg_bin(data, parsed_view_id, parsed_vec, parsed_eps);
	entity_label_answer = (int) (drand48() * 2);
	if(build_read_se_answer(entity_label_answer, answer, nSize) < 0) {
	  cout << "[Worker] problem in build msg";
	  return NULL;
	}
      }
      break;
    case READ_SE_LAZY_NAIVE_BIN:
      {
	parse_se_lazy_naive_msg_bin(data, parsed_view_id, parsed_vec);
	entity_label_answer = (int) (drand48() * 2);
	if(build_read_se_answer(entity_label_answer, answer, nSize) < 0) {
	  cout << "[Worker] problem in build msg";
	  return NULL;
	}
      }
      break;
    case UPDATE_BIN:
      {
	parse_update_msg_bin(data, parsed_view_id, parsed_vec, parsed_label);
	if(build_update_answer(answer, nSize) < 0) {
	  cout << "[Worker] problem in build msg";
	  return NULL;
	}
      }
      break;
    case DELETE_BIN:
      {
	parse_delete_msg_bin(data, parsed_view_id, parsed_deleted_id, parsed_db_name, parsed_rel_name);
	if(build_delete_answer(answer, nSize) < 0) {
	  cout << "[Worker] problem in build msg";
	  return NULL;
	}
      }
      break;
    case CREATE_VIEW:
      {
	parsed_cv_struct = new struct create_view_bin_str();
	parse_create_view_msg(data, parsed_cv_struct);
	if(build_create_view_answer(answer, nSize) < 0) {
	  cout << "[Worker] problem in build msg" << endl;
	  return NULL;
	}

	//this delete statement has to be here, but for testing purpose, we comment it
	//delete parsed_cv_struct;
      }
      break;
    case READ_SE_EAGER:
      parse_se_eager_msg(data, parsed_view_id, parsed_entity_id);
      entity_label_answer = (int) (drand48() * 2);
      if(build_read_se_answer(entity_label_answer, answer, nSize) < 0) {
	cout << "[Worker] problem in build msg";
	return NULL;
      }
      break;
    case READ_COUNT:
      parse_count_msg(data, parsed_view_id);
      count_answer = (int) (drand48() * range);
      if(build_read_count_answer(count_answer, answer, nSize) < 0) {
	cout << "[Worker] problem in build msg";
	return NULL;
      }
      break;
    case UPDATE:
      cout << "[update] incoming msg: " << data << endl;
      parse_update_msg(data, parsed_view_id, parsed_vec, parsed_label);
      if(build_update_answer(answer, nSize) < 0) {
	cout << "[Worker] problem in build msg";
	return NULL;
      }
      break;
    case QUERY_SERVER_ID:
      parsed_server_id = time ( NULL );
      if(build_query_server_answer(parsed_server_id, answer, nSize) < 0) {
	cout << "[Worker] problem in build msg" << endl;
	return NULL;
      }
      break;
    case CLOSE_CONNECTION:
      parsed_close_conn = 1;
      if(build_close_connection_answer(parsed_close_conn, answer, nSize) < 0) {
	cout << "[Worker] problem in build msg" << endl;;
	return NULL;
      }
      break;
    default: 
      cout << "[Worker] problem in msg type" << get_message_type(data) << endl;
      cout << "\t MSG: " << data << endl;
    }
    
    int success = hw->writeData(answer, strlen((char *) answer), false);
    if(success < 0)
      cout << "[Worker] error when writing to the client: " << strerror(errno);
  }
  hw->closeClientConnection();
  delete hw;
  return NULL;
}


template<class T>
struct server_thread_info {
  T& _parsed_vec;
  Hazy_Server *serv;
  vector<pthread_t> &workers;
  server_thread_info(Hazy_Server *srv, T& pv,   vector<pthread_t> &work) : _parsed_vec(pv), serv(srv), workers(work) {    
  }
};

template<class T>
void *server_thread ( void *ptr ) {
  server_thread_info<T> *sti = (server_thread_info<T>*) ptr;
  Hazy_Server *serv = sti->serv;
  vector<pthread_t> &workers = sti->workers;
  T& parsed_vec = sti->_parsed_vec;
  cout << "server thread started...." << endl;
  
  while(1) {
    cout << "Waiting to accept connection" << endl;
    Hazy_Worker *hw = serv->acceptConnection();
    if (hw == NULL) {
      printf("Error at accept() : %s\n", strerror(errno));
      continue;
    } 
    pthread_t t;
    cout << "Got a connection!" << endl;
    worker_thread_info<T> *wit = new worker_thread_info<T>(hw,parsed_vec);
    cout << "About to create thread..." << endl;
    int iret = pthread_create(&t, NULL, worker_thread<T>, (void *) wit);
    if(iret !=0 ) {
      cout << "pthread create failed " << iret << endl;
      break;
    }
    workers.push_back(t);
    cout << "Kicked off new worker thread: " << workers.size() << endl;   
  }
  
  // Server connection is closed
  // THIS WILL BE KILLED MOST LIKELY!!
  serv->closeConnection();
  /*for(vector<pthread_t>::iterator i = workers.begin(); 
    i != workers.end(); i++) {
    cout << "\tWaiting on a thread..." << endl;
    pthread_kill(*i, SIGALRM);
    }*/
  return NULL;
}

/**
 * Runs all regression tests
 */
template<class T>
void run_server() {
  T parsed_vec;
  
  Hazy_Server *serv;
  serv = new Hazy_Server();
  cout << "server started. Kicking off its thread..." << endl;
  
  pthread_t thread;
  vector<pthread_t> workers;
  server_thread_info<T> *sti = new server_thread_info<T>(serv, parsed_vec, workers);
  pthread_create(&thread, NULL, server_thread<T>, (void *) sti);  
  
  sleep(1);
  
  bool success_val = false;
  test_delete_bin(success_val);
  cout << "test_delete_bin test:" << ((success_val) ? "successful" : "fail") << endl;
  test_se_eager(success_val);
  cout << "test_se_eager test:" << ((success_val) ? "successful" : "fail") << endl;
  test_se_lazy_hazy(success_val, &parsed_vec);
  cout << "test_se_lazy_hazy:" << ((success_val) ? "successful" : "fail") << endl;
  test_se_lazy_hazy_bin(success_val, &parsed_vec);
  cout << "test_se_lazy_hazy_bin:" << ((success_val) ? "successful" : "fail") << endl;
  test_se_lazy_naive_bin(success_val, &parsed_vec);
  cout << "test_se_lazy_naive_bin:" << ((success_val) ? "successful" : "fail") << endl;
  test_update_bin(success_val, &parsed_vec);
  cout << "update_bin test:" << ((success_val) ? "successful" : "fail") << endl;
  test_se_lazy_naive(success_val, &parsed_vec);
  cout << "test_se_lazy_naive:" << ((success_val) ? "successful" : "fail") << endl;
  test_count(success_val);
  cout << "test_se_count:" << ((success_val) ? "successful" : "fail") << endl;
  test_query_server_id(success_val);
  cout << "query_server_id:" << ((success_val) ? "successful" : "fail") << endl;
  test_update(success_val, &parsed_vec);
  cout << "test_update:" << ((success_val) ? "successful" : "fail") << endl;  
  test_close_server_conn(success_val);
  cout << "close_server_conn:" << ((success_val) ? "successful" : "fail") << endl;
  test_create_view(success_val);
  cout << "test_create_view:" << ((success_val) ? "successful" : "fail") << endl;
  
  ///Database connection is closed
  db_conn->closeConnection();	
  cout << "Finished Killing threads" << endl;
  // KILL OFF THE OTHER THREADS!
  //pthread_kill(thread, SIGALRM);
  sleep(1);
  serv->closeConnection();
  /* for(vector<pthread_t>::iterator i = workers.begin(); 
     i != workers.end(); i++) {
     cout << "\tWaiting on a thread..." << endl;
     pthread_kill(*i, SIGALRM);
     } */
  cout << "Cleanup done" << endl;
}

int main (int argc, char **argv) {
  char *usage_str = "<dbname> [--dense]";
  
  if(argc > 3) {
    cout << argv[0] << " " << usage_str << endl;
    exit(1);
  }
  
  if(argc == 3) {
    db_name = argv[1];
    bDense = 1;
  }
  else if(argc == 2) {
    db_name = argv[1];
    bDense = 0;
  }
  else {
    cout << argv[0] << " " << usage_str << endl;
    exit(1);
  }
  
  /*	static struct extended_option long_options[] = {
	{"dense"    , no_argument, &bDense, 1, "if input contains dense vectors"},
	//{NULL,0,NULL,0,0} 
	};
	
	struct option* lo = convert_extended_options(long_options);
	int c = 0;
	while( (c = getopt_long(argc, argv, "r:d:m:o:l:", lo, &option_index)) != -1)
	{
	switch (c) {
	case 'r':
        rsize = atoi(optarg);
        break;
	case 'd':
        dim  = atoi(optarg);
        break;
	case 'l':
        lambda = atof(optarg);
        break;
	case 's':
        strategy = atoi(optarg);
        break;
	case 'o':
        db_name = optarg;
        break;
	case 'e':
        entity_table_name = optarg;
        break;
	case ':':
	case '?':
        print_usage(long_options, argv[0], usage_str);
        exit(-1);
	break;
	}
	}*/
  
  /*  if(optind == argc - 1) {
      db_name = argv[optind];
      } else {
      print_usage(long_options, argv[0], usage_str);
      exit(-1);
      }*/
  
  db_conn = new Hazy_Database(db_name);
  fillVectors();
  
  if(bDense)
    run_server<FVector>();
  else
    run_server<SVector>();
  
  return 0;
}
