#include "hazy_message.h"
#include "multi_class.h"
#include "postgres_util.h"
#include "server_conn.h"
#include "hashing.h"
#include "hazy_sgd.h"
#include "globals.h"
#include "frontend_util.h"
#include <errno.h>
#include <string>
#include <pthread.h>

#define HAZY_HYBRID "hazy_hybrid"
#define HAZY_ON_DISK "hazy_on_disk"
#define HAZY_MM "hazy_mm"

/**
 * This class implements multi-class classification that uses Main Memory architecture
 */

const string EXTERNAL_TABLE_BASE_NAME = "sm_test_external_table";
const string INTERNAL_TABLE_BASE_NAME = "internal_table";

const int r_size = 1000;

void usage() {
  cerr << "usage: " << endl;
  return;
}

using namespace std;

///Hazy Server
Hazy_Server *serv;

template<class T>
void startServer(int server_id, string dbname, int dim, int nLabels, const char *entity_file, const char *train_file_for_hot_run, hazy_model::strategy s, struct skiing_spec &ski_spec, bool should_build_ova);

int main(int argc, char **argv) {
  opterr = 1;
  int server_id = 0;
  int iDense  = 0, nLabels = 0;
  int dim     = 10, ova = 0, epochs = 5;
  char *entity_file = NULL;
  char *train_file_for_hot_run = NULL;
  string dbname = "";
  double m_factor = 0;
  int strategy_algorithm = 0;
  static struct extended_option long_options[] = {
    {"dense"    , no_argument, &iDense, 1, "if input contains dense vectors"},
    {"ova"      , no_argument, &ova, 1, "use one-versus-all classification"},
    {"labels"   , required_argument, NULL, 'l', "number of labels"},
    {"strategy", required_argument, NULL, 's', "strategy:0:LAZY_NAIVE (default),1: EAGER_NAIVE,2:LAZY_HAZY,3:EAGER_HAZY"},
    {NULL,0,NULL,0,0}
  };
  char usage[] = "<dim> <dbname> <m_factor> <entity_file> <train_file_for_hot_run> <server_id>";
  
  int c = 0, option_index;
  while( (c = getopt_long(argc, argv, "m:", convert_extended_options(long_options), &option_index)) != -1) {
    switch(c) {
    case 'e':
      epochs = atoi(optarg);
      break;
    case 'l':
      nLabels = atoi(optarg);
      break;
    case 's':
      strategy_algorithm = atoi(optarg);
      break;
    case ':':
    case '?':
      print_usage(long_options, argv[0], usage);
    exit(-1);
    break;
    default:
      break;
    }
  }
  if(optind == argc - 6) {
    dim = atoi(argv[optind]);
    dbname = argv[optind + 1];
    m_factor = atof(argv[optind + 2]);
    entity_file = argv[optind + 3];
    train_file_for_hot_run = argv[optind + 4];
    server_id  = atoi(argv[optind+5]);
  } else {
    print_usage(long_options, argv[0], usage);
    exit(-1);
  }
  
  main_memory_storage_spec *spec = new main_memory_storage_spec();
  spec->unique_id               = 0;
  spec->bRebuild                = true; 
  switch(strategy_algorithm) {
  case 0: spec->strategy      = hazy_model::LAZY_NAIVE;
    break;
  case 1: spec->strategy      = hazy_model::EAGER_NAIVE;
    break;
  case 2: spec->strategy      = hazy_model::LAZY_HAZY;
    break;
  case 3: 
    {
      cout << "in eager hazy" << endl;
      spec->strategy      = hazy_model::EAGER_HAZY;
    }
    break;
  default: cerr << "problem with strategy & algorithm: " << usage << endl;
  }
  skiing_spec ski_spec;
  ski_spec.initial_resort = 1.0;
  ski_spec.taoVal = 0.001;
  ski_spec.m_factor = m_factor;
  ski_spec.ratio = 1.0;
  if(iDense)
    startServer<FVector>(server_id, dbname, dim, nLabels, entity_file, train_file_for_hot_run, spec->strategy, ski_spec, ova == 1);
  else
    startServer<SVector>(server_id, dbname, dim, nLabels, entity_file, train_file_for_hot_run, spec->strategy, ski_spec, ova == 1);
  
  return 0;
}

template<class T>
struct worker_thread_info {
  Hazy_Worker *hw;
  Hazy_Node<T> *tree;
  int server_id;
  
  worker_thread_info(Hazy_Worker *_hw, Hazy_Node<T> *_tree, int _server_id) {
    hw = _hw;
    tree = _tree;
    server_id = _server_id;
  }
};

template<class T>
void*
worker_thread(void *data) {
  worker_thread_info<T> *wit = (worker_thread_info<T> *) data;
  Hazy_Worker *hw = wit->hw;
  Hazy_Node<T> *tree = wit->tree;
  int server_id = wit->server_id;
  
  int nSize = 2048;
  message_data answer = (char *) malloc(nSize * sizeof(char));
  bool close_conn = false;
  
  while (1) {
    void *data = hw->readData();
    if(data == NULL) {
      break;
    }
    
    switch(get_message_type(data)) {
    case READ_SE_EAGER:
      {
	int view_id = 0;
	int entity_id = 0;
	parse_se_eager_msg(data, view_id, entity_id);
	int entity_label = -1;
	tree->single_entity_read(entity_id, entity_label);
	LOGGING_ONLY(cout << "label: " << entity_label << endl;);
	
      	if(build_read_se_answer(entity_label, answer, nSize) < 0) {
	  cerr << "problem in build msg";
          return NULL;
      	}
      }
      break;
    case READ_COUNT:
      {
	/*int view_id = 0;
	  parse_count_msg(data, view_id);
	  int numInClass = 0;
	  sgd->readNumInClass(1, numInClass);
	  LOGGING_ONLY(cout << "count: " << numInClass << endl;);
	  
	  if(build_read_count_answer(numInClass, answer, nSize) < 0) {
	  cerr << "problem in build msg";
          return NULL;
	  }*/
      }
      break;
    case UPDATE:
      {
	int view_id = 0;
	T vec;
	int classOfExample = 0;
	parse_update_msg(data, view_id, vec, classOfExample);
	tree->insert_example(classOfExample, vec);
	
	if(build_update_answer(answer, nSize) < 0) {
	  cerr << "problem in build msg";
          return NULL;
      	}	
      }
      break;
    case UPDATE_BIN:
      {
        int view_id = 0;
        T vec;
        int classOfExample = 0;
	LOGGING_ONLY(Timer update_timer(true););	
        parse_update_msg_bin(data, view_id, vec, classOfExample);
	
	LOGGING_ONLY(Timer insert_time(true););
	VERBOSE_ONLY(std::cout << "tree insert called" << std::endl;);
        tree->insert_example(classOfExample, vec);
	LOGGING_ONLY(std::cout << "Insert timer:" << insert_time.stop() << std::endl;);
	
        if(build_update_answer(answer, nSize) < 0) {
          cerr << "problem in build msg";
          return NULL;
        }
	LOGGING_ONLY(std::cout << "Update Bin" << update_timer.stop() << std::endl;);
      }   
      break;
    case QUERY_SERVER_ID:
      {
	if(build_query_server_answer(server_id, answer, nSize) < 0) {
	  cerr << "problem in build msg";
          return NULL;
	}
      }
      break;
    case CLOSE_CONNECTION:
      {
	close_conn = true;
	if(build_close_connection_answer((int) close_conn, answer, nSize) < 0) {
	  cerr << "problem in build msg";
	  return NULL;
	}
      }
      break;
    case SERVER_NOOP:
      {
	int ack = 1;
	if(build_noop_server_answer(ack, answer, nSize) < 0) {
	  cerr << "problem in build msg";
          return NULL;
	}
      }
      break;
    default: 
      cerr << "there is no such function code (" << get_message_type(data) << ")" << endl;
    }
    
    int success = hw->writeData(answer, strlen((char *) answer), false);
    if(success < 0)
      cerr << "error when writing to the client: " << strerror(errno);
    if(close_conn)
      break;
  }
  
  ///Server's client connection is closed
  hw->closeClientConnection();
  delete hw;
  if(close_conn) {
    ///Server connection is closed
    cout << "server is being closed" << endl;
    serv->closeConnection();
    exit(0);
  }
  return NULL;
}


template<class T>
void startServer(int server_id, string dbname, int dim, int nLabels, const char *entity_file, const char *train_file_for_hot_run, hazy_model::strategy s, struct skiing_spec &ski_spec, bool should_build_ova) {
  Hazy_Database *db_conn = new Hazy_Database(dbname);
  vector<pthread_t> workers;
  
  Storage_Manager<T> **managers_lazy = new Storage_Manager<T>*[nLabels];
  IncrementalSGD<T> **sgds      = new IncrementalSGD<T>*[nLabels];
  Hazy_Sgd<T> **hazy_sgds_lazy       = new Hazy_Sgd<T>*[nLabels];
  Skiing ski_lazy(ski_spec);
  cout << "Building storage managers" << endl;
  struct main_memory_storage_spec mm_spec_lazy;
  mm_spec_lazy.unique_id = 0;
  mm_spec_lazy.bRebuild = false;
  mm_spec_lazy.strategy = hazy_model::LAZY_HAZY;
  
  cout << "strategy of mm is: " << s << endl;
  bool bPureSgd = true;
  for(int i = 0; i < nLabels - 1; i++) {
    managers_lazy[i] = new STLMainMemory_Storage_Manager<T>(&mm_spec_lazy);
    sgds[i]     = (bPureSgd) ? (IncrementalSGD<T>*) new TrueSGD<T>(dim, 1e-4) : (IncrementalSGD<T>*) new ReservoirSGD<T>(dim, 1e-4, r_size);
    hazy_model hm(sgds[i]->getModel(), hazy_model::LAZY_HAZY);
    managers_lazy[i]->loadFromFile(hm, entity_file);
    hazy_sgds_lazy[i] = new Multiclass_Specialized_Hazy_Sgd<T>(managers_lazy[i], sgds[i], ski_lazy, hazy_model::LAZY_HAZY);
    cout << (i + 1) << " sgd load completed" << endl;
  }
  cout << "We loaded everything!" << endl;
  Hazy_Node<T> *tree_lazy = should_build_ova ?  build_ova<T>(hazy_sgds_lazy, nLabels) : build_root_node<T>(hazy_sgds_lazy, nLabels);
  cout << "Tree is built." << endl;
  
  ifstream our_file(train_file_for_hot_run);
  if(our_file.is_open()) {
    std::string line;
    int line_num = 1;
    while(! our_file.eof() && getline(our_file, line) ){
      int label;
      T vec;
      parse_labeled_vector<T>(line, label, vec);
      tree_lazy->insert_example(label, vec);
      if(line_num % 1000 == 0)
	cout << "[multiclass_main] " << line_num << " loaded" << endl;
      
      line_num ++;
    }
    our_file.close();
    
  } else {
    cout << "Failed to open " << train_file_for_hot_run << endl;
    exit(-1);
  }
  
  cout << "Starting cleanup...." << endl;
  delete [] managers_lazy;
  delete [] hazy_sgds_lazy;
  cout << "done." << endl;
  
  Storage_Manager<T> **managers = new Storage_Manager<T>*[nLabels];
  Hazy_Sgd<T> **hazy_sgds       = new Hazy_Sgd<T>*[nLabels];
  Skiing ski(ski_spec);
  struct main_memory_storage_spec mm_spec;
  mm_spec.unique_id = 0;
  mm_spec.bRebuild = false;
  mm_spec.strategy = s;
  
  for(int i = 0; i < nLabels - 1; i++) {
    managers[i] = new STLMainMemory_Storage_Manager<T>(&mm_spec);
    hazy_model hm(sgds[i]->getModel(), s);
    managers[i]->loadFromFile(hm, entity_file);
    hazy_sgds[i] = new Multiclass_Specialized_Hazy_Sgd<T>(managers[i], sgds[i], ski, s);
    cout << (i + 1) << " sgd load completed" << endl;
  }
  cout << "We loaded everything!" << endl;
  Hazy_Node<T> *tree = should_build_ova ?  build_ova<T>(hazy_sgds, nLabels) : build_root_node<T>(hazy_sgds, nLabels);
  cout << "Tree is built." << endl;
  
  for(int i = 0; i < nLabels - 1; i ++) {
    ((Multiclass_Specialized_Hazy_Sgd<T>*) hazy_sgds[i])->resort();
  }
  
  cout << "tree is trained for hot run" << endl;  
  
  ///Server is initialized
  serv = new Hazy_Server();
  cout << "server started" << endl;
  
  while(1) {
    Hazy_Worker* hw = serv->acceptConnection();
    if (hw == NULL) {
      printf("Error at accept() : %s\n", strerror(errno));
      continue;
    }
    
    pthread_t t;
    worker_thread_info<T> *wit = new worker_thread_info<T>(hw, tree, server_id);
    int iret = pthread_create(&t, NULL, worker_thread<T>, (void *) wit);
    if(iret != 0 ) {
      cerr << "pthread create failed " << iret << endl;
      break;
    }
    workers.push_back(t);
  }
  
  ///Server connection is closed
  serv->closeConnection();
  db_conn->closeConnection();
}
