#include "multi_class.h"
#include "incremental_sgd.h"
#include "globals.h"
#include "timer.h"
#include "frontend_util.h"
#include <iostream>
#include <fstream>
#include "vectors.h"
#include "skiing.h"

using namespace std;

/**
 * This class implements multi-class classification that uses Main Memory architecture
 */


template
<class T>
void
insert_examples_from_file(Hazy_Node<T>* tree, const char* file_name) {
  ifstream our_file(file_name);
  if(our_file.is_open()) {
    std::string line;
    while(! our_file.eof() && getline(our_file, line) ){
      int label;
      T vec;
      parse_labeled_vector<T>(line, label, vec);
      tree->insert_example(label, vec);
    }
    our_file.close();
    
  } else {
    cout << "Failed to open " << file_name << endl;
    exit(-1);
  }
}

template
<class T>
void
print_entity_labels(Hazy_Node<T>* tree, const char* file_name) {
  ifstream our_file(file_name);
  int nExamples = 0;
  if(our_file.is_open()) {
    std::string line;
    while(! our_file.eof() && getline(our_file, line) ){
      int key;
      T vec;
      // here we assume they are labeled with their key
      parse_labeled_vector<T>(line, key, vec);
      int class_value = 0;
      tree->single_entity_read(key, class_value);
      cout << "\t key=" << key << " class=" << class_value << endl;
      nExamples++;
    }
    our_file.close();
    
  } else {
    cout << "Failed to open " << file_name << endl;
    exit(-1);
  }
}


template
<class T>
void
run_test(int dim, int nLabels, const char* train_file, const char *entity_file, hazy_model::strategy s, struct skiing_spec &ski_spec, bool should_build_ova, int epochs) {
  Timer run_timer(true);
  Storage_Manager<T> **managers = new Storage_Manager<T>*[nLabels];
  IncrementalSGD<T> **sgds      = new IncrementalSGD<T>*[nLabels];
  Hazy_Sgd<T> **hazy_sgds       = new Hazy_Sgd<T>*[nLabels];
  Skiing ski(ski_spec);
  cout << "Building storage managers" << endl;
  struct main_memory_storage_spec mm_spec;
  mm_spec.unique_id = 0;
  mm_spec.bRebuild = false;
  mm_spec.strategy = s;
  
  for(int i = 0; i < nLabels - 1; i++) {
    managers[i] = new STLMainMemory_Storage_Manager<T>(&mm_spec);
    sgds[i]     = new TrueSGD<T>(dim, 1e-4);
    hazy_model hm(sgds[i]->getModel(), s);
    managers[i]->loadFromFile(hm, entity_file);
    hazy_sgds[i] = new Hazy_Sgd<T>(managers[i], sgds[i], ski, s);
  }
  cout << "We loaded everything!" << endl;
  Hazy_Node<T> *tree = should_build_ova ?  build_ova<T>(hazy_sgds, nLabels) : build_root_node<T>(hazy_sgds, nLabels); 
  cout << "Tree is built." << endl;
  
  for(int k = 0; k < epochs; k++) {
    Timer training_timer(true);
    insert_examples_from_file<T>(tree, train_file);
    cout << " [epoch=" << k << "] We added examples in " << training_timer.stop() << endl;
  }
  
  Timer printing_timer(true);
  print_entity_labels<T>(tree, entity_file);
  cout << "Finished printing in " << printing_timer.stop() << endl;
  
  cout << "Finished run in " << run_timer.stop() << endl;
}

int main(int argc, char **argv) {
  opterr = 1;
  main_memory_storage_spec *spec = new main_memory_storage_spec();
  spec->bRebuild                = true; 
  spec->strategy                = hazy_model::LAZY_HAZY;
  skiing_spec ski_spec = {2,1e-2,1, 1.0};
  int iDense  = 0, nLabels = 0;
  int dim     = 10, ova = 0, epochs = 5;
  char *entity_file = NULL, *train_file = NULL;
  static struct extended_option long_options[] = {
    {"dense"    , no_argument, &iDense, 1, "if input contains dense vectors"},
    {"ova"      , no_argument, &ova, 1, "use one-versus-all classification"},
    {"labels"   , required_argument, NULL, 'l', "number of labels"},
    {"epochs"   , required_argument, NULL, 'e', "number of epochs: default is 5"},
    {"dim"      , required_argument, NULL, 'd', "the dimension of the model"},
    {NULL,0,NULL,0,0} 
  };
  char usage[] = "<entity_file> <train_file>";
  
  int c = 0, option_index;
  while( (c = getopt_long(argc, argv, "m:", convert_extended_options(long_options), &option_index)) != -1) {
    switch(c) {      
    case 'e':
      epochs = atoi(optarg);
      break;
    case 'd':
      dim = atoi(optarg);
      break;
    case 'l':
      nLabels = atoi(optarg);
      break;
    case ':':
    case '?':
      print_usage(long_options, argv[0], usage);
    exit(-1);
    break;
    default:
      break;
    }
  }
  if(optind == argc - 2) {
    entity_file = argv[optind];
    train_file  = argv[optind+1];
  } else {
    print_usage(long_options, argv[0], usage);
    exit(-1);
  }
  if(iDense == 1) {
    run_test<FVector>(dim, nLabels, train_file, entity_file, spec->strategy, ski_spec, ova == 1, epochs );
  } else {
    run_test<SVector>(dim, nLabels, train_file, entity_file, spec->strategy, ski_spec, ova == 1, epochs );
  }
}
