#include "postgres_util.h"
#include "storage_manager.h"
#include "ondisk_storage_manager.h"
#include "hybrid_storage_manager.h"
#include "main_memory_storage_manager.h"
#include "incremental_sgd.h"
#include "globals.h"
#include "frontend_util.h"
#include <errno.h>
#include <string>
#include <vector>

using namespace std;

/**
 * This class implements regression tests for the storage manager
 */

static int bDense = 0;
static int bPureSGD = 0;
static int mm = 0;
static int hybrid = 0;
static int rsize = 100;
static int dim = 100;
static string db_name = "";
static int test_num = 0;
static int strategy = 0;
static int _useepsmap = 1;
int buffer_size = 10;

enum t_storage_manager { SM_HYBRID, SM_MAIN_MEMORY, SM_ONDISK };

/**
 * Reads entities from the file
 * \param[in] filename name of the file that contains entities
 * \param[out] entity_ids the vector that contains entity ids
 * \param[out] entity_vectors the vector that contains entity vectors
 */
template<class T>
void read_file_into_vectors(const char *filename, vector<int> &entity_ids, vector<T> &entity_vectors) {
  ifstream our_file(filename);
  
  if(our_file.is_open()) {
    std::string line;
    while(! our_file.eof() && getline(our_file, line) ){
      int label;
      T vec;
      parse_labeled_vector<T>(line, label, vec);
      entity_ids.push_back(label);
      entity_vectors.push_back(vec);
    }
    our_file.close();
  } else {
    cout << "Failed to open " << filename << endl;
    exit(-1);
  }
  
}

/**
 * Classifies given entity vector with the given model and returns the class
 * \param[in] hm hazy model
 * \param[in] s feature vector of the entity
 * \return class of the entity
 */
template<class T>
int getLabel(struct hazy_model &hm, T s) {
  double prediction = IncrementalSGD<T>::classifyExample(hm._model, s);
  return (prediction > 0) ? 1 : 0;
}

/**
 * Reads entities from the file and load them into both storage manager and vectors
 * \param[in] sm storage manager
 * \param[in] hm hazy model
 * \param[in] filename name of the file that contains entities
 * \param[out] entity_ids the vector for entity ids
 * \param[out] entity_vectors the vector for entity vectors
 */
template<class T>
void loadEntitiesToSMAndVectors(Storage_Manager<T> *sm, struct hazy_model &hm, const char *filename, vector<int> &entity_ids, vector<T> &entity_vectors) {
  //sm->loadFromFile(hm, filename);
  //sm->loadFromDatabase(hm);
  read_file_into_vectors(filename, entity_ids, entity_vectors);
}

/**
 * For all entities in the vector, gets their labels from storage manager and compare the classes retrieved by this class and check whether they are same or not
 * \param[in] sm storage manager
 * \param[in] hm hazy model
 * \param[in] entity_ids id of the entities
 * \param[in] entity_vectors vectors of the entities
 * \param[out] method_successful success
 */
template<class T>
void testGetEntityClass(Storage_Manager<T> *sm, struct hazy_model &hm, vector<int> entity_ids, vector<T> entity_vectors, bool &method_successful) {
  for(unsigned int i = 0; i < entity_ids.size(); i ++) {
    int entity_id = entity_ids.at(i);
    
    int memory_label = getLabel(hm, entity_vectors.at(i));
    int storage_label = 0;
    sm->getEntityClass(entity_id, storage_label, hm);
    
    if(storage_label != memory_label) {
      method_successful = false;      
      cerr << "[ERROR] error in get entity class:" << entity_id << " storage=" << storage_label << " memory_label=" << memory_label << endl;
      cerr << entity_vectors.at(i) << endl;
      cerr << "[ERROR] eps=" << IncrementalSGD<T>::classifyExample(hm._model, entity_vectors.at(i)) << endl;
      cerr << "[ERROR] hm model: " << hm._model << endl;
      break;
    }
  }
}

/**
 * For all entities in the vector, gets their labels from storage manager and compare the classes retrieved by this class and check whether they are same or not
 * Compares # of the entities in class 1 that come from storage manager and calculated in this class
 * First, num in class is retrieved from storage manager and then for all entities label is calculated in this class. Then, we find the num in class and we compare it with the result of storage manager
 * \param[in] sm storage manager
 * \param[in] hm hazy model
 * \param[in] entity_ids id of the entities
 * \param[in] entity_vectors vectors of the entities
 * \param[out] method_successful success
 */
template<class T>
void testGetNumInClass(Storage_Manager<T> *sm, struct hazy_model &hm, vector<int> entity_ids, vector<T> entity_vectors, bool &method_successful) {
  for(unsigned int i = 0; i < entity_ids.size(); i ++) {
    int memory_label1Count = 0;
    for(unsigned int j = 0; j < entity_vectors.size(); j ++)
      memory_label1Count += getLabel(hm, entity_vectors.at(j));
    
    int storage_label1Count = 0;
    double waste = 0;
    sm->getNumInClass(1, storage_label1Count, hm, waste);
    
    if(memory_label1Count != storage_label1Count) {
      method_successful = false;
      cerr << "error in get num in class test=" << memory_label1Count << " storage=" << storage_label1Count << endl;
      break;
    }
  }
}

/**
 * Tests whether for all entities, we correctly get the class of the entity correctly
 * \param[in] sm storage manager
 * \param[in] filename name of the file that contains entities
 * \param[in] hm hazy model
 * \param[in] sgd sgd model
 * \param[in] training_file_name name of the file that contains training examples
 * \param[out] method_successful success
 */
template<class T>
void test1(Storage_Manager<T> *sm, const char* filename, struct hazy_model& hm, IncrementalSGD<T> *sgd, const char *training_file_name, bool &method_successful) {
  method_successful = true;
  vector<int> entity_ids;
  vector<T> entity_vectors;
  
  loadEntitiesToSMAndVectors(sm, hm, filename, entity_ids, entity_vectors);
  
  testGetEntityClass(sm, hm, entity_ids, entity_vectors, method_successful);	
}


//with given model and entities, only test get num in class
//currently only class 1 is being tested, since in storage manager code we have restriction on that.
/**
 * Tests whether we correctly get the number in class
 * \param[in] sm storage manager
 * \param[in] filename name of the file that contains entities
 * \param[in] hm hazy model
 * \param[in] sgd sgd model
 * \param[in] training_file_name name of the file that contains training examples
 * \param[out] method_successful success
 */
template<class T>
void test2(Storage_Manager<T> *sm, const char* filename,  struct hazy_model& hm, IncrementalSGD<T> *sgd, const char *training_file_name, bool &method_successful) {
  method_successful = true;
  vector<int> entity_ids;
  vector<T> entity_vectors;
  
  loadEntitiesToSMAndVectors(sm, hm, filename, entity_ids, entity_vectors);
  testGetNumInClass(sm, hm, entity_ids, entity_vectors, method_successful);
}

/**
 * Tests that incremental updates are working without any error
 * \param[in] sm storage manager
 * \param[in] filename name of the file that contains entities
 * \param[in] hm hazy model
 * \param[in] sgd sgd model
 * \param[in] training_file_name name of the file that contains training examples
 * \param[out] method_successful success
 */
template<class T>
void test3(Storage_Manager<T> *sm, const char* filename,  struct hazy_model& hm, IncrementalSGD<T> *sgd, const char *training_file_name, bool &method_successful) {
  method_successful = true;
  
  if(hm.isEager()) {
    vector<int> entity_ids;
    vector<T> entity_vectors;
    
    loadEntitiesToSMAndVectors(sm, hm, filename, entity_ids, entity_vectors);
    
    ifstream our_file2(training_file_name);
    
    if(our_file2.is_open()) {
      std::string line;
      
      while(! our_file2.eof() && getline(our_file2, line) ){
	int label;
	T vec;
	parse_labeled_vector<T>(line, label, vec);
	sgd->addExample(label, vec);
	double waste = 0;
	
	sm->incrementalUpdate(hm, waste);
      }
    }
  }
}

//only test whether resort is working without giving error
/**
 * Tests that resort at each step is working without any error
 * \param[in] sm storage manager
 * \param[in] filename name of the file that contains entities
 * \param[in] hm hazy model
 * \param[in] sgd sgd model
 * \param[in] training_file_name name of the file that contains training examples
 * \param[out] method_successful success
 */
template<class T>
void test4(Storage_Manager<T> *sm, const char* filename,  struct hazy_model& hm, IncrementalSGD<T> *sgd, const char *training_file_name, bool &method_successful) {
  method_successful = true;
  
  if(hm._strategy == hazy_model::EAGER_HAZY) {
    vector<int> entity_ids;
    vector<T> entity_vectors;
    
    loadEntitiesToSMAndVectors(sm, hm, filename, entity_ids, entity_vectors);
    
    ifstream our_file2(training_file_name);
    
    if(our_file2.is_open()) {
      std::string line;
      
      while(! our_file2.eof() && getline(our_file2, line) ){
      	int label;
	T vec;
      	parse_labeled_vector<T>(line, label, vec);
      	sgd->addExample(label, vec);
	sm->resort(hm);
      }
    }
  }
}

/**
 * This method tests whether we successfully update and read together
 * Tests that when we do incremental update at each step, then either get entity test or get num in class test works correctly (determined by isTest1 boolean)
 * Meaning that whether we successfully read after each update or not.
 * \param[in] sm storage manager
 * \param[in] filename name of the file that contains entities
 * \param[in] hm hazy model
 * \param[in] sgd sgd model
 * \param[in] training_file_name name of the file that contains training examples
 * \param[out] method_successful success
 * \param[in] isTest1 if this is true, get entity class is tested, else get num in class is tested
 */
template<class T>
//test whether at each step, if we incremental update, then still we get correct answers for single entity read or count reads
void test5(Storage_Manager<T> *sm, const char* filename,  struct hazy_model& hm, IncrementalSGD<T> *sgd, const char *training_file_name, bool &method_successful, bool isTest1) {
  DEBUG_ONLY(cout << "in test5 " << isTest1 << endl;);
  method_successful = true;
  
  vector<int> entity_ids;
  vector<T> entity_vectors;
  
  loadEntitiesToSMAndVectors(sm, hm, filename, entity_ids, entity_vectors);
  
  ifstream our_file2(training_file_name);
  
  if(our_file2.is_open()) {
    std::string line;
    
    while(! our_file2.eof() && getline(our_file2, line) ){
      int label;
      T vec;
      parse_labeled_vector<T>(line, label, vec);
      sgd->addExample(label, vec);
      double waste = 0;
      
      LOGGING_ONLY(cout << "sgd is: " << sgd->getModel().w << endl;);
      //if(hm.isEager()) {
      sm->incrementalUpdate(hm, waste);
      //}
      if(isTest1)
	testGetEntityClass(sm, hm, entity_ids, entity_vectors, method_successful);
      else
	testGetNumInClass(sm, hm, entity_ids, entity_vectors, method_successful);
      if(!method_successful)
	break;
    }
  }
}

/**
 * This method tests whether we successfully resort and read together
 * Tests that when we do resort at each step, then either get entity test or get num in class test works correctly (determined by isTest1 boolean)
 * Meaning that whether we successfully read after each resort or not.
 * \param[in] sm storage manager
 * \param[in] filename name of the file that contains entities
 * \param[in] hm hazy model
 * \param[in] sgd sgd model
 * \param[in] training_file_name name of the file that contains training examples
 * \param[out] method_successful success
 * \param[in] isTest1 if this is true, get entity class is tested, else get num in class is tested
 */
//test whether at each step, if we resort, then still we get correct answers for single entity read or count reads
template<class T>
void test6(Storage_Manager<T> *sm, const char* filename,  struct hazy_model& hm, IncrementalSGD<T> *sgd, const char *training_file_name, bool &method_successful, bool isTest1) {
  method_successful = true;
  
  if(hm._strategy == hazy_model::EAGER_HAZY) {
    vector<int> entity_ids;
    vector<T> entity_vectors;
    
    loadEntitiesToSMAndVectors(sm, hm, filename, entity_ids, entity_vectors);
    
    ifstream our_file2(training_file_name);
    
    if(our_file2.is_open()) {
      std::string line;
      
      while(! our_file2.eof() && getline(our_file2, line) ){
	int label;
	T vec;
	parse_labeled_vector<T>(line, label, vec);
	sgd->addExample(label, vec);
	sm->resort(hm);
	if(isTest1)
	  testGetEntityClass(sm, hm, entity_ids, entity_vectors, method_successful);
	else
	  testGetNumInClass(sm, hm, entity_ids, entity_vectors, method_successful);
	if(!method_successful)
	  break;
      }
    }
  }
}

/**
 * Copies into model from hazy model
 * \param[out] model model to be copied from given hazy model
 * \param[in] hm hazy model
 */
void copyModel(struct model &model, struct hazy_model &hm) {
  model.bias = hm._model.bias;
  model.wscale = hm._model.wscale;
  for(int i = 0; i < hm._model.w.size(); i ++)
    model.w.set(i, hm._model.w.get(i));
}


/**
 * Calculates delta differences from the given models
 * \param[in] current current model
 * \param[in] old last model from the sort
 * \param[in] m_factor m factor
 * \param[out] delta_w delta value between w's
 * \param[out] delta_b delta value between biases
 */
void
holder_difference_models(struct model current, struct model old, double m_factor, double &delta_w, double &delta_b) {// outputs
  delta_b = current.bias - old.bias;
  delta_w = 0.0;
  for(int d = 0; d < dim; d ++) {
    delta_w = std::max(delta_w, fabs((current.w.get(d) * current.wscale - old.w.get(d)*old.wscale)));
  }
  // Multiply with the max-1 norm of the dataset (why is this stored in skiing?)
  // Seems like this should be stored here....
  delta_w *= m_factor;
}

/**
 * This method tests whether we successfully update and read together
 * It simulates hazy algorithm by doing both resort and incremental update together. However, instead of skiing strategy, it does resort at every 5 step and incremental update otherwise.
 * Tests that when we do incremental update or resort at each step, then either get entity test or get num in class test works correctly (determined by isTest1 boolean)
 * Meaning that whether we successfully read after each update or not.
 * \param[in] sm storage manager
 * \param[in] filename name of the file that contains entities
 * \param[in] hm hazy model
 * \param[in] sgd sgd model
 * \param[in] training_file_name name of the file that contains training examples
 * \param[out] method_successful success
 * \param[in] isTest1 if this is true, get entity class is tested, else get num in class is tested
 */
//test whether at each step, if we incrementally update, then still we get correct answers for entity class reads or count reads
template<class T>
void test7(Storage_Manager<T> *sm, const char* filename,  struct hazy_model& hm, IncrementalSGD<T> *sgd, const char *training_file_name, bool &method_successful, bool isTest1) {
  struct model _last_sort_model;
  //using _last_sort_model = model(hm._model); ?
  double m_factor = 1;
  copyModel(_last_sort_model, hm);
  
  method_successful = true;
  
  if(hm._strategy == hazy_model::EAGER_HAZY) {
    vector<int> entity_ids;
    vector<T> entity_vectors;
    
    loadEntitiesToSMAndVectors(sm, hm, filename, entity_ids, entity_vectors);
    
    ifstream our_file2(training_file_name);
    
    if(our_file2.is_open()) {
      std::string line;
      int example_num = 1;
      
      while(! our_file2.eof() && getline(our_file2, line) ){
        int label;
        T vec;
        parse_labeled_vector<T>(line, label, vec);
        sgd->addExample(label, vec);
	
	// compute the low and high water.
	double delta_bias = 0.0, delta_w = 0.0;
	holder_difference_models(sgd->getModel(), _last_sort_model, m_factor, delta_w, delta_bias);
	
	double cond_plus = -(delta_bias - delta_w);
	double cond_neg = -(delta_bias + delta_w);     
	hm.low_water  = std::min(hm.low_water, cond_neg);
	hm.high_water = std::max(hm.high_water, cond_plus);
	
	if(example_num % 5 == 0) {
	  sm->resort(hm);
	  copyModel(_last_sort_model, hm);				
	} else {	  
	  double waste_time = 0.0;
	  sm->incrementalUpdate(hm, waste_time);
	}
	
	if(isTest1)
	  testGetEntityClass(sm, hm, entity_ids, entity_vectors, method_successful);
	else
	  testGetNumInClass(sm, hm, entity_ids, entity_vectors, method_successful);
        if(!method_successful)
          break;
	
	example_num ++;
      }
    }
  }
}


/**
 * Executes the test with the given parameter
 * \param[in] sm storage manager
 * \param[in] hm hazy model
 * \param[in] filename name of the file that contains the entities
 * \param[in] sgd sgd model
 * \param[in] training_file_name name of the training file
 * \param[out] method_successful success
 * \param[in] testNum which test to execute
 */
template<class T>
void run_test(Storage_Manager<T> *sm, struct hazy_model &hm, const char *filename, IncrementalSGD<T> *sgd, const char *training_file_name, bool &method_successful, int testNum) {
  switch(testNum) {
  case 1: test1<T>(sm, filename, hm, sgd, training_file_name, method_successful);
    break;
  case 2: test2<T>(sm, filename, hm, sgd, training_file_name, method_successful);
    break;
  case 3: test3<T>(sm, filename, hm, sgd, training_file_name, method_successful);
    break;
  case 4: test4<T>(sm, filename, hm, sgd, training_file_name, method_successful);
    break;
  case 5: test5<T>(sm, filename, hm, sgd, training_file_name, method_successful, true);
    break;
  case 6: test5<T>(sm, filename, hm, sgd, training_file_name, method_successful, false);
    break;
  case 7: test6<T>(sm, filename, hm, sgd, training_file_name, method_successful, true);
    break;
  case 8: test6<T>(sm, filename, hm, sgd, training_file_name, method_successful, false);
    break;
  case 9: test7<T>(sm, filename, hm, sgd, training_file_name, method_successful, true);
    break;
  case 10: test7<T>(sm, filename, hm, sgd, training_file_name, method_successful, false);
    break;
  default: cout << "test is not available" << endl;
  }
}


/**
 * Creates ondisk storage manager
 * \param[in] s ondisk storage spec
 * \param[in] sgd sgd model
 * \param[in] filename name of the file that contains entities
 * \param[in] training_file_name name of the file that contains training examples
 * \param[out] method_successful success
 * \param[in] testNum which test to execute
 * \param[in] hm hazy model
 * \param[in] dense whether feature vectors are dense or not
 * \param[in] db_conn database connection
 */ 
template<class T>
void prepare_ondisk_storage_manager(struct ondisk_storage_spec *s, IncrementalSGD<T> *sgd, const char *filename, const char *training_file_name, bool &method_successful, int testNum, struct hazy_model &hm, bool dense, Hazy_Database *db_conn) {
  Ondisk_Storage_Manager_Factory<T>      x;
  Storage_Manager<T>* sm          = (Storage_Manager<T>*) x.build_storage_manager(db_conn, s, hm);
  run_test<T>(sm, hm, filename, sgd, training_file_name, method_successful, testNum);
}

/**
 * Creates hybrid storage manager
 * \param[in] s ondisk storage spec
 * \param[in] sgd sgd model
 * \param[in] filename name of the file that contains entities
 * \param[in] training_file_name name of the file that contains training examples
 * \param[out] method_successful success
 * \param[in] testNum which test to execute
 * \param[in] hm hazy model
 * \param[in] dense whether feature vectors are dense or not
 * \param[in] db_conn database connection
 * \param[in] nBuffer buffer size
 * \param[in] useeps whether to use eps map or not
 */ 
template<class T>
void prepare_hybrid_storage_manager(struct ondisk_storage_spec *s, IncrementalSGD<T> *sgd, const char*filename, const char *training_file_name, bool &method_successful, int testNum, struct hazy_model &hm, bool dense, Hazy_Database *db_conn, int nBuffer, bool useeps) {
  Ondisk_Storage_Manager_Factory<T>      x;
  Ondisk_Storage_Manager<T> *disk = x.build_storage_manager(db_conn, s, hm);
  Storage_Manager<T> *sm   = (Storage_Manager<T>*) new Hybrid_Storage_Manager<T>(db_conn, s, disk, nBuffer, useeps);
  run_test<T>(sm, hm, filename, sgd, training_file_name, method_successful, testNum);
}

/**
 * Creates main memory storage manager
 * \param[in] s main memory storage spec
 * \param[in] sgd sgd model
 * \param[in] filename name of the file that contains entities
 * \param[in] training_file_name name of the file that contains training examples
 * \param[out] method_successful success
 * \param[in] testNum which test to execute
 * \param[in] hm hazy model
 * \param[in] db_conn database connection
 */ 
template<class T>
void prepare_mm_storage_manager(struct main_memory_storage_spec *s, IncrementalSGD<T> *sgd, const char*filename, const char *training_file_name, bool &method_successful, int testNum, struct hazy_model &hm, Hazy_Database *db_conn) {
  Storage_Manager<T>* sm          = (Storage_Manager<T>*) new STLMainMemory_Storage_Manager<T>(db_conn, s, hm);
  
  run_test<T>(sm, hm, filename, sgd, training_file_name, method_successful, testNum);
}

/*template<class T>
  void test_spec(Hazy_Database *db_conn, struct ondisk_storage_spec &s, char* filename, IncrementalSGD<T> *sgd, char *training_file_name, bool &method_successful, int testNum) {
  struct hazy_model hm(sgd->getModel(), s.strategy);
  cout << "here" << endl;
  
  hm.low_water = 0;
  hm.high_water = 0;
  Storage_Manager<T>* sm          = (Storage_Manager<T>*) new Ondisk_Storage_Manager(db_conn, s);
  cout << "Constructed the storage manager " << endl;
  cout << "Test Concluded" << endl;
  
  switch(testNum) {
  case 1: test1(sm, filename, hm, sgd, training_file_name, method_successful);
  break;
  case 2: test2(sm, filename, hm, sgd, training_file_name, method_successful);
  break;
  case 3: test3(sm, filename, hm, sgd, training_file_name, method_successful);
  break;
  case 4: test4(sm, filename, hm, sgd, training_file_name, method_successful);
  break;
  case 5: test5(sm, filename, hm, sgd, training_file_name, method_successful, true);
  break;
  case 6: test5(sm, filename, hm, sgd, training_file_name, method_successful, false);
  break;
  case 7: test6(sm, filename, hm, sgd, training_file_name, method_successful, true);
  break;
  case 8: test6(sm, filename, hm, sgd, training_file_name, method_successful, false);
  break;
  case 9: test7(sm, filename, hm, sgd, training_file_name, method_successful, true);
  break;
  case 10: test7(sm, filename, hm, sgd, training_file_name, method_successful, false);
  break;
  default: cout << "test is not available" << endl;
  }
  }*/


/**
 * Initializes the storage manager, spec and others required for the test
 * \param[in] lambda
 * \param[in] entity_table_name name of the table that contains entities
 * \param[in] entity_file_name name of the file that contains entities
 * \param[in] training_file_name name of the file that contains training examples
 * \param[out] method_successful success
 * \param[in] bDense whether feature vectors are dene or not
 * \param[in] storage_manager storage manager
 */
template<class T>
void initialize(double lambda, string entity_table_name, string entity_file_name, string training_file_name, 
		bool &method_successful, bool bDense, t_storage_manager storage_manager) {
  Hazy_Database *db_conn = new Hazy_Database(db_name);
  
  IncrementalSGD<T> *sgd  = bPureSGD ? (IncrementalSGD<T>*) new TrueSGD<T>(dim,lambda) : (IncrementalSGD<T>*) new ReservoirSGD<T>(dim, lambda, rsize);
  
  hazy_model::strategy h_s = hazy_model::LAZY_NAIVE;
  
  switch(strategy) {
  case 0: h_s = hazy_model::LAZY_NAIVE;
    break;
  case 1: h_s = hazy_model::EAGER_NAIVE;
    break;
  case 2: h_s = hazy_model::LAZY_HAZY;
    break;
  case 3: h_s = hazy_model::EAGER_HAZY;
    break;
  default: cout << "there is no strategy";
  } 
  
  
  struct hazy_model hm(sgd->getModel(), h_s);
  hm.low_water = 0; 
  hm.high_water = 0;
  
  switch (storage_manager) {
  case SM_ONDISK: {
    cout << "on-disk version" << endl;
    struct ondisk_storage_spec *od_spec = new ondisk_storage_spec();
    od_spec->unique_id               = 0;//we need id for prepared statements
    od_spec->entity_table_name       = entity_table_name;
    od_spec->external_table_name     = std::string("sm_test_external_table");
    od_spec->intermediate_table_name = std::string("sm_test_intermediate_table");
    od_spec->bRebuild                = true;
    od_spec->strategy = h_s;
    
    prepare_ondisk_storage_manager<T>(od_spec, sgd, entity_file_name.c_str(), training_file_name.c_str(), method_successful, test_num, hm, bDense, db_conn);
    break;
  }
  case SM_MAIN_MEMORY: {
    cout << "main memory version" << endl;
    struct main_memory_storage_spec *mm_spec = new main_memory_storage_spec();
    mm_spec->unique_id               = 0;
    mm_spec->bRebuild                = true;
    mm_spec->entity_table_name       = entity_table_name;
    mm_spec->strategy = h_s;
    
    prepare_mm_storage_manager<T>(mm_spec, sgd, entity_file_name.c_str(), training_file_name.c_str(), method_successful, test_num, hm, db_conn);
    break;
  }
  case SM_HYBRID: {
    struct ondisk_storage_spec *h_spec = new ondisk_storage_spec();
    h_spec->unique_id               = 0;
    h_spec->entity_table_name       = entity_table_name;
    h_spec->external_table_name     = std::string("sm_test_external_table");
    h_spec->intermediate_table_name = std::string("sm_test_intermediate_table");
    h_spec->bRebuild                = true;
    h_spec->strategy = h_s;
    prepare_hybrid_storage_manager<T>(h_spec, sgd, entity_file_name.c_str(), training_file_name.c_str(), method_successful, test_num, hm, bDense, db_conn,buffer_size, _useepsmap == 1);
    break;
  }
  default:
    cout << "Unknown storage manager " << endl;
    abort();
    break;
  }
  
}

int
main(int argc, char**argv) {
  /*	Hazy_Database *db = new Hazy_Database("test_new_system");
	vector< sParse_IDvec<SVector> > _out;
	db->postgresTupleParser("SELECT id, feature_vector FROM sm_test_entity_table_name;", sParse_IDvec<SVector>::parse_tuple, _out);
	
	for(unsigned int i = 0; i < _out.size(); i ++) {
	cout << _out.at(i).id << ", " << _out.at(i).f << endl;
	cout << _out.at(i).f.get(1) << endl;
	}
	
	vector< sParse_IDvec<FVector> > _out2;
	db->postgresTupleParser("SELECT id, feature_vector FROM sm_test_entity_table_name2;", sParse_IDvec<FVector>::parse_tuple, _out2);
	
	for(unsigned int i = 0; i < _out2.size(); i ++) {
	cout << _out2.at(i).id << ", " << _out2.at(i).f << endl;
	cout << _out2.at(i).f.get(1) << endl;
	}*/
  
  /*std::ifstream f_stream("/tmp/test_file2");
    string table_name = "test_table";
    string delimiter = ";";
    db->copy_to_table(f_stream, table_name, delimiter);
    std::ofstream t;
    t.open("/tmp/file2");
    db->copy_from_table(t, table_name, delimiter);*/
  
  /*return 1;*/
  
  int c = 0, option_index = 0;
  double lambda = 1e-4;
  char* usage_str = (char *) " <entity_file> <train_file> <test_num>";
  opterr = 1;
  
  static struct extended_option long_options[] = {
    {"dense"    , no_argument, &bDense, 1, "if input contains dense vectors"},
    {"puresgd"  , no_argument, &bPureSGD, 1, "if want to use PureSGD (instead of reservoir)"},
    {"mainmemory"  , no_argument, &mm, 1, "if want to use main memory (instead of on-disk)"},
    {"hybrid"  , no_argument, &hybrid, 1, "if want to use hybrid (instead of on-disk)"},
    {"dim"      , required_argument, NULL, 'd', "the dimension of the model"},
    {"lambda"   , required_argument, NULL, 'l', "lambda (learning rate)"},
    {"reservoir", required_argument, NULL, 'r', "size of resevoir to be used by ResevoirSGD"},
    {"db_name", required_argument, NULL, 'o', "name of database"},
    {"entity_table_name", required_argument, NULL, 'e', "name of database"},
    {"noepsmap", no_argument, &_useepsmap, 0, "Do not use the epsmap in hybrid"},
    {"strategy", required_argument, NULL, 's', "strategy:\n0: LAZY_NAIVE\n1: EAGER_NAIVE\n2:LAZY_HAZY\n3:EAGER_HAZY"},
    {NULL,0,NULL,0,0} 
  };
  
  
  string entity_table_name = "";
  
  // Somehow this parsing is not working as advertised.
  struct option* lo = convert_extended_options(long_options);
  while( (c = getopt_long(argc, argv, "r:d:m:o:l:", lo, &option_index)) != -1)
    {
      switch (c) {
      case 'r':
	rsize = atoi(optarg);
	break;
      case 'd':
	dim  = atoi(optarg);
	break;
      case 'l':
	lambda = atof(optarg);
	break;
      case 's':
	strategy = atoi(optarg);
	break;
      case 'o':
	db_name = optarg;
	break;
      case 'e':
	entity_table_name = optarg;
	break;
      case ':':
      case '?':
	print_usage(long_options, argv[0], usage_str);
      exit(-1);
      break;
      }
    }
  
  
  string entity_file_name = "", training_file_name = "";
  if(optind == argc - 3) {
    entity_file_name = argv[optind];
    training_file_name  = argv[optind+1];
    test_num = atoi(argv[optind + 2]);
  } else {
    print_usage(long_options, argv[0], usage_str);
    exit(-1);
  }
  
  bool method_successful = true;
  t_storage_manager storage_manager = SM_ONDISK;
  if(mm)     { storage_manager = SM_MAIN_MEMORY; }
  if(hybrid) { storage_manager = SM_HYBRID    ; }
  
  if(bDense) initialize<FVector>(lambda, entity_table_name, entity_file_name, training_file_name, method_successful, bDense, storage_manager);
  else initialize<SVector>(lambda, entity_table_name, entity_file_name, training_file_name, method_successful, bDense, storage_manager);
  
  cout << "test " << test_num << " for strategy " << strategy << " (" << (bDense ? "dense" : "sparse") << ") run " << (method_successful ? "successfully" : "failed") << endl;
    
}
