#include "incremental_sgd.h"
#include "globals.h"
#include "timer.h"
#include "frontend_util.h"
#include <iostream>
#include <fstream>


/**
 * This class implements training and testing routines with sgd
 */

using namespace std;

template <class T>
int 
correct_label(IncrementalSGD<T> *sgd, int label, T ex) {
  int _class = sgd->classifyExample(ex) > 0 ? 1 : 0;
  return ((label * (2*_class - 1)) == 1) ? 1 : 0;
}
  
template <class T>
int 
train(IncrementalSGD<T>* sgd, int label, T ex) {
  sgd->addExample(label, ex);
  return 0;
}

template <class T>
void apply_to_file(char* file_name, IncrementalSGD<T>* sgd, int (*foo)(IncrementalSGD<T>*,int,T), int &nExamples, int &nTotal) {
  ifstream our_file(file_name);
  nExamples = 0;
  if(our_file.is_open()) {
    std::string line;
    while(! our_file.eof() && getline(our_file, line) ){
      int label;
      T vec;
      parse_labeled_vector<T>(line, label, vec);
      nTotal += foo(sgd, label, vec);
      nExamples++;
    }
    our_file.close();
    
  } else {
    cout << "Failed to open " << file_name << endl;
    exit(-1);
  }
  
}

template <class T>
void
train_and_test(char* train_name, char* test_file, char* input_model_file, char *out_file, int dim, double lambda, bool pureSGD, int rsize) {
  DEBUG_ONLY(cout << "training=" << train_name << " test=" << test_file << " dim=" << dim << " lambda=" << lambda << " rsize=" << rsize << endl;);
  IncrementalSGD<T> *sgd  = pureSGD ? 
    (IncrementalSGD<T>*) new TrueSGD<T>(dim,lambda)
    : (IncrementalSGD<T>*) new ReservoirSGD<T>(dim, lambda, rsize);
  model m;
  
  if(input_model_file) {
    DEBUG_ONLY(cout << "reading model from " << input_model_file << endl;);
    std::ifstream f(input_model_file);
    f >> m;     
    sgd->setModel(m);
  } else {
    DEBUG_ONLY(cout << "using fresh model " << endl;);
  }
  
  
  DEBUG_ONLY(Timer train_time(true););
  int nExamples = 0, nTotal = 0; 
  apply_to_file<T>(train_name,sgd, (int (*)(IncrementalSGD<T>*, int, T))&train<T>, nExamples, nTotal);
  DEBUG_ONLY(cout << "Finished Training in " << train_time.stop() << "s on " << nExamples << " examples " << endl;);
  
  int nCorrect = 0;
  apply_to_file<T>(test_file, sgd, (int (*)(IncrementalSGD<T>*, int, T))&correct_label<T>, nExamples, nCorrect);
  //apply_to_file<T>(test_file, sgd, (int (*)(IncrementalSGD<T>*, int, T))&print_eps<T>, nExamples, nCorrect);
  cout << " correct " << nCorrect << " of " << nExamples << endl;
  
  if(out_file) {
    DEBUG_ONLY(cout << "writing output to " << out_file << endl;);
    std::ofstream f(out_file);
    model m = sgd->getModel();
    // SCALE THE MODEL.   
    f << m;
  }
}

static int bDense   = 0;
static int bPureSGD = 0;
static int rsize    = 100;
static int dim      = 100;


int 
main(int argc, char* argv[]) {
  int c = 0, option_index = 0;
  char *input_model = NULL, *output_file = NULL;
  double lambda = 1e-2;
  char* usage_str = " <train_file> <test_file>";
  opterr = 1;
  static struct extended_option long_options[] = {
    {"dense"    , no_argument, &bDense, 1, "if input contains dense vectors"},
    {"puresgd"  , no_argument, &bPureSGD, 1, "if want to use PureSGD (instead of reservoir)"},
    {"dim"      , required_argument, NULL, 'd', "the dimension of the model"},
    {"model"    , required_argument, NULL, 'm', "existing model input filename"},
    {"output"   , required_argument, NULL, 'o', "name of file to output model"},
    {"lambda"   , required_argument, NULL, 'l', "lambda (learning rate)"},
    {"reservoir", required_argument, NULL, 'r', "size of resevoir to be used by ResevoirSGD"},
    {NULL,0,NULL,0,0} 
  };
  
  // Somehow this parsing is not working as advertised.
  struct option* lo = convert_extended_options(long_options);
  while( (c = getopt_long(argc, argv, "r:d:m:o:l:", lo, &option_index)) != -1) 
    {
      switch (c) {      
      case 'r':
	rsize = atoi(optarg);
	break;
      case 'd':
	dim  = atoi(optarg);
	break;
      case 'm':
	input_model = optarg;
	break;
      case 'o':
	output_file = optarg;
	break;
      case 'l':
	lambda = atof(optarg);
	break;      
      case ':':
      case '?':
	print_usage(long_options, argv[0], usage_str);
      exit(-1);
      break;
      }
    }
  
  
  char* train_name, *test_file;
  if(optind == argc - 2) {
    train_name = argv[optind];
    test_file  = argv[optind+1];
  } else {
    print_usage(long_options, argv[0], usage_str);
    exit(-1);
  }
  
  // Conditions for sane input
  if(bPureSGD != 0 && rsize == 0) {
    cout << "ERROR: Reservoir size is 0 and ReservoirSGD is selected" << endl;
    print_usage(long_options, argv[0], usage_str);
    exit(-1);
  }
  
  if(bDense > 0) {
    train_and_test<FVector>(train_name, test_file, input_model, output_file, dim, lambda, bPureSGD > 0, rsize);
  } else {
    train_and_test<SVector>(train_name, test_file, input_model, output_file, dim, lambda, bPureSGD > 0, rsize);
  }
  free(lo);
}
