#ifndef HAZY_MODEL_H
#define HAZY_MODEL_H

#include "model.h"

#include <sstream>

/**
 * This model allows enapsulates the state of the SVM and the high, and low water.
 * It also contains some helper functions for the enum, strategy
 */
struct hazy_model {
  struct model &_model;
  enum strategy { LAZY_NAIVE, EAGER_NAIVE,LAZY_HAZY, EAGER_HAZY } _strategy;
  double low_water, high_water;
  /**
   * This is used by the LAZY algorithms to determine if they need
   * to reupdate/cache the model inside the database.
   * It is to support an optimization.
   * The protocol is that a model is in the db, until an update arrives, then we can
   * no longer assume that the model is inside the databse.
   * when we execute the prepared stmt (or equivalent) of model caching
   * we will set this value to be true.
   */
  bool model_is_in_db;
  /// Returns true if the model is in the database.
  bool test_and_set_model_in_db() { bool ret = model_is_in_db; model_is_in_db = true; return ret; }
  /// Invalidates model_is_in_db variable
  void invalidate_db_model() { model_is_in_db = false; }
  
  /// Helper accessor methods
  /**
   * Returns whether the strategy is eager or not
   */
  bool isEager()  { return (_strategy == EAGER_NAIVE || _strategy == EAGER_HAZY); }
  /**
   * Returns whether the strategy is lazy or not
   */
  bool isLazy ()  { return (_strategy == LAZY_NAIVE  || _strategy == LAZY_HAZY); }
  /**
   * Returns whether the strategy is hazy or not
   */
  bool isHazy()   {return (_strategy == EAGER_HAZY || _strategy == LAZY_HAZY);}
  /**
   * Returns whether the strategy is naive or not
   */
  bool isNaive()  { return (_strategy == EAGER_NAIVE || _strategy == LAZY_NAIVE); }
  
  /// Static methods
  /**
   * Returns whether given strategy is eager or not
   * \param[in] s strategy
   * \return s is eager or not
   */
  static bool isEager(strategy s) { return (s == EAGER_NAIVE || s == EAGER_HAZY); }
  /**
   * Returns whether given strategy is lazy or not
   * \param[in] s strategy
   * \return s is lazy or not
   */
  static bool isLazy(strategy s)  {return (s == LAZY_NAIVE || s == LAZY_HAZY);}
  /**
   * Returns whether given strategy is hazy or not
   * \param[in] s strategy
   * \return s is hazy or not
   */
  static bool isHazy(strategy s)  { return (s == EAGER_HAZY || s == LAZY_HAZY); } 
  /**
   * Returns whether given strategy is naive or not
   * \param[in] s strategy
   * \return s is naive or not
   */
  static bool isNaive(strategy s) {return (s == EAGER_NAIVE || s == LAZY_NAIVE);}
  
  /// Constructor
  /**
   * hazy_model constructor
   * \param[in] model sgd model
   * \param[in] s strategy being used
   */
hazy_model(struct model &model, strategy s) : _model(model), _strategy(s), 
    low_water(0.0), high_water(0.0), model_is_in_db(false) { }
};

/// This is an exception that is used in case statements: C++ doesn't check
/// whether the cases of a switch statement are exhaustive, and so
/// this exception should never be seen by a user.
class UnknownStrategyException : public std::exception {
  std::string _filename, _msg;
  int _line;
 public:
  UnknownStrategyException(const char* file, int line, const char *msg) {
    _filename = std::string(file);
    _line     = line;
    _msg      = std::string(msg);
  }
  virtual ~UnknownStrategyException() throw() { } 
  
  virtual const char* what() const throw() {
    std::ostringstream s;
    s << "[Unknown Strategy Exception @ " << _filename << ":" << _line << "] msg: " << _msg;
    return s.str().c_str();
  }
  
};

#endif
