/**
 * \file result_table.h
 */
#include <string>
#include "vectors.h"


///This class represents entities to be kept in main memory
template<class T>
class ResultTable
{
  
  struct resultTuple{
    int id;
    double eps;
    int truthValue;
    const T *featureArray;
  };
  int
    resultTupleComp(const void*, const void*);
 private:
  ///size of tuple
  int tupleSize;
  ///# of entities
  int entity_size;
  ///all entities
  resultTuple *table;
  ///packed feature vectors
  T *pFeatures;
  ///packed ids
  int *packedIds;
  ///set this to recluster in memory on each sort
  bool bCluster;
 public:
  /**
   * Constructor
   * \param[in] entitySize # of entities
   */
  ResultTable(int entitySize);
  //~ResultTable() {free(table);}
  /**
   * Returns id of tuple at given index
   * \param[in] index index of entity in entity table
   * \return id of tuple
   */
  int getId(int index);
  /**
   * Returns feature vector of tuple at given index
   * \param[in] index index of entity in entity table
   * \return feature vector of tuple
   */
  const T* getFeatureArray(int index);
  /**
   * Returns eps of tuple at given index
   * \param[in] index index of entity in entity table
   * \return epsilon of tuple
   */
  double getEps(int index);
  /**
   * Returns truth value of tuple at given index
   * \param[in] index index of entity in entity table
   * \return truth value of tuple
   */
  int getTruthValue(int index);
  /**
   * Returns pointer to the id of tuple at given index
   * \param[in] index index of entity in entity table
   * \return pointer to the id of tuple
   */
  int* getPointer(int index);
  /**
   * Sets eps value of tuple at given index
   * \param[in] index index of entity in entity table
   * \param[in] new_eps new eps value
   */
  void setEps(int index, double new_eps);
  /**
   * Sets feature vector component of tuple at given index
   * \param[in] index index of entity in entity table
   * \param[in] featureIndex index of feature vector 
   * \param[in] featureValue value of feature vector at given featureIndex
   */
  void setFeatureVectorComponent(int index, int featureIndex, VFloat featureValue);
  /**
   * Sets feature vector of tuple at given index from string in svm format
   * \param[in] index index of entity in entity table
   * \param[in] str feature vector in svm format
   */
  void setFeatureVectorFromStr(int index, std::string str);
  /**
   * Sets truth value of tuple at given index
   * \param[in] index index of entity in entity table
   * \param[in] new_truth_value new truth value
   */
  void setTruthValue(int index, int new_truth_value);
  /**
   * Sets id of tuple at given index
   * \param[in] index index of entity in entity table
   * \param[in] id new id
   */
  void setId(int index, int id);
  /**
   * sorts entities based on epsilons
   */
  void sort();
};
