/**
 * skiing.h
 */
#ifndef _SKIING_H_
#define _SKIING_H_ 1
#include <algorithm>
#include "globals.h"

/**
 * This class handles skiing strategy
 */

/**
 * Skiing spec structure that keeps initial resort time, tao value, m factor and ratio
 */
struct skiing_spec {
  double initial_resort, taoVal, m_factor, ratio;
};

class Skiing
{
 private:
  ///tao
  double tao;
  ///m factor of dataset
  double m_factor;
  ///accumulated cost
  double acc_cost;
  ///resort cost
  double resort_cost;
  ///alpha ratio
  double ratio;
  ///svm baseline update (updating external table right after sort)
  double svm_baseline_update;
  
  
  /**
   * Resets accumulated cost to 0
   */
  void resetAccCost() {acc_cost = 0;}
  /**
   * Updates resort cost with given resort cost
   * \param[in] res_cost new resort cost
   */
  void updateResortCost(double res_cost) {resort_cost = res_cost;}
  
  /**
   * Keeps whether the current iteration is right after resort or not
   */
  bool firstAfterResort; // This is set to true by resort.
  /**
   * Performs the baseline adjustment
   * If iteration is right after resort, then we set this cost as baseline and return it
   * Else, we return adjusted cost
   * \param[in] cost iteration cost
   * \return adjusted cost
   */
  double adjustCost(double cost) {
    if(firstAfterResort) {
      setSVMBaseline(cost);
      firstAfterResort  = false;
      return cost;
    } else {
      double svm_baseline_update = getSVMBaselineUpdate();
      double adjustedCost = std::max(cost - svm_baseline_update, 0.0);
      setSVMBaseline(std::min(cost, svm_baseline_update));
      return adjustedCost;
    }
  }
  
  /**
   * Initializes class members
   */
  void setup() {
    acc_cost = 0.0;
    svm_baseline_update = 0.0;
    firstAfterResort = false;
  }
 public:
  /**
   * Initializes skiing object with the given skiing spec
   * \param[in] s skiing spec
   */
  Skiing(struct skiing_spec &s) {
    this->tao = s.taoVal;
    this->m_factor = s.m_factor;
    this->resort_cost = s.initial_resort;
    this->ratio = s.ratio;
    setup();
    LOGGING_ONLY(std::cout << "[skiing.h] ski constructor" << std::endl;);
    LOGGING_ONLY(std::cout << "\ttao: " << tao << std::endl;);
    LOGGING_ONLY(std::cout << "\tm_factor: " << m_factor << std::endl;);
    LOGGING_ONLY(std::cout << "\tresort_cost: " << resort_cost << std::endl;);
    LOGGING_ONLY(std::cout << "\tratio: " << ratio << std::endl;);
    LOGGING_ONLY(std::cout << "\tacc_cost: " << acc_cost << std::endl;);
    LOGGING_ONLY(std::cout << "\tsvm_baseline_update: " << svm_baseline_update << std::endl;);
    LOGGING_ONLY(std::cout << "\tfirstAfterResort: " << firstAfterResort << std::endl;);
  }
  ///Constructor initializes variables
  /**
   * Initializes skiing object with the given skiing values
   * \param[in] tao_val tao value
   * \param[in] m_f m factor
   * \param[in] resort initial resort cost
   * \param[in] ratio alpha
   */
  Skiing(double tao_val, double m_f, double resort, double ratio) {
    this->tao = tao_val;
    this->m_factor = m_f;
    this->resort_cost = resort;
    this->ratio = ratio;
    setup();
  }
  ///Returns m factor
  double getMFactor() {return m_factor;}
  ///Returns accumulated cost value
  double getAccCost() {return acc_cost;}
  ///Returns resort cost value
  double getResortCost() {return resort_cost;}
  
  ///Updates accumulated cost with incremental update time
  void updateAccCost(double waste_time) {
    double adjustedCost = adjustCost(waste_time);
    acc_cost = ((1 - tao) * acc_cost) + adjustedCost;
    LOGGING_ONLY(std::cout << "[skiing.h] updateAccCost" << std::endl;);
    LOGGING_ONLY(std::cout << "\ttao: " << tao << std::endl;);
    LOGGING_ONLY(std::cout << "\tm_factor: " << m_factor << std::endl;);
    LOGGING_ONLY(std::cout << "\tresort_cost: " << resort_cost << std::endl;);
    LOGGING_ONLY(std::cout << "\tratio: " << ratio << std::endl;);
    LOGGING_ONLY(std::cout << "\tacc_cost: " << acc_cost << std::endl;);
    LOGGING_ONLY(std::cout << "\tsvm_baseline_update: " << svm_baseline_update << std::endl;);
    LOGGING_ONLY(std::cout << "\tfirstAfterResort: " << firstAfterResort << std::endl;);
  }
  
  ///Checks whether hazy should resort or not
  bool shouldResort() {return acc_cost > resort_cost * ratio;}
  
  /**
   * This method informs skiing that a resort has occurred.
   * \param[in] time_to_resort how long the resort has taken.
   */
  void doResort(double time_to_resort) {
    updateResortCost(time_to_resort);
    resetAccCost();
    firstAfterResort = true;
  }
  
  // These are only for debugging.
  ///Updates svm baseline time
  void setSVMBaseline(double baseline) {svm_baseline_update = baseline;}
  ///Returns svm baseline time
  double getSVMBaselineUpdate() {return svm_baseline_update;}
  
};
#endif
