#include "postgres.h"
#include "funcapi.h"
#include "executor/executor.h"  /* for GetAttributeByName() */
#include <string.h>
#include "utils/array.h"
#include "catalog/pg_type.h"
#include <pthread.h>
#include "executor/spi.h"       /* this is what you need to work with SPI */
#include "commands/trigger.h"   /* ... and triggers */

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/socket.h>
#include <netdb.h>

#include <assert.h>

#include "fmgr.h"
#include "vector_bin.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

const char *szDelim = "|"; // make sure this is the same!

PG_FUNCTION_INFO_V1(auxF);

/**
 * Parses input string and outputs word with tf score
 * \param[in] parsedPubText to_ts_vector output of the input string
 * \param[in] distinctWordCount # of distinct words in the input string
 * \param[in] pubLength string length
 */
Datum auxF(PG_FUNCTION_ARGS)
{
  bool isnull;
  FuncCallContext     *funcctx;
  int                  call_cntr;
  int                  max_calls;
  TupleDesc            tupdesc;
  AttInMetadata       *attinmeta;
  int i;
  
  text *parsedPubText = PG_GETARG_TEXT_P(0);
  int distinctWordCount = PG_GETARG_INT32(1);
  int pubLength = PG_GETARG_INT32(2);
  
  int strlength = VARSIZE(parsedPubText)-VARHDRSZ;
  char *parsedPubName = palloc(sizeof(char) * (strlength + 1));
  memcpy(parsedPubName, VARDATA(parsedPubText), strlength);
  parsedPubName[strlength] = '\0';
  
  char *tempstr = palloc((pubLength + 1) * sizeof(char));
  char **str_array = palloc((sizeof (char *)) * distinctWordCount);
  int *count_array = palloc((sizeof (int)) * distinctWordCount);
  
  int index = 0;
  int strCount = 0;
  
  for (i = 0; i < strlen(parsedPubName); i ++) {
    if(parsedPubName[i] != '\'' && parsedPubName[i] != ':') {
      tempstr[index] = parsedPubName[i];
      index ++;
    }
    else if (parsedPubName[i] == ':') {
      tempstr[index] = '\0';
      int length = (strlen(tempstr) + 1) * sizeof(char);
      str_array[strCount] = (char*) palloc(length);
      strcpy(str_array[strCount], tempstr);
      int localCount = 0;
      while (i < strlen(parsedPubName) && parsedPubName[i] != '\'') {
	if(parsedPubName[i] == ',')
	  localCount ++;
	i ++;
      }
      localCount ++;
      count_array[strCount] = localCount;
      strCount ++;
      tempstr = palloc((pubLength + 1) * sizeof(char));
      index = 0;
    }
  }
  
  if (distinctWordCount != strCount)
    elog(ERROR, "problem");
  
  int wordCount = 0;
  for (i = 0; i < distinctWordCount; i ++)
    wordCount += count_array[i];
  
  /* stuff done only on the first call of the function */
  if (SRF_IS_FIRSTCALL())
    {
      MemoryContext   oldcontext;
      
      /* create a function context for cross-call persistence */
      funcctx = SRF_FIRSTCALL_INIT();
      
      /* switch to memory context appropriate for multiple function calls */
      oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
      
      /* total number of tuples to be returned */
      max_calls = distinctWordCount;
      funcctx->max_calls = max_calls;
      
      /* Build a tuple descriptor for our result type */
      if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE)
	ereport(ERROR,
		(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
		 errmsg("function returning record called in context "
			"that cannot accept type record")));
      
      /*
       * generate attribute metadata needed later to produce tuples from raw
       * C strings
       */
      attinmeta = TupleDescGetAttInMetadata(tupdesc);
      funcctx->attinmeta = attinmeta;
      
      MemoryContextSwitchTo(oldcontext);
    }
  
  /* stuff done on every call of the function */
  funcctx = SRF_PERCALL_SETUP();
  
  call_cntr = funcctx->call_cntr;
  max_calls = funcctx->max_calls;
  attinmeta = funcctx->attinmeta;
  
  if (call_cntr < max_calls)    /* do when there is more left to send */
    {
      char       **values;
      HeapTuple    tuple;
      Datum        result;
      
      /*
       * Prepare a values array for building the returned tuple.
       * This should be an array of C strings which will
       * be processed later by the type input functions.
       */
      values = (char **) palloc(2 * sizeof(char *));
      values[0] = (char *) palloc((strlen(str_array[call_cntr]) + 1) * sizeof(char));
      values[1] = (char *) palloc(16 * sizeof(char));
      
      snprintf(values[0], strlen(str_array[call_cntr]) + 1, "%s", str_array[call_cntr]);
      snprintf(values[1], 16, "%f", ((float) count_array[call_cntr] / wordCount));
      
      /* build a tuple */
      tuple = BuildTupleFromCStrings(attinmeta, values);
      
      /* make the tuple into a datum */
      result = HeapTupleGetDatum(tuple);
      
      /* clean up (this is not really necessary) */
      pfree(values[0]);
      pfree(values[1]);
      pfree(values);
      
      SRF_RETURN_NEXT(funcctx, result);
    }
  else    /* do when there is no more left */
    {
      SRF_RETURN_DONE(funcctx);
    }
}

PG_FUNCTION_INFO_V1(dotprdct_sparse_cached_waste0);
PG_FUNCTION_INFO_V1(dotprdct_sparse_cached_without_waste0);
PG_FUNCTION_INFO_V1(cache_model_waste0);

static float8 model_array0[800000];
static int waste_count0 = 0;
static int not_waste_count0 = 0;
static float8 bias0 = 0;

/**
 * Initializes waste count
 */
Datum initialize_waste_count0(PG_FUNCTION_ARGS) {
  waste_count0 = 0;
  not_waste_count0 = 0;
}

/**
 * Calculates waste ratio
 * \return waste ratio
 */
Datum get_waste_ratio0(PG_FUNCTION_ARGS) {
  double denom = ((double) waste_count0 + not_waste_count0);
  double ratio = (double) waste_count0 / denom;
  if(waste_count0 == 0 && not_waste_count0 == 0)
    ratio = 0;
  
  PG_RETURN_FLOAT8(ratio);
}

/**
 * Caches the given model
 * \param[in] w model vector
 * \param[in] bias0
 */
Datum cache_model_waste0(PG_FUNCTION_ARGS) {
  Datum *i_data1;
  ArrayType *input1;
  Oid i_eltype1;
  int16 i_typlen1;
  int ndims1, *dims1, *lbs1;
  bool i_typbyval1;
  char i_typalign1;
  bool *nulls1;
  int n1;
  int index;
	
  input1 = PG_GETARG_ARRAYTYPE_P(0);
  bias0 = PG_GETARG_FLOAT8(1);
  
  
  i_eltype1 = ARR_ELEMTYPE(input1);
  get_typlenbyvalalign(i_eltype1, &i_typlen1, &i_typbyval1, &i_typalign1);
  
  ndims1 = ARR_NDIM(input1);
  dims1 = ARR_DIMS(input1);
  lbs1 = ARR_LBOUND(input1);
  
  deconstruct_array(input1, i_eltype1, i_typlen1, i_typbyval1, i_typalign1, &i_data1, &nulls1, &n1);
  
  /*model_array0 = (float8 *) palloc(sizeof(float8) * n1);*/
  
  for(index = 0; index < n1; index ++)
    model_array0[index] = DatumGetFloat8(i_data1[index]);
}

/**
 * Calculates the dot product with the cached model and given feature vector and updates waste count
 * \param[in] input feature vector of the entity
 */
Datum dotprdct_sparse_cached_waste0(PG_FUNCTION_ARGS)
{
  ArrayType *input2;
  Datum *i_data2;
  Oid i_eltype2;
  int16 i_typlen2;
  int ndims2, *dims2, *lbs2;
  bool i_typbyval2;
  char i_typalign2;
  bool *nulls2;
  int n2;
  int i, j;
  j = 0;
  
  input2 = PG_GETARG_ARRAYTYPE_P(0);
  
  i_eltype2 = ARR_ELEMTYPE(input2);
  
  get_typlenbyvalalign(i_eltype2, &i_typlen2, &i_typbyval2, &i_typalign2);
  
  ndims2 = ARR_NDIM(input2);
  dims2 = ARR_DIMS(input2);
  lbs2 = ARR_LBOUND(input2);
  
  deconstruct_array(input2, i_eltype2, i_typlen2, i_typbyval2, i_typalign2, &i_data2, &nulls2, &n2);
  
  float8 sum = 0;
  
  for (i = 0; i < n2; i ++) {
    Pair *p = (Pair *) DatumGetPointer(i_data2[i]);
    int indexOfItem = p->index;
    float8 value1 = model_array0[indexOfItem];
    float8 value2 = p->value;
    
    sum += (value1 * value2);
  }
  
  float8 prediction = sum + bias0;
  if(prediction <= 0)
    waste_count0 ++;
  else
    not_waste_count0 ++;
  
  PG_RETURN_FLOAT8(prediction);
}

/**
 * Calculates the dot product with the cached model and given feature vector and do not update waste count
 * \param[in] input feature vector of the entity
 */
Datum dotprdct_sparse_cached_without_waste0(PG_FUNCTION_ARGS)
{
  ArrayType *input2;
  Datum *i_data2;
  Oid i_eltype2;
  int16 i_typlen2;
  int ndims2, *dims2, *lbs2;
  bool i_typbyval2;
  char i_typalign2;
  bool *nulls2;
  int n2;
  int i, j;
  j = 0;
  
  input2 = PG_GETARG_ARRAYTYPE_P(0);
  
  i_eltype2 = ARR_ELEMTYPE(input2);
  
  get_typlenbyvalalign(i_eltype2, &i_typlen2, &i_typbyval2, &i_typalign2);
  
  ndims2 = ARR_NDIM(input2);
  dims2 = ARR_DIMS(input2);
  lbs2 = ARR_LBOUND(input2);
  
  deconstruct_array(input2, i_eltype2, i_typlen2, i_typbyval2, i_typalign2, &i_data2, &nulls2, &n2);
  
  float8 sum = 0;
  
  for (i = 0; i < n2; i ++) {
    Pair *p = (Pair *) DatumGetPointer(i_data2[i]);
    int indexOfItem = p->index;
    float8 value1 = model_array0[indexOfItem];
    float8 value2 = p->value;
    
    sum += (value1 * value2);
  }
  
  float8 prediction = sum + bias0;
  
  PG_RETURN_FLOAT8(prediction);
}

PG_FUNCTION_INFO_V1(dotprdct_dense_cached_waste0);
PG_FUNCTION_INFO_V1(dotprdct_dense_cached_without_waste0);

/**
 * Calculates the dot product with the cached model and given dense feature vector and updates waste count
 * \param[in] input feature vector of the entity
 */
Datum dotprdct_dense_cached_waste0(PG_FUNCTION_ARGS)
{   
  ArrayType *input2;
  Datum *i_data2;
  Oid i_eltype2;
  int16 i_typlen2;
  int ndims2, *dims2, *lbs2;
  bool i_typbyval2;
  char i_typalign2;
  bool *nulls2;
  int n2;
  int i, j;
  j = 0;
  
  input2 = PG_GETARG_ARRAYTYPE_P(0);
  
  i_eltype2 = ARR_ELEMTYPE(input2);
  
  get_typlenbyvalalign(i_eltype2, &i_typlen2, &i_typbyval2, &i_typalign2);
  
  ndims2 = ARR_NDIM(input2);
  dims2 = ARR_DIMS(input2);
  lbs2 = ARR_LBOUND(input2);
  
  deconstruct_array(input2, i_eltype2, i_typlen2, i_typbyval2, i_typalign2, &i_data2, &nulls2, &n2);
  
  float8 sum = 0;
  
  for (i = 0; i < n2; i ++)
    sum += model_array0[i] * ((float8) (DatumGetFloat8(i_data2[i])));
  
  float8 prediction = sum + bias0;
  if(prediction <= 0)
    waste_count0 ++;
  else
    not_waste_count0 ++;
  
  PG_RETURN_FLOAT8(prediction);
}

/**
 * Calculates the dot product with the cached model and given dense feature vector and do not update waste count
 * \param[in] input feature vector of the entity
 */
Datum dotprdct_dense_cached_without_waste0(PG_FUNCTION_ARGS)
{   
  ArrayType *input2;
  Datum *i_data2;
  Oid i_eltype2;
  int16 i_typlen2;
  int ndims2, *dims2, *lbs2;
  bool i_typbyval2;
  char i_typalign2;
  bool *nulls2;
  int n2;
  int i, j;
  j = 0;
  
  input2 = PG_GETARG_ARRAYTYPE_P(0);
  
  i_eltype2 = ARR_ELEMTYPE(input2);
  
  get_typlenbyvalalign(i_eltype2, &i_typlen2, &i_typbyval2, &i_typalign2);

  ndims2 = ARR_NDIM(input2);
  dims2 = ARR_DIMS(input2);
  lbs2 = ARR_LBOUND(input2);
  
  deconstruct_array(input2, i_eltype2, i_typlen2, i_typbyval2, i_typalign2, &i_data2, &nulls2, &n2);
  
  float8 sum = 0;
  
  for (i = 0; i < n2; i ++) 
    sum += model_array0[i] * ((float8) (DatumGetFloat8(i_data2[i])));
  
  float8 prediction = sum + bias0;
  
  PG_RETURN_FLOAT8(prediction);
}

