#include "hazy_message.h"
#include "low_level_net.h"
#include "vector_bin.h"
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include "postgres.h"
#include "funcapi.h"
#include "executor/executor.h"  /* for GetAttributeByName() */
#include <string.h>
#include "utils/array.h"
#include "catalog/pg_type.h"
#include "executor/spi.h"       /* this is what you need to work with SPI */
#include "commands/trigger.h"   /* ... and triggers */

#include "fmgr.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

/**
 * Methods for read queries are implemented
 */

#define BUFFER_SIZE 1000
#define DEFAULT_MESSAGE_SIZE 1000
char buf_in_functions[BUFFER_SIZE];

int view_id = 0;

/**
 * Writes the given msg to the server and retrieves integer from the received msg
 * \param[in] msg msg to be written to the server
 * \param[in] msg_size size of the msg
 * \param[out] label integer parsed from the server's incoming msg
 */
int write_func(message_data msg, int msg_size, int *label) {
  static struct addrinfo *res = NULL;
  static int sockfd = -1;
  static struct buffer_handle buf;
  buf = initialize_buffer();
  pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_lock( &mutex1 );
  
  static int buffer_initialized = 0;
  static struct buffer_handle update_trig_buf;
  if (buffer_initialized == 0) { update_trig_buf = initialize_buffer(); buffer_initialized = 1; }
  
  int i = 0, status =0 ;
  struct addrinfo hints;
  
  if(sockfd < 0) {
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags  = AI_NUMERICSERV;
    if( getaddrinfo("localhost", "10003", &hints, &res) != 0 ) {
      elog(INFO, "[all_frontend_sparse_on_disk_functions.c] Getaddrinfo fails\n");
      res = NULL;
      pthread_mutex_unlock( &mutex1 );
      return -1;
    }
    
    sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    
    if(sockfd < 0) {
      elog(INFO , "[all_frontend_sparse_on_disk_functions.c] msg error: %s \n", strerror(errno));
      sockfd = -1;
      pthread_mutex_unlock( &mutex1 );
      
      elog(INFO, "[all_frontend_sparse_on_disk_functions.c] Connect failed!");
      return -1;
    }
    
    int flag = 1;
    if(setsockopt( sockfd, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag) ) == -1) {
      elog(INFO, "Couldn't setsockopt(TCP_NODELAY)\n");
      destroy_handle(buf);
      pthread_mutex_unlock( &mutex1 );
      elog(ERROR,  "EXIT_FAILURE" );
    }
    
    if(connect(sockfd, res->ai_addr, res->ai_addrlen) < 0) {
      elog(INFO, "[all_frontend_sparse_on_disk_functions.c] msg error: %s \n", strerror(errno));
      sockfd = -1;
      pthread_mutex_unlock( &mutex1 );
      
      elog(INFO,"[all_frontend_sparse_on_disk_functions.c] Connect failed\n");
      return -1;
    }
    
  }
  
  if(write_wire_message(sockfd, msg, msg_size) != msg_size) {
    elog(INFO, "msg length: %d\n", msg_size);
    elog(INFO, "msg error: %s \n", strerror(errno));
    sockfd = -1;
    destroy_handle(buf);
    pthread_mutex_unlock( &mutex1 );
    
    elog(ERROR, "[all_frontend_sparse_on_disk_functions.c] Write failed!");
    return -1;
  }
  
  char *pData = read_wire_message(&buf, sockfd);
  if(pData == NULL) {
    elog(INFO, "msg error: %s\n", strerror(errno));
    sockfd = -1;
    pthread_mutex_unlock( &mutex1 );
    
    destroy_handle(buf);
    elog(ERROR, "[all_frontend_sparse_on_disk_functions.c] Read Failed!");
  }
  
  char *str1 = strtok(pData, "|");
  *label=atoi(str1);
  
  //LWLockRelease(AddinShmemInitLock);
  pthread_mutex_unlock( &mutex1 );
  
  return 0;
}



PG_FUNCTION_INFO_V1(getEntityLabel);

/**
 * Retrieves class of the entity
 * \param[in] id entity id
 * \return class of the entity
 */
Datum getEntityLabel(PG_FUNCTION_ARGS)
{
  int idToQuery = PG_GETARG_INT32(0);
  static pthread_mutex_t read_mutex = PTHREAD_MUTEX_INITIALIZER;
  static struct buffer_handle bh;
  static int    buffer_initialized = 0;
  pthread_mutex_lock( &read_mutex );
  if (buffer_initialized == 0) { bh = initialize_buffer(); buffer_initialized = 1; }
  
  update_buffer_size(&bh, DEFAULT_MESSAGE_SIZE);
  if(build_se_eager_msg(view_id, idToQuery, bh.pData, DEFAULT_MESSAGE_SIZE) < 0) {
    destroy_handle(bh);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "[getEntityLabel] Could not create the message (too big?)");	  
  }
  int label = -1;
  int retVal = write_func(bh.pData, strlen((char *) bh.pData), &label);
  if(retVal < 0) {
    destroy_handle(bh);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "label cannot be < 0 in getEntityLabel");
  }
  
  pthread_mutex_unlock( &read_mutex );
  PG_RETURN_INT32(label);
}

PG_FUNCTION_INFO_V1(getView);

/**
 * Retrieves number of entities that belong to the class 1
 * \return number of entities that belong to the class 1
 */
Datum getView(PG_FUNCTION_ARGS)
{
  static pthread_mutex_t read_mutex = PTHREAD_MUTEX_INITIALIZER;
  static struct buffer_handle bh;
  static int    buffer_initialized = 0;
  pthread_mutex_lock( &read_mutex );
  if (buffer_initialized == 0) { bh = initialize_buffer(); buffer_initialized = 1; }
  
  update_buffer_size(&bh, DEFAULT_MESSAGE_SIZE);
  
  if(build_count_msg(view_id, bh.pData, DEFAULT_MESSAGE_SIZE) < 0 ) {
    destroy_handle(bh);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "[getView] Could not create the message (too big?)");
  }
  int numOfPosEntities = -1;
  int retVal = write_func(bh.pData, strlen((char *) bh.pData), &numOfPosEntities);
  
  pthread_mutex_unlock( &read_mutex );
  PG_RETURN_INT32(numOfPosEntities);
}


PG_FUNCTION_INFO_V1(getServerId);

/**
 * Retrieves server id from the server
 * \return id of the server if server is up, -1 otherwise
 */
Datum getServerId(PG_FUNCTION_ARGS) {
  static pthread_mutex_t read_mutex = PTHREAD_MUTEX_INITIALIZER;
  static struct buffer_handle bh;
  static int    buffer_initialized = 0;
  pthread_mutex_lock( &read_mutex );
  if (buffer_initialized == 0) { bh = initialize_buffer(); buffer_initialized = 1; }
  
  update_buffer_size(&bh, DEFAULT_MESSAGE_SIZE);
  
  if(build_query_server_id_msg(bh.pData, DEFAULT_MESSAGE_SIZE) < 0) {
    destroy_handle(bh);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "[getServerId] Could not create the message (too big?)");
  }
  
  int server_id = -1;
  int retVal = write_func(bh.pData, strlen((char *) bh.pData), &server_id);
  
  pthread_mutex_unlock( &read_mutex );
  PG_RETURN_INT32(server_id);
}

PG_FUNCTION_INFO_V1(closeServerConn);

/**
 * Sends a close msg to the server
 * \return success
 */
Datum closeServerConn(PG_FUNCTION_ARGS) {
  static pthread_mutex_t read_mutex = PTHREAD_MUTEX_INITIALIZER;
  static struct buffer_handle bh;
  static int    buffer_initialized = 0;
  pthread_mutex_lock( &read_mutex );
  if (buffer_initialized == 0) { bh = initialize_buffer(); buffer_initialized = 1; }
  
  update_buffer_size(&bh, DEFAULT_MESSAGE_SIZE);
  
  if(build_close_connection_msg(bh.pData, DEFAULT_MESSAGE_SIZE) < 0) {
    destroy_handle(bh);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "[closeServerConn] Could not create the message (too big?)");
  }
  
  int close_conn = -1;
  int retVal = write_func(bh.pData, strlen((char *) bh.pData), &close_conn);
  
  pthread_mutex_unlock( &read_mutex );
  PG_RETURN_INT32(close_conn);
}

PG_FUNCTION_INFO_V1(getEntityLabelWithSQL_Text);

/**
 * Sends eps and feature vector of the entity as a text to the server and retrieves the class of the entity
 * \param[in] eps eps value of the entity
 * \param[in] w feature vector of the entity in text format
 * \return class of the entity
 */
Datum getEntityLabelWithSQL_Text(PG_FUNCTION_ARGS)
{
  float8 eps = PG_GETARG_FLOAT8(0);
  text *w = PG_GETARG_TEXT_P(1);
  
  int strlength = VARSIZE(w)-VARHDRSZ;
  char *w_str = (char *) palloc(sizeof(char) * (strlength + 1));
  memcpy(w_str, VARDATA(w), strlength);
  w_str[strlength] = '\0';
  
  int label = -1;
  
  static pthread_mutex_t read_mutex = PTHREAD_MUTEX_INITIALIZER;
  static struct buffer_handle bh;
  static int    buffer_initialized = 0;
  pthread_mutex_lock( &read_mutex );
  if (buffer_initialized == 0) { bh = initialize_buffer(); buffer_initialized = 1; }
  
  update_buffer_size(&bh, DEFAULT_MESSAGE_SIZE);
  
  if( build_se_lazy_hazy_msg(view_id, w_str, eps, bh.pData, (DEFAULT_MESSAGE_SIZE + strlength)) < 0 ){
    destroy_handle(bh);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "[getEntityLabelWithSQL] Could not create the message (too big?)");
  }
  
  int retVal = write_func(bh.pData, strlen((char *) bh.pData), &label);
  if(retVal < 0) {
    destroy_handle(bh);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "label cannot be < 0 in getEntityLabelWithSQL");
  }
  
  pfree(w_str);
  
  pthread_mutex_unlock( &read_mutex );
  PG_RETURN_INT32(label);
}

PG_FUNCTION_INFO_V1(getEntityLabelWithSQL_Sparse);

/**
 * Sends eps and feature vector of the entity as a binary data to the server and retrieves the class of the entity
 * This method is used for sparse entities
 * \param[in] eps eps value of the entity
 * \param[in] w feature vector of the entity in binary format
 * \return class of the entity
 */
Datum getEntityLabelWithSQL_Sparse(PG_FUNCTION_ARGS)
{
  float8 eps = PG_GETARG_FLOAT8(0);
  ArrayType *input = PG_GETARG_ARRAYTYPE_P(1);
  
  int n       = ARR_DIMS(input)[0];
  Pair *_data  = (Pair*) ARR_DATA_PTR(input);
  
  static pthread_mutex_t read_mutex = PTHREAD_MUTEX_INITIALIZER;
  static struct buffer_handle bh;
  static struct buffer_handle vec_buf;
  static int    buffer_initialized = 0;
  pthread_mutex_lock( &read_mutex );
  if (buffer_initialized == 0) { bh = initialize_buffer(); vec_buf = initialize_buffer(); buffer_initialized = 1; }
  
  int vec_size = calculate_svector_total_size(n);
  int data_size = vec_size + DEFAULT_MESSAGE_SIZE;
  
  update_buffer_size(&vec_buf, vec_size);
  update_buffer_size(&bh, data_size);	//?
  
  write_svector_bin_psql_pair(vec_buf.pData, vec_size, n, _data);
  
  size_t out_size = 0;
  if( build_se_lazy_hazy_msg_bin(view_id, eps, vec_buf.pData, vec_size, bh.pData, &out_size, data_size) < 0 ){ 
    destroy_handle(bh);
    destroy_handle(vec_buf);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "[getEntityLabelWithSQL] Could not create the message (too big?)");
  }
  
  int label = -1;
  int retVal = write_func(bh.pData, out_size, &label);
  if(retVal < 0) {
    destroy_handle(bh);
    destroy_handle(vec_buf);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "label cannot be < 0 in getEntityLabelWithSQL");
  }
  
  PG_RETURN_INT32(label);
}

PG_FUNCTION_INFO_V1(getEntityLabelWithSQL_Dense);

/**
 * Sends eps and feature vector of the entity as a binary data to the server and retrieves the class of the entity
 * This method is used for dense entities
 * \param[in] eps eps value of the entity
 * \param[in] w feature vector of the entity in binary format
 * \return class of the entity
 */
Datum getEntityLabelWithSQL_Dense(PG_FUNCTION_ARGS)
{
  float8 eps = PG_GETARG_FLOAT8(0);
  ArrayType *input = PG_GETARG_ARRAYTYPE_P(1);
  
  int n            = ARR_DIMS(input)[0];
  double *_data    = (double *) ARR_DATA_PTR(input);
  
  static pthread_mutex_t read_mutex = PTHREAD_MUTEX_INITIALIZER;
  static struct buffer_handle bh;
  static struct buffer_handle vec_buf;
  static int    buffer_initialized = 0;
  pthread_mutex_lock( &read_mutex );
  if (buffer_initialized == 0) { bh = initialize_buffer(); vec_buf = initialize_buffer(); buffer_initialized = 1; }
  
  int vec_size = calculate_fvector_total_size(n);
  int data_size = vec_size + DEFAULT_MESSAGE_SIZE;
  
  update_buffer_size(&vec_buf, vec_size);
  update_buffer_size(&bh, data_size);	//?
  
  write_fvector_bin_psql_double(vec_buf.pData, vec_size, n, _data);
  
  size_t out_size = 0;
  if( build_se_lazy_hazy_msg_bin(view_id, eps, vec_buf.pData, vec_size, bh.pData, &out_size, data_size) < 0 ){ 
    destroy_handle(bh);
    destroy_handle(vec_buf);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "[getEntityLabelWithSQL] Could not create the message (too big?)");
  }
  
  int label = -1;
  int retVal = write_func(bh.pData, out_size, &label);
  if(retVal < 0) {
    destroy_handle(bh);
    destroy_handle(vec_buf);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "label cannot be < 0 in getEntityLabelWithSQL");
  }
  
  PG_RETURN_INT32(label);
}


PG_FUNCTION_INFO_V1(getEntityLabelWithSQLNOEPS_Text);

/**
 * Sends feature vector of the entity as a text to the server and retrieves the class of the entity
 * \param[in] w feature vector of the entity in text format
 * \return class of the entity
 */
Datum getEntityLabelWithSQLNOEPS_Text(PG_FUNCTION_ARGS)
{
  text *w = PG_GETARG_TEXT_P(0);
  
  int strlength = VARSIZE(w)-VARHDRSZ;
  char *w_str = (char *) palloc(sizeof(char) * (strlength + 1));
  memcpy(w_str, VARDATA(w), strlength);
  w_str[strlength] = '\0';
  
  int label = -1;
  
  static pthread_mutex_t read_mutex = PTHREAD_MUTEX_INITIALIZER;
  static struct buffer_handle bh;
  static int    buffer_initialized = 0;
  pthread_mutex_lock( &read_mutex );
  if (buffer_initialized == 0) { bh = initialize_buffer(); buffer_initialized = 1; }
  
  update_buffer_size(&bh, DEFAULT_MESSAGE_SIZE + strlen(w_str));
  
  if(build_se_lazy_naive_msg(view_id, w_str, bh.pData, (DEFAULT_MESSAGE_SIZE + strlength)) < 0) {
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "[getEntityLabelWithSQLNOEPS] Could not create the message (too big?)");
  }
  int retVal = write_func(bh.pData, strlen((char *) bh.pData), &label);
  if(retVal < 0) {
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "label cannot be < 0 in getEntityLabelWithSQLNOEPS");
  }
  
  pfree(w_str);
  
  PG_RETURN_INT32(label);
}


PG_FUNCTION_INFO_V1(getEntityLabelWithSQLNOEPS_Dense);

/**
 * Sends feature vector of the entity as a binary data to the server and retrieves the class of the entity
 * This method is used for dense entities
 * \param[in] w feature vector of the entity in binary format
 * \return class of the entity
 */
Datum getEntityLabelWithSQLNOEPS_Dense(PG_FUNCTION_ARGS)
{ 
  ArrayType *input = PG_GETARG_ARRAYTYPE_P(0);
  
  int n         = ARR_DIMS(input)[0];
  double *_data = (double *) ARR_DATA_PTR(input);
  
  static pthread_mutex_t read_mutex = PTHREAD_MUTEX_INITIALIZER;
  static struct buffer_handle bh;
  static struct buffer_handle vec_buf;
  static int    buffer_initialized = 0;
  pthread_mutex_lock( &read_mutex );
  if (buffer_initialized == 0) { bh = initialize_buffer(); vec_buf = initialize_buffer(); buffer_initialized = 1; }
  
  int vec_size = calculate_fvector_total_size(n);
  int data_size = vec_size + DEFAULT_MESSAGE_SIZE;
  
  update_buffer_size(&vec_buf, vec_size);
  update_buffer_size(&bh, data_size); //?
  
  write_fvector_bin_psql_double(vec_buf.pData, vec_size, n, _data);
  
  size_t out_size = 0;
  if( build_se_lazy_naive_msg_bin(view_id, vec_buf.pData, vec_size, bh.pData, &out_size, data_size) < 0 ){
    destroy_handle(bh);
    destroy_handle(vec_buf);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "[getEntityLabelWithSQL] Could not create the message (too big?)");
  }
  
  int label = -1;
  int retVal = write_func(bh.pData, out_size, &label);
  if(retVal < 0) {
    destroy_handle(bh);
    destroy_handle(vec_buf);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "label cannot be < 0 in getEntityLabelWithSQL");
  }
  
  PG_RETURN_INT32(label);
}

PG_FUNCTION_INFO_V1(getEntityLabelWithSQLNOEPS_Sparse);

/**
 * Sends feature vector of the entity as a binary data to the server and retrieves the class of the entity
 * This method is used for sparse entities
 * \param[in] w feature vector of the entity in binary format
 * \return class of the entity
 */
Datum getEntityLabelWithSQLNOEPS_Sparse(PG_FUNCTION_ARGS)
{
  ArrayType *input = PG_GETARG_ARRAYTYPE_P(0);
  
  int n = ARR_DIMS(input)[0];
  Pair *_data = (Pair *) ARR_DATA_PTR(input);
  
  static pthread_mutex_t read_mutex = PTHREAD_MUTEX_INITIALIZER;
  static struct buffer_handle bh;
  static struct buffer_handle vec_buf;
  static int    buffer_initialized = 0;
  pthread_mutex_lock( &read_mutex );
  if (buffer_initialized == 0) { bh = initialize_buffer(); vec_buf = initialize_buffer(); buffer_initialized = 1; }
  
  int vec_size = calculate_svector_total_size(n);
  int data_size = vec_size + DEFAULT_MESSAGE_SIZE;
  
  update_buffer_size(&vec_buf, vec_size);
  update_buffer_size(&bh, data_size); //?
  
  write_svector_bin_psql_pair(vec_buf.pData, vec_size, n, _data);
  
  size_t out_size = 0;
  if( build_se_lazy_naive_msg_bin(view_id, vec_buf.pData, vec_size, bh.pData, &out_size, data_size) < 0 ){
    destroy_handle(bh);
    destroy_handle(vec_buf);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "[getEntityLabelWithSQL] Could not create the message (too big?)");
  }
  
  int label = -1;
  int retVal = write_func(bh.pData, out_size, &label);
  if(retVal < 0) {
    destroy_handle(bh);
    destroy_handle(vec_buf);
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "label cannot be < 0 in getEntityLabelWithSQL");
  }
  
  PG_RETURN_INT32(label);
}

PG_FUNCTION_INFO_V1(function_noop);

/**
 * Implements noop
 * \param[in] integer to simulate single entity read but without more operation
 * \return the input parameter
 */
Datum function_noop(PG_FUNCTION_ARGS)
{
  int input_arg = PG_GETARG_INT32(0);
  PG_RETURN_INT32(input_arg);
}

PG_FUNCTION_INFO_V1(function_noop_withserver);

/**
 * Implements noop with server connection
 * Writes the given input to the server and in server, noop occurs and then ack  is sent back to here
 * \param[in] integer to simulate single entity read but without more operation
 * \return ack
 */
Datum function_noop_withserver(PG_FUNCTION_ARGS)
{
  int input_arg = PG_GETARG_INT32(0);
  static pthread_mutex_t read_mutex = PTHREAD_MUTEX_INITIALIZER;
  static struct buffer_handle bh;
  static int    buffer_initialized = 0;
  pthread_mutex_lock( &read_mutex );
  if (buffer_initialized == 0) { bh = initialize_buffer(); buffer_initialized = 1; }
  
  update_buffer_size(&bh, DEFAULT_MESSAGE_SIZE);
  
  if(build_noop_server_msg(bh.pData, DEFAULT_MESSAGE_SIZE) < 0) {
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "[function_noop] Could not create the message (too big?)");
  }
  int ack = -1;
  int retVal = write_func(bh.pData, strlen((char *) bh.pData), &ack);
  if(retVal < 0) {
    pthread_mutex_unlock( &read_mutex );
    elog(ERROR, "label cannot be < 0 in getEntityLabel");
  }
  
  PG_RETURN_INT32(ack); 
}

