#include "postgres.h"
#include "funcapi.h"
#include "executor/executor.h"  /* for GetAttributeByName() */
#include <string.h>
#include "utils/array.h"
#include "catalog/pg_type.h"
#include <pthread.h>
#include "executor/spi.h"       /* this is what you need to work with SPI */
#include "commands/trigger.h"   /* ... and triggers */

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/socket.h>
#include <netdb.h>

#include <assert.h>

#include "fmgr.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

const char *szDelim = "|"; // make sure this is the same!

PG_FUNCTION_INFO_V1(auxF);

Datum auxF(PG_FUNCTION_ARGS)
{
  bool isnull;
  FuncCallContext     *funcctx;
  int                  call_cntr;
  int                  max_calls;
  TupleDesc            tupdesc;
  AttInMetadata       *attinmeta;
  int i;
  
  text *parsedPubText = PG_GETARG_TEXT_P(0);
  int distinctWordCount = PG_GETARG_INT32(1);
  int pubLength = PG_GETARG_INT32(2);
  
  int strlength = VARSIZE(parsedPubText)-VARHDRSZ;
  char *parsedPubName = palloc(sizeof(char) * (strlength + 1));
  memcpy(parsedPubName, VARDATA(parsedPubText), strlength);
  parsedPubName[strlength] = '\0';
  
  char *tempstr = palloc((pubLength + 1) * sizeof(char));
  char **str_array = palloc((sizeof (char *)) * distinctWordCount);
  int *count_array = palloc((sizeof (int)) * distinctWordCount);
  
  int index = 0;
  int strCount = 0;
  
  for (i = 0; i < strlen(parsedPubName); i ++) {
    if(parsedPubName[i] != '\'' && parsedPubName[i] != ':') {
      tempstr[index] = parsedPubName[i];
      index ++;
    }
    else if (parsedPubName[i] == ':') {
      tempstr[index] = '\0';
      int length = (strlen(tempstr) + 1) * sizeof(char);
      str_array[strCount] = (char*) palloc(length);
      strcpy(str_array[strCount], tempstr);
      int localCount = 0;
      while (i < strlen(parsedPubName) && parsedPubName[i] != '\'') {
	if(parsedPubName[i] == ',')
	  localCount ++;
	i ++;
      }
      localCount ++;
      count_array[strCount] = localCount;
      strCount ++;
      tempstr = palloc((pubLength + 1) * sizeof(char));
      index = 0;
    }
  }
  
  if (distinctWordCount != strCount)
    elog(ERROR, "problem");
  
  int wordCount = 0;
  for (i = 0; i < distinctWordCount; i ++)
    wordCount += count_array[i];
  
  /* stuff done only on the first call of the function */
  if (SRF_IS_FIRSTCALL())
    {
      MemoryContext   oldcontext;
      
      /* create a function context for cross-call persistence */
      funcctx = SRF_FIRSTCALL_INIT();
      
      /* switch to memory context appropriate for multiple function calls */
      oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
      
      /* total number of tuples to be returned */
      max_calls = distinctWordCount;
      funcctx->max_calls = max_calls;
      
      /* Build a tuple descriptor for our result type */
      if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE)
	ereport(ERROR,
		(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
		 errmsg("function returning record called in context "
			"that cannot accept type record")));
      
      /*
       * generate attribute metadata needed later to produce tuples from raw
       * C strings
       */
      attinmeta = TupleDescGetAttInMetadata(tupdesc);
      funcctx->attinmeta = attinmeta;
      
      MemoryContextSwitchTo(oldcontext);
    }
  
  /* stuff done on every call of the function */
  funcctx = SRF_PERCALL_SETUP();
  
  call_cntr = funcctx->call_cntr;
  max_calls = funcctx->max_calls;
  attinmeta = funcctx->attinmeta;
  
  if (call_cntr < max_calls)    /* do when there is more left to send */
    {
      char       **values;
      HeapTuple    tuple;
      Datum        result;
      
      /*
       * Prepare a values array for building the returned tuple.
       * This should be an array of C strings which will
       * be processed later by the type input functions.
       */
      values = (char **) palloc(2 * sizeof(char *));
      values[0] = (char *) palloc((strlen(str_array[call_cntr]) + 1) * sizeof(char));
      values[1] = (char *) palloc(16 * sizeof(char));
      
      snprintf(values[0], strlen(str_array[call_cntr]) + 1, "%s", str_array[call_cntr]);
      snprintf(values[1], 16, "%f", ((float) count_array[call_cntr] / wordCount));
      
      /* build a tuple */
      tuple = BuildTupleFromCStrings(attinmeta, values);
      
      /* make the tuple into a datum */
      result = HeapTupleGetDatum(tuple);
      
      /* clean up (this is not really necessary) */
      pfree(values[0]);
      pfree(values[1]);
      pfree(values);
      
      SRF_RETURN_NEXT(funcctx, result);
    }
  else    /* do when there is no more left */
    {
      SRF_RETURN_DONE(funcctx);
    }
}

PG_FUNCTION_INFO_V1(dotprdctint_cached_waste);
PG_FUNCTION_INFO_V1(dotprdctint_cached_without_waste);
PG_FUNCTION_INFO_V1(cache_model_waste);

///static float8 *model_array = NULL;
static float8 model_array[800000];
static int waste_count = 0;
static int not_waste_count = 0;
static float8 bias = 0;

Datum initialize_sparse_waste_count(PG_FUNCTION_ARGS) {
  waste_count = 0;
  not_waste_count = 0;
}

Datum get_sparse_waste_ratio(PG_FUNCTION_ARGS) {
  double ratio = (double) waste_count / ((double) waste_count + not_waste_count);
  if(waste_count == 0 && not_waste_count == 0)
    ratio = 0;
  
  PG_RETURN_FLOAT8(ratio);
}


Datum cache_model_waste(PG_FUNCTION_ARGS) {
  Datum *i_data1;
  ArrayType *input1;
  Oid i_eltype1;
  int16 i_typlen1;
  int ndims1, *dims1, *lbs1;
  bool i_typbyval1;
  char i_typalign1;
  bool *nulls1;
  int n1;
  int index;
  
  input1 = PG_GETARG_ARRAYTYPE_P(0);
  bias = PG_GETARG_FLOAT8(1);
  
  
  i_eltype1 = ARR_ELEMTYPE(input1);
  get_typlenbyvalalign(i_eltype1, &i_typlen1, &i_typbyval1, &i_typalign1);
  
  ndims1 = ARR_NDIM(input1);
  dims1 = ARR_DIMS(input1);
  lbs1 = ARR_LBOUND(input1);
  
  deconstruct_array(input1, i_eltype1, i_typlen1, i_typbyval1, i_typalign1, &i_data1, &nulls1, &n1);
  
  /*model_array = (float8 *) palloc(sizeof(float8) * n1);*/
  
  for(index = 0; index < n1; index ++)
    model_array[index] = DatumGetFloat8(i_data1[index]);
}

Datum dotprdctint_cached_waste(PG_FUNCTION_ARGS)
{
  ArrayType *input2, *input3;
  Datum *i_data2, *i_data3;
  Oid i_eltype2, i_eltype3;
  int16 i_typlen2, i_typlen3;
  int ndims2, ndims3, *dims2, *dims3, *lbs2, *lbs3;
  bool i_typbyval2, i_typbyval3;
  char i_typalign2, i_typalign3;
  bool *nulls2, *nulls3;
  int n2, n3;
  int i, j;
  j = 0;
  
  input2 = PG_GETARG_ARRAYTYPE_P(0);
  input3 = PG_GETARG_ARRAYTYPE_P(1);
  
  i_eltype2 = ARR_ELEMTYPE(input2);
  i_eltype3 = ARR_ELEMTYPE(input3);
  
  get_typlenbyvalalign(i_eltype2, &i_typlen2, &i_typbyval2, &i_typalign2);
  get_typlenbyvalalign(i_eltype3, &i_typlen3, &i_typbyval3, &i_typalign3);
  
  ndims2 = ARR_NDIM(input2);
  ndims3 = ARR_NDIM(input3);
  dims2 = ARR_DIMS(input2);
  dims3 = ARR_DIMS(input3);
  lbs2 = ARR_LBOUND(input2);
  lbs3 = ARR_LBOUND(input3);
  
  deconstruct_array(input2, i_eltype2, i_typlen2, i_typbyval2, i_typalign2, &i_data2, &nulls2, &n2);
  deconstruct_array(input3, i_eltype3, i_typlen3, i_typbyval3, i_typalign3, &i_data3, &nulls3, &n3);
  
  float8 sum = 0;
  
  for (i = 0; i < n3; i ++) {
    int indexOfItem = DatumGetInt32(i_data2[i]);
    float8 value1 = model_array[indexOfItem];
    float8 value2 = (float8) DatumGetFloat8(i_data3[i]);
    
    sum += (value1 * value2);
  }
  
  float8 prediction = sum + bias;
  if(prediction <= 0)
    waste_count ++;
  else
    not_waste_count ++;
  
  PG_RETURN_FLOAT8(prediction);
}

Datum dotprdctint_cached_without_waste(PG_FUNCTION_ARGS)
{
  ArrayType *input2, *input3;
  Datum *i_data2, *i_data3;
  Oid i_eltype2, i_eltype3;
  int16 i_typlen2, i_typlen3;
  int ndims2, ndims3, *dims2, *dims3, *lbs2, *lbs3;
  bool i_typbyval2, i_typbyval3;
  char i_typalign2, i_typalign3;
  bool *nulls2, *nulls3;
  int n2, n3;
  int i, j;
  j = 0;
  
  input2 = PG_GETARG_ARRAYTYPE_P(0);
  input3 = PG_GETARG_ARRAYTYPE_P(1);
  
  i_eltype2 = ARR_ELEMTYPE(input2);
  i_eltype3 = ARR_ELEMTYPE(input3);
  
  get_typlenbyvalalign(i_eltype2, &i_typlen2, &i_typbyval2, &i_typalign2);
  get_typlenbyvalalign(i_eltype3, &i_typlen3, &i_typbyval3, &i_typalign3);
  
  ndims2 = ARR_NDIM(input2);
  ndims3 = ARR_NDIM(input3);
  dims2 = ARR_DIMS(input2);
  dims3 = ARR_DIMS(input3);
  lbs2 = ARR_LBOUND(input2);
  lbs3 = ARR_LBOUND(input3);
  
  deconstruct_array(input2, i_eltype2, i_typlen2, i_typbyval2, i_typalign2, &i_data2, &nulls2, &n2);
  deconstruct_array(input3, i_eltype3, i_typlen3, i_typbyval3, i_typalign3, &i_data3, &nulls3, &n3);
  
  float8 sum = 0;
  
  for (i = 0; i < n3; i ++) {
    int indexOfItem = DatumGetInt32(i_data2[i]);
    float8 value1 = model_array[indexOfItem];
    float8 value2 = (float8) DatumGetFloat8(i_data3[i]);
    
    sum += (value1 * value2);
  }
  
  float8 prediction = sum + bias;
  
  PG_RETURN_FLOAT8(prediction);
}

#define BUFFER_SIZE 1000

pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
static struct addrinfo *res = NULL;
static int sockfd = -1;

int write_func(const char* strings[], int nStrings, int *label) {
  int i = 0, status =0 ;
  char buf[BUFFER_SIZE];
  struct addrinfo hints;
  
  if(nStrings < 1) {
    // we don't write anything
    return 0;
  }
  
  pthread_mutex_lock( &mutex1 );
  
  if(res == NULL) {
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags  = AI_NUMERICSERV;
    if( getaddrinfo("localhost", "10003", &hints, &res) != 0 ) {
      elog(INFO, "Getaddrinfo fails\n");
      res = NULL;
      pthread_mutex_unlock( &mutex1 );
      return -1;
    }
    
    sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    status = connect(sockfd, res->ai_addr, res->ai_addrlen);
    if(status < 0) {
      elog(INFO, "Connect failed\n");
      res = NULL;
      pthread_mutex_unlock( &mutex1 );
      return -1;
    }
  }
  strcpy(buf, "");
  for(i = 0; i < nStrings; i++) {
    strcat(buf, strings[i]);
    strcat(buf, szDelim);
  }
  strcat(buf, "\n"); // for readline
  
  write(sockfd, buf, strlen(buf));
  bzero(buf, BUFFER_SIZE);
  read(sockfd, buf, BUFFER_SIZE);
  char *str1 = strtok(buf, "|");
  if(strcmp(str1, "done") == 0) {
    str1 = strtok(NULL, "|");
    if(str1 != NULL)
      *label=atoi(str1);
  }
  
  //LWLockRelease(AddinShmemInitLock);
  pthread_mutex_unlock( &mutex1 );
  
  return 0;
}



PG_FUNCTION_INFO_V1(getEntityLabel);

Datum getEntityLabel(PG_FUNCTION_ARGS)
{
  int idToQuery = PG_GETARG_INT32(0);
  
  //code to retrieve label
  char *idStr = palloc(sizeof(char) * 1000);
  sprintf(idStr, "%d", idToQuery);
  
  const char* methodArg[] = {"0", "read", "single_ent", idStr};
  int label = -1;
  int retVal = write_func(methodArg, 4, &label);
  if(retVal < 0)
    elog(ERROR, "label cannot be < 0");
  
  pfree(idStr);
  
  PG_RETURN_INT32(label);
}


static int alg = 0;
static double elapsedTime = 0;
static TupleDesc tupdesc;
static SPITupleTable *tuptable;

int buildsocket() {
  struct addrinfo hints;
  if(res == NULL) {
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags  = AI_NUMERICSERV;
    if( getaddrinfo("localhost", "10003", &hints, &res) != 0 ) {
      elog(INFO, "Getaddrinfo fails\n");
      res = NULL;
      pthread_mutex_unlock( &mutex1 );
      return -1;
    }
    
    sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    int status = connect(sockfd, res->ai_addr, res->ai_addrlen);
    if(status < 0) {
      elog(INFO, "Connect failed\n");
      res = NULL;
      pthread_mutex_unlock( &mutex1 );
      return -1;
    }
  }
  
}

int write_func2(const char* strings[], int nStrings, int *numOfPosEntities) {
  int i = 0, status =0 ;
  int ret;
  char buf[BUFFER_SIZE];
  
  if(nStrings < 1) {
    // we don't write anything
    return 0;
  }
  
  pthread_mutex_lock( &mutex1 );
  
  int retVal = buildsocket();
  if(retVal == -1)
    return -1;
  
  strcpy(buf, "");
  for(i = 0; i < nStrings; i++) {
    strcat(buf, strings[i]);
    strcat(buf, szDelim);
  }
  strcat(buf, "\n"); // for readline
  
  ///int write_size = write(sockfd, buf, strlen(buf));
  int write_size = write(sockfd, buf, BUFFER_SIZE);
  if(write_size <= 0) {
    elog(ERROR, "error during write to the server: %s", strerror(errno));
  }
  
  bzero(buf, BUFFER_SIZE);
  int numOfBytesRecv = read(sockfd, buf, BUFFER_SIZE);
  if(numOfBytesRecv < 0) {
    elog(INFO, "error during read");
  }
  
  int countOfPos = atoi(buf);
  *numOfPosEntities = atoi(buf);
  
  //LWLockRelease(AddinShmemInitLock);
  pthread_mutex_unlock( &mutex1 );
  
  return 0;
}

PG_FUNCTION_INFO_V1(getView);

Datum getView(PG_FUNCTION_ARGS)
{
  //code to retrieve label
  const char* methodArg[] = {"0", "read", "all_ent"};
  int numOfPosEntities = -1;
  int retVal = write_func2(methodArg, 3, &numOfPosEntities);
  
  PG_RETURN_INT32(numOfPosEntities);
}

int write_func3(const char* strings[], int nStrings, int *label) {
  int i = 0, status =0 ;
  char buf[BUFFER_SIZE];
  struct addrinfo hints;
  
  if(nStrings < 1) {
    // we don't write anything
    return 0;
  }
  
  pthread_mutex_lock( &mutex1 );
  
  if(res == NULL) {
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags  = AI_NUMERICSERV;
    if( getaddrinfo("localhost", "10003", &hints, &res) != 0 ) {
      elog(INFO, "Getaddrinfo fails\n");
      res = NULL;
      pthread_mutex_unlock( &mutex1 );
      return -1;
    }
    
    sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    status = connect(sockfd, res->ai_addr, res->ai_addrlen);
    if(status < 0) {
      elog(INFO, "Connect failed\n");
      res = NULL;
      pthread_mutex_unlock( &mutex1 );
      return -1;
    }
  }
  strcpy(buf, "");
  for(i = 0; i < nStrings; i++) {
    strcat(buf, strings[i]);
    strcat(buf, szDelim);
  }
  strcat(buf, "\n"); // for readline
  
  write(sockfd, buf, strlen(buf));
  bzero(buf, BUFFER_SIZE);
  read(sockfd, buf, BUFFER_SIZE);
  char *str1 = strtok(buf, "|");
  if(strcmp(str1, "done") == 0) {
    str1 = strtok(NULL, "|");
    if(str1 != NULL)
      *label=atoi(str1);
  }
  
  //LWLockRelease(AddinShmemInitLock);
  pthread_mutex_unlock( &mutex1 );
  
  return 0;
}



PG_FUNCTION_INFO_V1(getEntityLabelWithSQL);

Datum getEntityLabelWithSQL(PG_FUNCTION_ARGS)
{
  float8 eps = PG_GETARG_FLOAT8(0);
  text *w = PG_GETARG_TEXT_P(1);
  text *w2 = PG_GETARG_TEXT_P(2);
  
  int strlength = VARSIZE(w)-VARHDRSZ;
  char *w_str = palloc(sizeof(char) * (strlength + 1));
  memcpy(w_str, VARDATA(w), strlength);
  w_str[strlength] = '\0';
  
  int strlength2 = VARSIZE(w2)-VARHDRSZ;
  char *w2_str = palloc(sizeof(char) * (strlength2 + 1));
  memcpy(w2_str, VARDATA(w2), strlength2);
  w2_str[strlength2] = '\0';
  
  //code to retrieve label
  char *epsStr = palloc(sizeof(char) * 1000);
  sprintf(epsStr, "%f", eps);
  
  const char* methodArg[] = {"0", "read", "single_ent", epsStr, w_str, w2_str};
  int label = -1;
  int retVal = write_func3(methodArg, 6, &label);
  if(retVal < 0)
    elog(ERROR, "label cannot be < 0");
  
  pfree(epsStr);
  pfree(w_str);
  pfree(w2_str);
  
  PG_RETURN_INT32(label);
}

int write_func4(const char* strings[], int nStrings, int *label) {
  int i = 0, status =0 ;
  char buf[BUFFER_SIZE];
  struct addrinfo hints;
  
  if(nStrings < 1) {
    // we don't write anything
    return 0;
  }
  
  pthread_mutex_lock( &mutex1 );
  
  if(res == NULL) {
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags  = AI_NUMERICSERV;
    if( getaddrinfo("localhost", "10003", &hints, &res) != 0 ) {
      elog(INFO, "Getaddrinfo fails\n");
      res = NULL;
      pthread_mutex_unlock( &mutex1 );
      return -1;
    }
    
    sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    status = connect(sockfd, res->ai_addr, res->ai_addrlen);
    if(status < 0) {
      elog(INFO, "Connect failed\n");
      res = NULL;
      pthread_mutex_unlock( &mutex1 );
      return -1;
    }
  }
  strcpy(buf, "");
  for(i = 0; i < nStrings; i++) {
    strcat(buf, strings[i]);
    strcat(buf, szDelim);
  }
  strcat(buf, "\n"); // for readline
  
  write(sockfd, buf, strlen(buf));
  bzero(buf, BUFFER_SIZE);
  read(sockfd, buf, BUFFER_SIZE);
  char *str1 = strtok(buf, "|");
  if(strcmp(str1, "done") == 0) {
    str1 = strtok(NULL, "|");
    if(str1 != NULL)
      *label=atoi(str1);
  }
  
  //LWLockRelease(AddinShmemInitLock);
  pthread_mutex_unlock( &mutex1 );
  
  return 0;
}



PG_FUNCTION_INFO_V1(getEntityLabelWithSQLNOEPS);

Datum getEntityLabelWithSQLNOEPS(PG_FUNCTION_ARGS)
{
  text *w = PG_GETARG_TEXT_P(0);
  text *w2 = PG_GETARG_TEXT_P(1);
  
  int strlength = VARSIZE(w)-VARHDRSZ;
  char *w_str = palloc(sizeof(char) * (strlength + 1));
  memcpy(w_str, VARDATA(w), strlength);
  w_str[strlength] = '\0';
  
  int strlength2 = VARSIZE(w2)-VARHDRSZ;
  char *w2_str = palloc(sizeof(char) * (strlength2 + 1));
  memcpy(w2_str, VARDATA(w2), strlength2);
  w2_str[strlength2] = '\0';
  
  const char* methodArg[] = {"0", "read", "single_ent", w_str, w2_str};
  int label = -1;
  int retVal = write_func4(methodArg, 5, &label);
  if(retVal < 0)
    elog(ERROR, "label cannot be < 0");
  
  pfree(w_str);
  pfree(w2_str);
  
  PG_RETURN_INT32(label);
}
