#include "hazy_message.h"
#include "low_level_net.h"
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include "postgres.h"
#include "funcapi.h"
#include "executor/executor.h"  /* for GetAttributeByName() */
#include <string.h>
#include "utils/array.h"
#include "catalog/pg_type.h"
#include "executor/spi.h"       /* this is what you need to work with SPI */
#include "commands/trigger.h"   /* ... and triggers */

#include "fmgr.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif


const int DEFAULT_STRING_SIZE=1000;

/**
 * Writes the given msg to the server and retrieves integer from the received msg
 * \param[in] msg msg to be written to the server
 */
int write_func_trig_noop_server(message_data msg) {
  static struct addrinfo *res = NULL;
  static int sockfd = -1;
  static pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_lock( &mutex1 );
  
  static int buffer_initialized = 0;
  static struct buffer_handle trig_noop_server_buf;
  if (buffer_initialized == 0) { trig_noop_server_buf = initialize_buffer(); buffer_initialized = 1; } 
  
  int i = 0, status =0 ;
  struct addrinfo hints;
  
  if(sockfd < 0) {
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags  = AI_NUMERICSERV;
    if( getaddrinfo("localhost", "10003", &hints, &res) != 0 ) {
      elog(INFO, "[trig_noop_server.c] Getaddrinfo fails\n");
      res = NULL;
      pthread_mutex_unlock( &mutex1 );
      return -1;
    }
    
    sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    
    if(sockfd < 0) {
      elog(INFO , "[trig_noop_server.c] msg error: %s \n", strerror(errno));
      sockfd = -1;
      pthread_mutex_unlock( &mutex1 );
      
      elog(INFO, "[trig_noop_server.c] Connect failed!");
      return -1;
    }
    
    int flag = 1;
    if(setsockopt( sockfd, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag) ) == -1) {
      elog(INFO, "Couldn't setsockopt(TCP_NODELAY)\n");
      elog(ERROR,  "EXIT_FAILURE" );
    }
    
    
    if(connect(sockfd, res->ai_addr, res->ai_addrlen) < 0) {
      elog(INFO, "[trig_noop_server.c] msg error: %s \n", strerror(errno));
      sockfd = -1;
      pthread_mutex_unlock( &mutex1 );
      
      elog(ERROR,"[trig_noop_server.c] Connect failed\n"); 
      return -1;
    }
  }
  
  if(write_wire_message(sockfd, msg, strlen(msg)) != strlen(msg)) {
    elog(INFO, "msg length: %d\n", strlen(msg));
    elog(INFO, "msg error: %s \n", strerror(errno));
    sockfd = -1;
    pthread_mutex_unlock( &mutex1 );
    
    elog(ERROR, "[trig_noop_server.c] Write failed!");
    return -1;
  }
  
  char*pData = read_wire_message(&trig_noop_server_buf, sockfd);
  
  if(pData == NULL) {
    elog(INFO, "msg error: %s\n", strerror(errno));
    sockfd = -1;
    pthread_mutex_unlock( &mutex1 );
    
    destroy_handle(trig_noop_server_buf);
    elog(ERROR, "[trig_noop_server.c] Read Failed!");
  }
  pthread_mutex_unlock( &mutex1 );
  
  return 0;	
}

extern Datum trigf(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(trig_noop_server);

/**
 * When new training example is inserted, then a noop msg is sent to the server. Server does not do anything, but send an ack back
 */
Datum
trig_noop_server(PG_FUNCTION_ARGS)
{
  HeapTuple   rettuple;
  TriggerData *trigdata = (TriggerData *) fcinfo->context;
  TupleDesc   tupdesc;
  bool        checknull = false;
  bool        isnull;
  int         ret, i;
  
  static pthread_mutex_t update_mutex = PTHREAD_MUTEX_INITIALIZER;
  static struct buffer_handle bh;
  static int    buffer_initialized = 0;
  pthread_mutex_lock( &update_mutex );
  if (buffer_initialized == 0) { bh = initialize_buffer(); buffer_initialized = 1; }
  
  
  // make sure it's called as a trigger at all
  if (!CALLED_AS_TRIGGER(fcinfo)) {
    pthread_mutex_unlock( &update_mutex );
    elog(ERROR, "trigf: not called by trigger manager");
  }
  
  if (TRIGGER_FIRED_BY_INSERT(trigdata->tg_event)) {
    rettuple = trigdata->tg_trigtuple;
    checknull = true;
    
    tupdesc = trigdata->tg_relation->rd_att;
    
    char* relname = SPI_getrelname(trigdata->tg_relation);
    int id =  atoi(SPI_getvalue(rettuple, tupdesc, 1));
    
    // connect to SPI manager
    if ((ret = SPI_connect()) < 0) {
      pthread_mutex_unlock( &update_mutex );
      elog(INFO, "trigf (fired): SPI_connect returned %d", ret);
    }
    
    int view_id = 0;
    update_buffer_size(&bh, DEFAULT_STRING_SIZE);
    if(build_noop_server_msg(bh.pData, DEFAULT_STRING_SIZE) < 0) {
      pthread_mutex_unlock( &update_mutex );
      elog(ERROR, "[trig_noop_server] Could not create the message (too big?)");
    }
    
    int retVal = write_func_trig_noop_server(bh.pData);
    pfree(relname);
    
    if (checknull) {   
      SPI_getbinval(rettuple, tupdesc, 1, &isnull);
      if (isnull)
	rettuple = NULL;
    }
    SPI_finish();
  }
  pthread_mutex_unlock( &update_mutex );
  return PointerGetDatum(rettuple);
}

