#include "vector_bin.h"
#include "hazy_message.h"
#include "low_level_net.h"
#include "errno.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <pthread.h>
#include "postgres.h"
#include "funcapi.h"
#include "executor/executor.h"  /* for GetAttributeByName() */
#include <string.h>
#include "utils/array.h"
#include "catalog/pg_type.h"
#include "executor/spi.h"       /* this is what you need to work with SPI */
#include "commands/trigger.h"   /* ... and triggers */

#include "fmgr.h"


#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif


const int DEFAULT_SIZE=1000;

/**
 * Writes the given msg to the server and retrieves integer from the received msg
 * \param[in] msg msg to be written to the server
 * \param[in] length size of the msg
 */
int write_func_update_trig(message_data msg, size_t length) {
  static struct addrinfo *addrinfo_res = NULL;
  static int sockfd = -1;
  static pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_lock( &mutex1 );
  
  static int buffer_initialized = 0;
  static struct buffer_handle update_trig_buf;
  if (buffer_initialized == 0) {
    update_trig_buf = initialize_buffer();
    buffer_initialized = 1;
  } 
  
  int i = 0, status =0 ;
  struct addrinfo hints;
  
  if(sockfd < 0) {
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags  = AI_NUMERICSERV;
    
    if( getaddrinfo("localhost", "10003", &hints, &addrinfo_res) != 0 ) {
      elog(INFO, "[update_trig.c] Getaddrinfo fails\n");
      addrinfo_res = NULL;
      destroy_handle(update_trig_buf);
      pthread_mutex_unlock( &mutex1 );
      return -1;
    }
    
    sockfd = socket(addrinfo_res->ai_family, 
		    addrinfo_res->ai_socktype, 
		    addrinfo_res->ai_protocol);
    
    if(sockfd < 0) {
      elog(INFO , "[update_trig.c] msg error: %s \n", strerror(errno));
      sockfd = -1;
      destroy_handle(update_trig_buf);
      pthread_mutex_unlock( &mutex1 );
      
      elog(INFO, "[update_trig.c] Connect failed!");
      return -1;
    }
    
    int flag = 1;
    if(setsockopt( sockfd, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag) ) == -1) {
      elog(INFO, "Couldn't setsockopt(TCP_NODELAY)\n");
      destroy_handle(update_trig_buf);
      pthread_mutex_unlock( &mutex1 );
      elog(ERROR,  "EXIT_FAILURE" );
    }
    
    if(connect(sockfd, 
	       addrinfo_res->ai_addr, 
	       addrinfo_res->ai_addrlen) < 0) {
      elog(INFO, "[update_trig.c] msg error: %s \n", strerror(errno));
      sockfd = -1;
      pthread_mutex_unlock( &mutex1 );
      
      elog(ERROR,"[update_trig.c] Connect failed\n"); 
      return -1;
    }
  }
  
  if(write_wire_message(sockfd, msg, length) != length) {
    elog(INFO, "msg length: %d\n", length);
    elog(INFO, "msg error: %s \n", strerror(errno));
    sockfd = -1;
    destroy_handle(update_trig_buf);
    pthread_mutex_unlock( &mutex1 );
    
    elog(ERROR, "[update_trig.c] Write failed!");
    return -1;
  }	
  
  char*pData = read_wire_message(&update_trig_buf,
				 sockfd);
  
  if(pData == NULL) {
    elog(INFO, "msg error: %s\n", strerror(errno));
    sockfd = -1;
    destroy_handle(update_trig_buf);
    pthread_mutex_unlock( &mutex1 );
    
    elog(ERROR, "[update_trig.c] Read Failed!");
  }
  
  pthread_mutex_unlock( &mutex1 );
  
  return 0;
}



extern Datum trigf(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(update_trig);

/**
 * Sends update data to the server as text data
 * \param[in] training_example training example in the svm text format
 */
Datum
update_trig(PG_FUNCTION_ARGS)
{
  HeapTuple   rettuple;
  TriggerData *trigdata = (TriggerData *) fcinfo->context;
  TupleDesc   tupdesc;
  bool        checknull = false;
  bool        isnull;
  int         ret, i;
  static pthread_mutex_t update_mutex = PTHREAD_MUTEX_INITIALIZER;
  static struct buffer_handle bh;
  static int    buffer_initialized = 0;
  pthread_mutex_lock( &update_mutex );
  if (buffer_initialized == 0) {
    bh = initialize_buffer();
    buffer_initialized = 1;
  } 
  
  // make sure it's called as a trigger at all
  if (!CALLED_AS_TRIGGER(fcinfo)) {
    destroy_handle(bh);
    pthread_mutex_unlock( &update_mutex );
    elog(ERROR, "trigf: not called by trigger manager");
  }
  
  if (TRIGGER_FIRED_BY_INSERT(trigdata->tg_event)) {
    rettuple = trigdata->tg_trigtuple;
    checknull = true;
    
    tupdesc = trigdata->tg_relation->rd_att;
    
    char* relname = SPI_getrelname(trigdata->tg_relation);
    int id =  atoi(SPI_getvalue(rettuple, tupdesc, 1));
    
    char *newExample;
    
    // connect to SPI manager
    if ((ret = SPI_connect()) < 0)
      elog(INFO, "trigf (fired): SPI_connect returned %d", ret);
    
    newExample = SPI_getvalue(rettuple, tupdesc, 2);
    
    int view_id = 0;
    int msg_size = strlen(newExample) + DEFAULT_SIZE;
    update_buffer_size(&bh, msg_size);
    
    if(build_update_msg(view_id, newExample, bh.pData, msg_size) < 0) {
      destroy_handle(bh);
      pthread_mutex_unlock( &update_mutex );
      elog(ERROR, "[update_trig] Could not create the message (too big?)");
    }
    
    int retVal = write_func_update_trig(bh.pData, 
					strlen((char*) bh.pData));
    
    pfree(newExample);
    pfree(relname);
    
    if (checknull) {   
      SPI_getbinval(rettuple, tupdesc, 1, &isnull);
      if (isnull)
	rettuple = NULL;
    }
    SPI_finish();
  }
  
  pthread_mutex_unlock( &update_mutex );
  return PointerGetDatum(rettuple);
}


PG_FUNCTION_INFO_V1(update_trig_bin);

/**
 * Sends update data to the server as a binary
 * \param[in] class class of the training example
 * \param[in] vector feature vector of the training example
 */
Datum
update_trig_bin(PG_FUNCTION_ARGS)
{
  HeapTuple   rettuple;
  TriggerData *trigdata = (TriggerData *) fcinfo->context;
  TupleDesc   tupdesc;
  bool        checknull = false;
  bool        isnull;
  int         ret, i;
  static pthread_mutex_t update_mutex = PTHREAD_MUTEX_INITIALIZER;
  static struct buffer_handle bh;
  static struct buffer_handle vec_buf;
  static int    buffer_initialized = 0;
  pthread_mutex_lock( &update_mutex );
  if (buffer_initialized == 0) {
    bh = initialize_buffer();
    vec_buf = initialize_buffer();
    buffer_initialized = 1;
  } 
  
  // make sure it's called as a trigger at all
  if (!CALLED_AS_TRIGGER(fcinfo)) {
    destroy_handle(bh);
    
    pthread_mutex_unlock( &update_mutex );
    elog(ERROR, "trigf: not called by trigger manager");
  }
  
  if (TRIGGER_FIRED_BY_INSERT(trigdata->tg_event)) {
    rettuple = trigdata->tg_trigtuple;
    checknull = true;
    
    tupdesc = trigdata->tg_relation->rd_att;
    
    char* relname = SPI_getrelname(trigdata->tg_relation);
    int id        =  atoi(SPI_getvalue(rettuple, tupdesc, 1));
    int label     =  atoi(SPI_getvalue(rettuple, tupdesc, 2));
    
    // connect to SPI manager
    if ((ret = SPI_connect()) < 0)
      elog(INFO, "trigf (fired): SPI_connect returned %d", ret);
    
    bool isnull = false;
    Datum newExample = SPI_getbinval(rettuple, tupdesc, 3, &isnull);
    if(isnull) {
      pthread_mutex_unlock( &update_mutex );
      elog(ERROR, "Feature vector is a null array");
    }
    
    ArrayType *a     = DatumGetArrayTypeP(newExample);
    int n_size       = ARR_DIMS(a)[0];
    char* t_name = SPI_gettype(tupdesc, 3);
    
    int vec_size = 0;
    int data_size = 0;
    
    // double array.
    if(strcmp(t_name, "_float8") == 0) {
      double* _data = (double*) ARR_DATA_PTR(a);
      
      vec_size = calculate_fvector_total_size(n_size);
      data_size = vec_size + DEFAULT_SIZE;
      update_buffer_size(&vec_buf, vec_size);
      update_buffer_size(&bh, data_size);
      
      write_fvector_bin_psql_double(vec_buf.pData, vec_size, n_size, _data);
    }
    else if(strcmp(t_name, "_pair") == 0) {
      Pair* _data = (Pair*) ARR_DATA_PTR(a);
      
      vec_size = calculate_svector_total_size(n_size);
      data_size = vec_size + DEFAULT_SIZE;
      update_buffer_size(&vec_buf, vec_size);
      update_buffer_size(&bh, data_size);
      
      write_svector_bin_psql_pair(vec_buf.pData, vec_size, n_size, _data);
    }
    
    size_t out_size = 0;
    
    int view_id = 0; // curious... set this to a real value??
    if( build_update_msg_bin(view_id, label, vec_buf.pData, 
			     vec_size, bh.pData, &out_size, data_size) < 0 ){ 
      destroy_handle(bh);
      destroy_handle(vec_buf);
      pthread_mutex_unlock( &update_mutex );
      elog(ERROR, "[update_trig_bin] Could not create the message (too big?)");
    }
    
    int retVal = write_func_update_trig(bh.pData, out_size);
    if(retVal < 0) {
      destroy_handle(bh);
      destroy_handle(vec_buf);
      pthread_mutex_unlock( &update_mutex );
      elog(ERROR, "label cannot be < 0 in update_trig_bin");
    }
    
    pfree(t_name);
    pfree(relname);
    
    if (checknull) {  
      SPI_getbinval(rettuple, tupdesc, 1, &isnull);
      if (isnull)
	rettuple = NULL;
    }
    
    SPI_finish();
  }

  else if(TRIGGER_FIRED_BY_DELETE(trigdata->tg_event)) {
    // connect to SPI manager
    if ((ret = SPI_connect()) < 0)
      elog(INFO, "trigf (fired): SPI_connect returned %d", ret);
    
    rettuple = trigdata->tg_trigtuple;
    checknull = true;
    
    tupdesc = trigdata->tg_relation->rd_att;
    
    char db_name[] = "VLDB11_Exp";

    char *relname = SPI_getrelname(trigdata->tg_relation);
    int entity_id = atoi(SPI_getvalue(rettuple, tupdesc, 1));
    
    int view_id = 0;
    int msg_size = strlen(db_name) + strlen(relname) + DEFAULT_SIZE;
    update_buffer_size(&bh, msg_size);
    
    size_t out_size = 0;

    if(build_delete_msg_bin(view_id, entity_id, db_name, strlen(db_name), relname, strlen(relname), bh.pData, &out_size, msg_size) < 0) {
      destroy_handle(bh);
      pthread_mutex_unlock( &update_mutex );
      elog(ERROR, "[update_trig] Could not create the message (too big?)");
    }

    int retVal = write_func_update_trig(bh.pData, out_size);

    pfree(relname);
    
    if (checknull) {   
      SPI_getbinval(rettuple, tupdesc, 1, &isnull);
      if (isnull)
	rettuple = NULL;
    }
    SPI_finish();

  }
  
  pthread_mutex_unlock( &update_mutex );
  return PointerGetDatum(rettuple);
}
