#include "hazy_sgd.h"
/**
 * This class implements multi-class classification by using decision trees
 */

// NB:
// the updates use labels in {-1,1} but 
// we return labels in {0,1}.. this causes problems. You'd think we would have fixed that by now.
/**
 * This is a generic Hazy_Node class that is extended by Internal or Leaf Hazy_Nodes
 */
template <class T>
class Hazy_Node {   
 protected:
  ///List of classifiers
  Hazy_Sgd<T>* _classifier;  
 public:  
  /**
   * Retrieves class of the entity
   * \param[in] entity_id id of the entity
   * \param[out] sClass class of the entity
   */
  virtual void single_entity_read(int entity_id, int &sClass) = 0;
  /**
   * Inserts given training examples to the tree and model is updated with this example
   * \param[in] label class of the training example
   * \param[in] vec feature vector of the given example
   */
  virtual void insert_example(int label, T& vec) = 0;
  /**
   * Destructor
   */
  virtual ~Hazy_Node() { //delete _classifier; 
  }
};

/**
 * This class extends Hazy_Node and represents Internal node in the tree
 */
template <class T>
class MultiClass_Internal_Node : public Hazy_Node<T> {
  Hazy_Node<T> *l, *r;
  int search_key;
  /**
   * Considering the label, decisdes which sub-tree to search
   * \param[in] label class of the entity
   * \return 0 for left sub-tree, 1 for right sub-tree
   */
  int find_direction(int label) {return (label <= search_key) ? 0 : 1;}
 public:
  /**
   * Constructor that initializes MultiClass_Internal_Node
   * \param[in] search_key
   * \param[in] l left sub-tree
   * \param[in] r right sub-tree
   * \param[in] s list of classifiers
   */
 MultiClass_Internal_Node(int search_key, Hazy_Node<T> *l, Hazy_Node<T> *r, Hazy_Sgd<T>* s) :
  l(l), r(r), search_key(search_key) {
    this->_classifier = s;
    
  }
  /**
   * Retrieves class of the entity
   * \param[in] entity_id id of the entity
   * \param[out] sClass class of the entity
   */
  void single_entity_read(int entity_id, int &sClass) {
    int this_class = 0;
    this->_classifier->readEntityClass(entity_id, this_class);
    if (this_class == 0) {
      l->single_entity_read(entity_id, sClass);
    } else {
      r->single_entity_read(entity_id, sClass);
    }
  }
  /**
   * Inserts given training examples to the tree and model is updated with this example
   * \param[in] label class of the training example
   * \param[in] vec feature vector of the given example
   */
  void insert_example(int label, T& vec) {
    int direction = find_direction(label);
    LOGGING_ONLY(Timer total_time(true););
    this->_classifier->updateModel(vec, (direction == 0) ? -1 : 1);
    LOGGING_ONLY(std::cout << "\tUpdating with " << label << " sk=" << search_key << " train time=" << total_time.stop() << " direction=" << direction << std::endl;)
      if(direction == 0) {
	l->insert_example(label, vec);
      } else {
	r->insert_example(label, vec);
      }
    LOGGING_ONLY(std::cout << "\t[Multiclass] Insert to children for " << search_key << " took " << total_time.stop() << std::endl;);
  }
};

/**
 * This class extends Hazy_Node and represents Leaf node in the tree
 */
template
<class T>
class MultiClass_Leaf_Node : public Hazy_Node<T> {
  int global_label, local_label;
  
 public:
  /**
   * Constructor that initializes MultiClass_Leaf_Node
   * \param[in] global_label represents the actual class that comes from training example
   * \param[in] local_label represents the label that leaf node keeps: either -1 or 1
   * \param[in] s list of classifiers
   */
  MultiClass_Leaf_Node(int global_label, int local_label, Hazy_Sgd<T>* s) 
    { this->global_label = global_label; this->local_label = local_label; this->_classifier = s; }
  /**
   * Retrieves class of the entity
   * \param[in] entity_id id of the entity
   * \param[out] sClass class of the entity
   */
  void single_entity_read(int entity_id, int &sClass) {sClass = global_label;}
  /**
   * Inserts given training examples to the tree and model is updated with this example
   * \param[in] label class of the training example
   * \param[in] vec feature vector of the given example
   */
  void insert_example(int label, T& vec) { this->_classifier->updateModel(vec, (local_label == 0) ? -1 : 1); } 
};

// models are indexed by search key which we assume are dense.
/**
 * Constructs the tree
 * \param[in] models list of hazy sgd classifiers
 * \param[in] low_value
 * \param[in] high_value
 * \param[in] parents_search_key
 * \param[in] parents_label_of_us
 */
template 
<class T>
Hazy_Node<T>*
construct_tree(Hazy_Sgd<T> **models, int low_value, int high_value, int parents_search_key, int parents_label_of_us) {
  assert(low_value <= high_value);
  if(low_value == high_value) {
    // here we are leaf
    return new MultiClass_Leaf_Node<T>(low_value, parents_label_of_us, models[parents_search_key]);
  } else {
    // build our children
    int mid_point = (high_value + low_value)/2;
    int our_search_key = mid_point;
    VERBOSE_ONLY(std::cout << "\t our_search_key = " << our_search_key << " [" << low_value << ", " << high_value << "] " << std::endl;);
    VERBOSE_ONLY(std::cout << "\t\tmidpoint = " << mid_point << std::endl;);
    VERBOSE_ONLY(std::cout << "\t\tparent=" << parents_search_key << " parents opinion=" << parents_label_of_us << std::endl;);
    Hazy_Node<T> *l = construct_tree<T>(models, low_value  , mid_point , our_search_key, 0);
    Hazy_Node<T> *r = construct_tree<T>(models, mid_point+1, high_value, our_search_key, 1);
    return new MultiClass_Internal_Node<T>(our_search_key, l, r, models[our_search_key]);
  }
}

/**
 * Builds tree by using balanced-tree
 * \param[in] models list of hazy sgd classifiers
 * \param[in] nLabels number of labels
 */
template
<class T>
Hazy_Node<T> *
build_root_node(Hazy_Sgd<T> **models, int nLabels) {
  int parent_id = 1;
  // find the next highest power of two.
  while(parent_id < nLabels) { parent_id = parent_id << 1; }
  return construct_tree<T>(models, 0, nLabels - 1, parent_id, 0);
}
// constructr_tree(models, 0, 7, 8, 0);


/**
 * Builds tree by using one-versus all tree
 * \param[in] models list of hazy sgd classifiers
 * \param[in] nLabels number of labels
 */
template
<class T>
Hazy_Node<T> *
build_ova(Hazy_Sgd<T> **models, int nLabels) {
  DEBUG_ONLY(std::cout << "Building one-versus-all tree" << std::endl;);
  Hazy_Node<T>** leaf = new Hazy_Node<T>*[nLabels];
  for(int i = 0; i < nLabels - 1; i++) {
    leaf[i]   = new MultiClass_Leaf_Node<T>(i, 0, models[i]);  
  }
  leaf[nLabels-1] = new MultiClass_Leaf_Node<T>(nLabels-1, 1, models[nLabels-2]); 
  Hazy_Node<T> *last_node = leaf[nLabels - 1];
  for(int i = nLabels - 2; i >= 0 ; i--) {
    Hazy_Node<T> *this_node = new MultiClass_Internal_Node<T>(i, leaf[i], last_node, models[i]);
    last_node = this_node;
  }
  
  DEBUG_ONLY(std::cout << "Finished building one-versus-all tree" << std::endl;);
  return last_node;
}
