#ifndef HYBRID_STORAGE_MANAGER_H
#define HYBRID_STORAGE_MANAGER_H

#include "storage_manager.h"
#include "ondisk_storage_manager.h"
#include "hybrid_storage_manager.h"
#include "incremental_sgd.h"
#include "postgres_util.h"
#include "lvector.h"
#include "timer.h"
#include "vectors.h"
#include <vector>
#include <map>


/**
 * Hybrid_Storage_Manager extends Storage_Manager and is used for hybrid architecture
 */
template<class T>
class Hybrid_Storage_Manager : Storage_Manager<T> {
  ///Ondisk_Storage_Manager is used
  Ondisk_Storage_Manager<T> *_disk;
  ///Database connection
  Hazy_Database *_conn;

  ///binary count queries
  std::ostringstream search_disk_region_one_lazy_query;
  std::ostringstream search_disk_region_one_eager_query;
  std::ostringstream search_disk_region_three_lazy_query;
  std::ostringstream search_disk_region_three_eager_query;

  typedef std::vector<struct _full_entity_entry<T> > _tbuffer;
  typedef typename std::vector<struct _full_entity_entry<T> >::iterator _tbuffer_it;
  _tbuffer _buffer;
  
  typedef std::map<key, int> _tentitymap;
  _tentitymap _entity_map; // the integer here is a "pointer" into _buffer. This in variant is maintained during resorts.
  
  // WARNING: May do not assume that this map exists.
  bool _epsused;// set to false to remove it.
  typedef std::map<key, double> _tepsmap;
  
  typedef typename std::map<key,double>::iterator _tepsmap_iterator;
  _tepsmap _epsmap;
  // it's parsing function
  static int parse_eps_map(PGresult *res, int index, key &k, double &eps) {
    k = atoi(PQgetvalue(res, index, 0));
    eps = atof(PQgetvalue(res, index, 1));
    return 0;
  }
  
  ///buffer size
  int _buffer_size;
  ///low & high eps
  double eps_high, eps_low;
  ///low & high water iterators
  _tbuffer_it low_water_it, high_water_it;
  
  bool _lazy;
  /**
   * Returns whether approach is lazy or not
   * \return approach is lazy
   */
  bool isLazy() { return _lazy; }
  /**
   * Returns whether approach is eager or not
   * \return approach is eager
   */
  bool isEager() { return !isLazy(); }

  ///unique id for prepared statements in order to prevent any conflict
  ///in the same database connection, there cannot be prepared statements with exact same name
  int unique_id_for_ps;
  
  ///type definition for string
  typedef std::ostringstream prepared_stmt;
  ///prepared statements
  prepared_stmt query_buffer_fill, query_eps_exec;
  
  /**
   * Calculates space used by the eps map
   * \param[out] spaceUsed space used by the eps map
   */
  void calculateSpaceUsedByEpsMap(double &spaceUsed);
  /**
   * Calculates space used by the buffer
   * \param[out] spaceUsed space used by the buffer
   */
  void calculateSpaceUsedByBuffer(double &spaceUsed);
  /**
   * Calculates space used by entity map
   * \param[out] spaceUsed space used by entity map
   */
  void calculateSpaceUsedByEntityMap(double &spaceUsed);
  /**
   * Calculates total space used in memory by eps map, entity map and the buffer
   * \param[out] spaceUsed total space used in bytes
   */
  void calculateSpaceUsed(double &spaceUsed);
  
  ///This is required for the theory to hold.
  bool _force_monotone; //When false, be more aggressive in pruning the bounds.
  // ***********************************
  ///Logging variables
  ///Single Entity Counters
#ifdef __LOGGING
  int log_se_eps_hit, log_se_eps_miss;
  int log_se_buffer_hit, log_se_buffer_miss;
  // Incremental Update Counters
  int log_iu_buffer_hit, log_iu_buffer_miss;
#endif
  // ***********************************
  
 public:
  /**
   * Constructor that initializes Hybrid_Storage_Manager
   * \param[in] db_conn databae connection
   * \param[in] spec ondisk_storage_spec
   * \param[in] disk Ondisk_Storage_Manager
   * \param[in] nBuffer buffer size
   * \param[in] bUseEps whether use eps map or not
   */
  Hybrid_Storage_Manager(Hazy_Database *db_conn, ondisk_storage_spec *spec, Ondisk_Storage_Manager<T> *disk, int nBuffer, bool bUseEps);
  /**
   * Makes a resort to reorganize entities according to their epsilon values
   * \param[in] m hazy model that we use to re-calculate epsllon values
   */
  void resort(struct hazy_model &m);
  /**
   * Makes an incremental update on the entities by using hazy model
   * \param[in] m hazy model that we use to update entities classes
   * \param[out] waste_time waste time calcaulted during the method execution
   */
  void incrementalUpdate(struct hazy_model &m, double &waste_time);
  /**
   * Inserts the entity to the storage manager
   * \param[in] m hazy model which is used when the approach is hazy, because we need to calculate eps value
   * \param[in] e id of the entity
   * \param[in] f feature vector of the entity
   */
  void insertEntity(struct hazy_model &m, key e, T f);
  /**
   * Retrieves the class of the entity
   * \param[in] e entitiy id
   * \param[out] c class of the entity
   * \param[in] m hazy model that we use to classify the entity
   */
  void getEntityClass(key e, sClass &c, struct hazy_model &m);
  /**
   * Retrieves # of elements that belong to a given class
   * \param[in] c class that we ask # of elements belong it
   * \param[out] nClass # of elements in the class
   * \param[in] m hazy model that we use to classify entities
   * \param[out] waste_time waste time calculated during the method execution
   */
  void getNumInClass(sClass c, int &nClass, struct hazy_model &m, double &waste_time);
};

#include "hybrid_storage_manager.hxx"

#endif

