#include "storage_manager.h"
#include "incremental_sgd.h"
#include <map>
#include <vector>

struct main_memory_storage_spec : storage_spec {
  
};

/**
 * STLMainMemory_Storage_Manager extends Storage_Manager and is used for main memory architecture
 */
template<class T>
class STLMainMemory_Storage_Manager : public Storage_Manager<T> {
  
  ///variable that keeps all entity information 
  std::map<key, T> _entity_table;
  ///external table that is used for eager approach
  std::map<key, bool>    _eager_external_table;
  
  hazy_model::strategy strategy;
  ///Maintained only by hazy strategies
  struct internal_record {
    key k;
    float eps;
    T v;
  };
  
  ///internal table structure that is used by Hazy
  std::vector<internal_record> _itable;
  ///To get around C++ compliler parsing issue
  typedef typename std::vector<struct STLMainMemory_Storage_Manager<T>::internal_record>::iterator internal_iterator;
  typedef typename std::map<key,T>::iterator key_iterator_type;
  
  internal_iterator low_water_it_cached, high_water_it_cached;
  
  static bool cluster_sort_predicate(const internal_record &x, const internal_record &y) {
    return x.eps < y.eps;
  }
  
  /**
   * Classifies given entity with the given model
   * \param[in] m given model
   * \param[in] v feature vector
   * \return eps calculated by the given model
   */ 
  double
    classify(struct model& m, T &v) {
    double ret= IncrementalSGD<T>::classifyExample(m, v);
    
    VERBOSE_ONLY(std::cout << "classify " << ret << std::endl;);
    return ret;
  }
 public:
  /**
   * Constructor that loads entities with insertEntity method and do not use database connection
   * \param[in] s main memory storage spec
   */
  STLMainMemory_Storage_Manager(struct main_memory_storage_spec *s);
  /**
   * Constructor that loads entities from the database
   * \param[in] db_conn database connection
   * \param[in] s main memory storage spec
   * \param[in] hm hazy model to calculate entity epsilons
   */
  STLMainMemory_Storage_Manager(Hazy_Database *db_conn, struct main_memory_storage_spec *s, struct hazy_model &hm);
  /**
   * Makes a resort to reorganize entities according to their epsilon values
   * \param[in] hm hazy model that we use to re-calculate epsllon values
   */
  void resort(struct hazy_model &hm);
  
  /**
   * Makes an incremental update on the entities by using hazy model
   * \param[in] m hazy model that we use to update entities classes
   * \param[out] waste_time waste time calcaulted during the method execution
   */
  void incrementalUpdate(struct hazy_model &m, double &waste_time);
  /**
   * Inserts the entity to the storage manager
   * \param[in] m hazy model which is used when the approach is hazy, because we need to calculate eps value
   * \param[in] e id of the entity
   * \param[in] x feature vector of the entity
   */
  void insertEntity(struct hazy_model &m, key e, T x);
  
  /**
   * Retrieves the class of the entity
   * \param[in] e entitiy id
   * \param[out] c class of the entity
   * \param[in] m hazy model that we use to classify the entity
   */
  void getEntityClass(key e, sClass &c, struct hazy_model &m);
  /**
   * Retrieves # of elements that belong to a given class
   * \param[in] c class that we ask # of elements belong it
   * \param[out] nClass # of elements in the class
   * \param[in] m hazy model that we use to classify entities
   * \param[out] waste_time waste time calculated during the method execution
   */
  void getNumInClass(sClass c, int &nClass, struct hazy_model &m, double &waste_time);      
  /**
   * Destructor
   */
  ~STLMainMemory_Storage_Manager() { } 
  
};

// This is apparently best practice to deal with some linker problems for templtes.
// since templates are not code themselves (but allow one to define code)
#include "main_memory_storage_manager.hxx"

