/**
 * \file storage_manager.h
 */
#ifndef ONDISK_STORAGE_MANAGER_H
#define ONDISK_STORAGE_MANAGER_H

#include "storage_manager.h"
#include "vectors.h"
#include "timer.h"

///This the ondisk storage spec
struct ondisk_storage_spec : storage_spec  {
  ///the view that hazy must maintain
  std::string external_table_name;
  ///only used by hazy on disk
  std::string intermediate_table_name;
};

template <class T>
class Hybrid_Storage_Manager;

///Storage manager that interacts with either main memory entities or on-disk entities (or both for hybrid)
template <class T>
class Ondisk_Storage_Manager : public Storage_Manager<T> {
 protected:
  ///External results table name for on-disk
  std::string external_table_name;
  ///Internal results table for on-disk
  std::string hazy_intermediate_table_name;
  
  hazy_model::strategy _strategy;

  ///unique id for prepared statements in order to prevent any conflict
  ///in the same database connection, there cannot be prepared statements with exact same name
  int unique_id_for_ps;

  /**
   * Updates internal table from given model and bias
   * First, the internal table is clustered on new epsilon values based on the hazy model. Then necessary indexes are created
   * \param[in] m hazy_model
   */
  void updateInternalTable(struct hazy_model &m);
  
  /**
   * This function updates a disk region
   * The hybrid calls it (via a friend method) : *** This is factored this way for optimization.
   * The input is a model and X = [lower_bound, upper_bound]  NB: The region is inclusive.
   * for each tuple t s.t. t.eps in X, we reclassify them with the model.
   * \param[in] hm hazy model
   * \param[in] lower_bound low water
   * \param[in] upper_bound high water
   * \param[out] waste_time
   */
  void update_disk_region(struct hazy_model &hm, double lower_bound, double upper_bound, double &waste_time);
  /**
   * Helper function for naive eager that recreates the external table
   * \param[in] hm hazy model
   */
  void rescan(struct hazy_model &hm);
  
  /**
   * Does initial setup
   * If bRebuild is true, recreates all prepared statements and tables. Then, fills internal and external tables by using hazy model
   * \param[in] bRebuild If this is true, then all tables and prepared statements are created
   * \param[in] hm hazy model
   */
  void setup(bool bRebuild, struct hazy_model &hm);
  
 public:
  /**
   * Constructor that takes database connector and ondisk storage spec
   * \param[in] db_conn database connection
   * \param[in] spec ondisk_storage_spec
   */
  Ondisk_Storage_Manager(Hazy_Database *db_conn, struct ondisk_storage_spec *spec);
  
  /**
   * Makes a resort to reorganize entities according to their epsilon values
   * \param[in] m hazy model that we use to re-calculate epsllon values
   */
  void resort(struct hazy_model &m);
  
  /**
   * Makes an incremental update on the entities by using hazy model
   * \param[in] m hazy model that we use to update entities classes
   * \param[out] waste_time waste time calcaulted during the method execution
   */
  void incrementalUpdate(struct hazy_model &m, double &waste_time);
  
  /**
   * Inserts the entity to the storage manager
   * \param[in] m hazy model which is used when the approach is hazy, because we need to calculate eps value
   * \param[in] e id of the entity
   * \param[in] f feature vector of the entity
   */
  virtual void insertEntity(struct hazy_model &m, key e, T f) = 0;
  // void loadFromFile(char*szFileName);
  
  // **********************
  // Read Methods
  // Single Entity Read
  /**
   * Retrieves the class of the entity
   * \param[in] e entitiy id
   * \param[out] c class of the entity
   * \param[in] m hazy model that we use to classify the entity
   */
  void getEntityClass(key e, sClass &c, struct hazy_model &m);
  // All Members Read
  /**
   * Retrieves # of elements that belong to a given class
   * \param[in] c class that we ask # of elements belong it
   * \param[out] nClass # of elements in the class
   * \param[in] m hazy model that we use to classify entities
   * \param[out] waste_time waste time calculated during the method execution
   */
  void getNumInClass(sClass c, int &nClass, struct hazy_model &m, double &waste_time);
  
  //Destructor
  ~Ondisk_Storage_Manager();
  // hybrid is a friend class so that it can access update_disk_region
  // friend class Hybrid_Storage_Manager<T>;
};

/**
 * Ondisk_Storage_Manager_Dense extends Ondisk_Storage_Manager and used for entities that have dense feature vectors
 */
class Ondisk_Storage_Manager_Dense : public Ondisk_Storage_Manager<FVector> {
 public:
 Ondisk_Storage_Manager_Dense(Hazy_Database *db_conn, struct ondisk_storage_spec *spec, struct hazy_model &hm) : Ondisk_Storage_Manager<FVector>(db_conn, spec) {setup(spec->bRebuild, hm); }
  void insertEntity(struct hazy_model &m, key e, FVector f);	
};

/**
 * Ondisk_Storage_Manager_Sparse extends Ondisk_Storage_Manager and used for entities that have sparse feature vectors
 */
class Ondisk_Storage_Manager_Sparse: public Ondisk_Storage_Manager<SVector> {
 public:
 Ondisk_Storage_Manager_Sparse(Hazy_Database *db_conn, struct ondisk_storage_spec *spec, struct hazy_model &hm) : Ondisk_Storage_Manager<SVector>(db_conn, spec) {setup(spec->bRebuild, hm); }
  void insertEntity(struct hazy_model &m, key e, SVector f);
};


/**
 * This class is used to build Ondisk_Storage_Manager that is for either sparse or dense features
 */
template<class T>
class Ondisk_Storage_Manager_Factory {
 protected:
  /**
   * Builds storage manager
   */
  Ondisk_Storage_Manager<T>* build_storage_manager(Hazy_Database *db_conn, struct ondisk_storage_spec *spec) = 0;
 public:
  Ondisk_Storage_Manager_Factory() {}
};

template<>
class Ondisk_Storage_Manager_Factory<SVector> {
 public:
  /**
   * Builds storage manager with type SVector
   */
  Ondisk_Storage_Manager<SVector>* build_storage_manager(Hazy_Database *db_conn, struct ondisk_storage_spec *spec, struct hazy_model &hm) { return new Ondisk_Storage_Manager_Sparse(db_conn, spec, hm); }
};

template<>
class Ondisk_Storage_Manager_Factory<FVector> {
 public:
  /**
   * Builds storage manager with type FVector
   */
  Ondisk_Storage_Manager<FVector>* build_storage_manager(Hazy_Database *db_conn, struct ondisk_storage_spec *spec, struct hazy_model &hm) { return new Ondisk_Storage_Manager_Dense(db_conn, spec, hm); }
};

#include "ondisk_storage_manager.hxx"

#endif
