/**
 * \file storage_manager.cxx
 */
#include <sstream>

/**
 * Does initial setup
 * If bRebuild is true, recreates all prepared statements and tables. Then, fills internal and external tables by using hazy model
 * \param[in] bRebuild If this is true, then all tables and prepared statements are created
 * \param[in[ hm hazy model
 */
template<class T>
void
Ondisk_Storage_Manager<T>::setup (bool bRebuild, struct hazy_model &hm)  { 
  // This is where we prepare the statements.
  std::ostringstream ostr;
  std::ostringstream schema_setup;
  
  ostr << "PREPARE model_initialization" << unique_id_for_ps << " AS SELECT initialize_waste_count" 
       << unique_id_for_ps << "();" << std::endl;;
  ostr << "PREPARE model_caching" << unique_id_for_ps << "(FLOAT8[], FLOAT8) AS SELECT cache_model_waste" << unique_id_for_ps << "($1, $2);" << std::endl;;
  DEBUG_ONLY(std::cout << ostr.str() << std::endl;);
  int retVal = this->db_conn->execute_statement_msg(ostr.str().c_str(), "[Ondisk_Storage_Manager::model_prepared_statements:__FILE__: __LINE__");
  checkQueryReturnValue(retVal, ostr.str());
  
  if(bRebuild) {
    // external table could be either view or table...
    //schema_setup << "DROP VIEW  IF EXISTS " << external_table_name << " CASCADE;" << std::endl;
    schema_setup << "SELECT delete_table_or_view('" << external_table_name << "');" ;
    if(hazy_model::isHazy(_strategy)) {
      std::string model_string, bias_string;
      
      schema_setup << "DROP TABLE IF EXISTS " << hazy_intermediate_table_name << " CASCADE;" << std::endl;
      schema_setup << "EXECUTE model_initialization" << unique_id_for_ps << ";";
      if(!hm.test_and_set_model_in_db()) { model_to_dbstring(hm._model, model_string, bias_string); schema_setup << "EXECUTE model_caching" << unique_id_for_ps << "(" << model_string << ", " << bias_string << ");";}
      schema_setup << "CREATE TABLE " << hazy_intermediate_table_name << " AS (SELECT id, (dotprdct_cached_without_waste" << unique_id_for_ps << "(feature_vector))::float8 as eps, feature_vector FROM " << this->entity_table_name << ");" << std::endl;
      schema_setup << "CREATE INDEX index1_" << hazy_intermediate_table_name << " ON " << hazy_intermediate_table_name << "(eps, id) WITH (FILLFACTOR=100);";
      schema_setup << "CREATE INDEX index2_" << hazy_intermediate_table_name << " ON " << hazy_intermediate_table_name << " USING hash(id);";
      schema_setup << "CLUSTER index1_" << hazy_intermediate_table_name << " ON " << hazy_intermediate_table_name << ";";
    }
    switch (_strategy) {
    case hazy_model::LAZY_NAIVE:
      schema_setup << "CREATE VIEW " << external_table_name << " AS SELECT id, getEntityLabelWithSQLNOEPS(feature_vector::text) as class from " << this->entity_table_name << ";" << std::endl;
      break;
    case hazy_model::LAZY_HAZY: 
      schema_setup << "CREATE VIEW " << external_table_name << " AS SELECT id, eps, getEntityLabelWithSQL(eps, feature_vector::text) as class from " << hazy_intermediate_table_name << ";" << std::endl;
      break;
      
    case hazy_model::EAGER_NAIVE:
      {		  
	schema_setup << "EXECUTE model_initialization" << unique_id_for_ps << ";";
	if(!hm.test_and_set_model_in_db()) { 
	  std::string model_string_naive, bias_string_naive;
	  model_to_dbstring(hm._model, model_string_naive, bias_string_naive); 
	  schema_setup << "EXECUTE model_caching" << unique_id_for_ps << "(" << model_string_naive << ", " << bias_string_naive << ");";
	} 
	schema_setup << "CREATE TABLE " << external_table_name << " AS (SELECT id, (dotprdct_cached_without_waste" << unique_id_for_ps << "(feature_vector) > 0)::int as class FROM " << this->entity_table_name << ");" << std::endl;
	schema_setup << "CREATE INDEX index2_" << external_table_name << " ON " << external_table_name << " USING hash(id);";   
      }
      break;
    case hazy_model::EAGER_HAZY:
      {
	schema_setup << "EXECUTE model_initialization" << unique_id_for_ps << ";";
	if(!hm.test_and_set_model_in_db()) { 
	  std::string model_string_hazy, bias_string_hazy;
	  model_to_dbstring(hm._model, model_string_hazy, bias_string_hazy); 
	  schema_setup << "EXECUTE model_caching" << unique_id_for_ps << "(" << model_string_hazy << ", " << bias_string_hazy << ");"; 
	}
	schema_setup << "CREATE TABLE " << external_table_name << " AS (SELECT id, (dotprdct_cached_without_waste" << unique_id_for_ps << "(feature_vector))::float8 as eps, (dotprdct_cached_without_waste" << unique_id_for_ps << "(feature_vector) > 0)::int as class FROM " << this->entity_table_name << ");" << std::endl;
	schema_setup << "CREATE INDEX index1_" << external_table_name << " ON " << external_table_name << "(eps) WITH (FILLFACTOR=100);";
	schema_setup << "CREATE INDEX index2_" << external_table_name << " ON " << external_table_name << " USING hash(id);";
      }
      break;
    default:
      std::cout << "[SETUP] Unknown Strategy " << std::endl;
      assert(false);      
    }
    schema_setup << "CREATE OR REPLACE FUNCTION update_external_table_inplace" << unique_id_for_ps << "(" << external_table_name << ", bool) RETURNS BOOLEAN AS 'inplace" << unique_id_for_ps << ".so', 'update_hack" << unique_id_for_ps << "' LANGUAGE C STRICT;"; 
    VERBOSE_ONLY(std::cout << "SETUP SQL Queries [Rebuild]" << std::endl << schema_setup.str() << std::endl << std::endl; );
    int retVal = this->db_conn->execute_statement_msg(schema_setup.str().c_str(), "[Ondisk_Storage_Manager::Setup:REBUILD]");
    checkQueryReturnValue(retVal, schema_setup.str());    
  }
  // This is for sparse vectors
  ostr.str("");
  if(hazy_model::isHazy(_strategy)) {
    ostr << "PREPARE view_recreate_in_hazy" << unique_id_for_ps << " AS SELECT id, eps, (eps > 0)::integer AS class INTO " << external_table_name << " FROM " << hazy_intermediate_table_name << " ORDER BY eps, id;" << std::endl;;
    ostr << "PREPARE single_hazy_lazy" << unique_id_for_ps << "(INT, FLOAT8, FLOAT8) AS SELECT CASE WHEN eps < $2 THEN 0 ELSE CASE WHEN eps > $3 THEN 1 ELSE (CASE WHEN (dotprdct_cached_without_waste" << unique_id_for_ps << "(feature_vector) > 0) THEN 1 ELSE 0 END) END END FROM " << hazy_intermediate_table_name << " WHERE id = $1;";
    ostr << "PREPARE count_hazy_lazy" << unique_id_for_ps << "(FLOAT8, FLOAT8) AS SELECT ((SELECT COUNT(*) FROM " << hazy_intermediate_table_name << " WHERE (eps BETWEEN $1 AND $2) AND ((dotprdct_cached_waste" << unique_id_for_ps << "(feature_vector)) > 0)) + (SELECT COUNT(*) FROM " << hazy_intermediate_table_name << " WHERE eps > $2));" << std::endl;;
    ostr << "PREPARE count_hazy_eager" << unique_id_for_ps << "(FLOAT8, FLOAT8) AS SELECT (SELECT COUNT(*) FROM " << external_table_name << " WHERE eps BETWEEN $1 AND $2 AND class = 1) + (SELECT COUNT(*) FROM " << external_table_name << " WHERE eps > $2);" << std::endl;;
    ostr << "PREPARE PREPARED_CREATE_VIEW2" << unique_id_for_ps << "(FLOAT8) AS SELECT id INTO " << external_table_name << " FROM " << hazy_intermediate_table_name << " WHERE eps > $1;" << std::endl;;
    ostr << "PREPARE PREPARED_CREATE_VIEW" << unique_id_for_ps << "(FLOAT8, FLOAT8) AS SELECT update_external_table_inplace" << unique_id_for_ps << "(e, dotprdct_cached_waste" << unique_id_for_ps << "(i.feature_vector) > 0) FROM " << hazy_intermediate_table_name << " i, " << external_table_name << " e ";
    //ostr << "WHERE (i.eps BETWEEN $1 AND $2) AND (e.eps BETWEEN $1 AND $2) AND (e.eps = i.eps) AND (e.id = i.id);" << std::endl;;
    ostr << "WHERE (i.eps BETWEEN $1 AND $2) AND (e.eps BETWEEN $1 AND $2) AND (e.id = i.id);" << std::endl;;
    ostr << "PREPARE internal_recreate" << unique_id_for_ps << " AS SELECT id, (dotprdct_cached_without_waste" << unique_id_for_ps << "(feature_vector)) AS eps, feature_vector INTO " << hazy_intermediate_table_name << "_copy" << " FROM " << hazy_intermediate_table_name << " ORDER BY eps, id;" << std::endl;;
  }
  else {
    ostr << "PREPARE view_recreate_in_naive" << unique_id_for_ps << " AS SELECT id, (dotprdct_cached_without_waste" << unique_id_for_ps << "(feature_vector) > 0)::integer as class INTO " << external_table_name << " FROM " << this->entity_table_name << ";" << std::endl;;
    ostr << "PREPARE single_naive_lazy" << unique_id_for_ps << "(INT) AS SELECT CASE WHEN (dotprdct_cached_without_waste" << unique_id_for_ps << "(feature_vector) > 0) THEN 1 ELSE 0 END FROM " << this->entity_table_name << " WHERE id = $1;";
    ostr << "PREPARE count_naive_lazy" << unique_id_for_ps << " AS SELECT COUNT(*) FROM " << this->entity_table_name << " WHERE (dotprdct_cached_without_waste" << unique_id_for_ps << "(feature_vector) > 0);" << std::endl;;
    ostr << "PREPARE count_naive_eager" << unique_id_for_ps << " AS SELECT COUNT(*) FROM " << external_table_name << " WHERE class = 1;" << std::endl;;
  }
  ostr << "PREPARE retrieve_ratio" << unique_id_for_ps << " AS SELECT get_waste_ratio" << unique_id_for_ps << "();" << std::endl;;
  ostr << "PREPARE single_entity_eager" << unique_id_for_ps << "(INT) AS SELECT class FROM " << external_table_name << " WHERE id = $1;" << std::endl;;
  
  // These prepare statements are for dense vectors.
  // ostr << "PREPARE model_initialization AS SELECT initialize_dense_waste_count();";
  // ostr << "PREPARE model_caching(FLOAT8[], FLOAT8) AS SELECT cache_dense_model_waste($1, $2);";
  // ostr << "PREPARE retrieve_ratio AS SELECT get_dense_waste_ratio();";
  // ostr << "PREPARE internal_recreate AS SELECT id, feature_value_array, dotprdctdense_cached_without_waste(feature_value_array) AS c INTO " << hazy_intermediate_table_name << "_copy" << " FROM " << hazy_intermediate_table_name << " ORDER BY c, id;";
  // ostr << "PREPARE view_recreate_in_hazy AS SELECT id, c, (c > 0)::boolean AS class INTO " << external_table_name << " FROM " << hazy_intermediate_table_name << " ORDER BY c, id;";
  // ostr << "PREPARE view_recreate_in_naive AS SELECT id, (dotprdctdense_cached_without_waste(feature_value_array) > 0)::boolean as class INTO " << external_table_name << " FROM " << entity_table_name << ";";
  // ostr << "PREPARE single_entity_eager(INT) AS SELECT class FROM " << external_table_name << " WHERE id = $1;";
  // ostr << "PREPARE count_hazy_lazy(FLOAT8, FLOAT8) AS SELECT ((SELECT COUNT(*) FROM " << hazy_intermediate_table_name << " WHERE (c BETWEEN $1 AND $2) AND ((dotprdctdense_cached_waste(feature_value_array)) > 0)) + (SELECT COUNT(*) FROM " << hazy_intermediate_table_name << " WHERE c > $2));";
  // ostr << "PREPARE count_naive_lazy AS SELECT COUNT(*) FROM " << entity_table_name << " WHERE (dotprdctdense_cached_without_waste(feature_value_array) > 0);";
  // ostr << "PREPARE count_hazy_eager(FLOAT8) AS SELECT COUNT(*) FROM " << external_table_name << " WHERE class='t' AND c >= $1;";
  // ostr << "PREPARE count_naive_eager AS SELECT COUNT(*) FROM " << external_table_name << " WHERE class = 't';";
  // ostr << "PREPARE PREPARED_CREATE_VIEW2(FLOAT4) AS  SELECT id INTO " << external_table_name << " FROM " << hazy_intermediate_table_name << " WHERE c > $1;";
  // ostr << "PREPARE PREPARED_CREATE_VIEW(FLOAT4, FLOAT4) AS SELECT update_external_table_inplace(e, dotprdctdense_cached_waste(i.feature_value_array) > 0) FROM " << hazy_intermediate_table_name << " i, " << external_table_name << " e ";
  // ostr << "WHERE (i.c BETWEEN $1 AND $2) AND (e.c BETWEEN $1 AND $2) AND (e.c = i.c) AND (e.id = i.id);";
  // end dense vector code
  DEBUG_ONLY(std::cout << ostr.str() << std::endl;)
    retVal = this->db_conn->execute_statement_msg(ostr.str().c_str(), "[Ondisk_Storage_Manager::Setup:__FILE__: __LINE__");
  checkQueryReturnValue(retVal, ostr.str());
  // what happens if it fails?
}

/**
 * Destructor
 */
template<class T>
Ondisk_Storage_Manager<T>::~Ondisk_Storage_Manager ()  { }

///Constructor that takes database connector, view id and entity table name
/**
 * Constructor that takes database connector and ondisk storage spec
 * \param[in] db_conn database connection
 * \param[in] spec ondisk_storage_spec
 */
template<class T>
Ondisk_Storage_Manager<T>
::Ondisk_Storage_Manager(Hazy_Database *db_conn, struct ondisk_storage_spec *spec) : Storage_Manager<T>(db_conn) {
  std::ostringstream ostr;
  
  this->entity_table_name            = spec->entity_table_name;
  this->external_table_name          = spec->external_table_name;
  this->hazy_intermediate_table_name = spec->intermediate_table_name;
  _strategy                          = spec->strategy;
  unique_id_for_ps                   = spec->unique_id;
  
  if(hazy_model::isHazy(_strategy) && hazy_intermediate_table_name.length() == 0) { 
    std::cout << "PANIC: We are a hazy ondisk strategy, but we have no (or invalid) interemediate table name!" << std::endl;
    assert(false);
  }
}


///Updates external results table with naive rescan approach
/**
 * Helper function for naive eager that recreates the external table
 * \param[in] hm hazy model
 */
template<class T>
void
Ondisk_Storage_Manager<T>::rescan(struct hazy_model &hm) {
  model &m       = hm._model;
  FVector w     = m.w; 
  
  std::ostringstream ostr,ostr2,ostr3;
  std::string w_string, bias_string;

  int retVal = 0;
  
  ostr << "DROP TABLE IF EXISTS " << external_table_name << " CASCADE;";
  ostr << "EXECUTE model_initialization" << unique_id_for_ps << ";";
  
  model_to_dbstring(hm._model, w_string, bias_string); 
  ostr << "EXECUTE model_caching" << unique_id_for_ps << "(" << w_string << ", " << bias_string << ");"; 
  
  LOGGING_ONLY(Timer t1(true););
  retVal = this->db_conn->execute_statement_msg(ostr.str().c_str(), "[naive rescan]");
  checkQueryReturnValue(retVal, ostr.str());
  LOGGING_ONLY(std::cout << "n-rescan drop table and caching: " << t1.stop(););
  
  ostr2 << "EXECUTE view_recreate_in_naive" << unique_id_for_ps << ";";
  LOGGING_ONLY(t1.start(););
  retVal = this->db_conn->execute_statement_msg(ostr2.str().c_str(), "[naive rescan]");
  checkQueryReturnValue(retVal, ostr2.str());
  LOGGING_ONLY(std::cout << "n-rescan create table: " << t1.stop(););
  
  ostr3 << "CREATE INDEX index2_" << external_table_name << " ON " << external_table_name << " USING hash(id);";   
  LOGGING_ONLY(t1.start());
  retVal = this->db_conn->execute_statement_msg(ostr3.str().c_str(), "[naive rescan]");
  checkQueryReturnValue(retVal, ostr3.str());
  LOGGING_ONLY(std::cout << "n-rescan create index: " << t1.stop(););	
}


///Resorts internal table and updates external results table
/**
 * Makes a resort to reorganize entities according to their epsilon values
 * \param[in] m hazy model that we use to re-calculate epsllon values
 */
template<class T>
void
Ondisk_Storage_Manager<T>::resort(struct hazy_model &hm) {
  updateInternalTable(hm);
  
  std::ostringstream ostr1;
  ostr1 << "DROP TABLE IF EXISTS " << external_table_name << " CASCADE;";
  ostr1 << "EXECUTE view_recreate_in_hazy" << unique_id_for_ps << ";";
  ostr1 << "CREATE OR REPLACE FUNCTION update_external_table_inplace" << unique_id_for_ps << "(" << external_table_name << ", bool) RETURNS BOOLEAN AS 'inplace" << unique_id_for_ps << ".so', 'update_hack" << unique_id_for_ps << "' LANGUAGE C STRICT;"; 
  ostr1 << "CREATE INDEX index1_" << external_table_name << " ON " << external_table_name << "(eps) WITH (FILLFACTOR=100);";
  ostr1 << "CREATE INDEX index2_" << external_table_name << " ON " << external_table_name << " USING hash(id);";
  
  LOGGING_ONLY(Timer t1(true););
  int retVal = this->db_conn->execute_statement_msg(ostr1.str().c_str(), "[Update:Rebuild]");
  checkQueryReturnValue(retVal, ostr1.str());
  LOGGING_ONLY(std::cout << "after sort, update view time: " << t1.stop() << std::endl;);	
}


///Incrementally updates external results table and returns waste ratio
/**
 * Makes an incremental update on the entities by using hazy model
 * \param[in] m hazy model that we use to update entities classes
 * \param[out] waste_time waste time calcaulted during the method execution
 */
template<class T>
void
Ondisk_Storage_Manager<T>::incrementalUpdate(struct hazy_model &hm, double &waste_time) {
  hm.invalidate_db_model();
  
  switch (hm._strategy) {
  case hazy_model::LAZY_NAIVE:
    break;
  case hazy_model::LAZY_HAZY: 
    VERBOSE_ONLY(std::cout << "[incrementalUpdate]: Lazy does nothing" << std::endl);
    break;
  case hazy_model::EAGER_NAIVE:
    rescan(hm);
    break;
  case hazy_model::EAGER_HAZY:
    update_disk_region(hm, hm.low_water, hm.high_water, waste_time);
    break;
  default:
    std::cout << "Unknown Strategy " << std::endl;
    assert(false);
  }
}

///Updates internal table from given model and bias
/**
 * Updates internal table from given model and bias
 * First, the internal table is clustered on new epsilon values based on the hazy model. Then necessary indexes are created
 * \param[in] m hazy_model
 */
template<class T>
void
Ondisk_Storage_Manager<T>::updateInternalTable(struct hazy_model &hm) {
  std::string w_string, bias_string;
  model_to_dbstring(hm._model, w_string, bias_string);
  
  std::ostringstream querystr;
  querystr << "EXECUTE model_initialization" << unique_id_for_ps << ";";
  querystr << "EXECUTE model_caching" << unique_id_for_ps << "(" << w_string << ", " << bias_string << ");";
  querystr << "EXECUTE internal_recreate" << unique_id_for_ps << ";";
  querystr << "DROP TABLE IF EXISTS " << hazy_intermediate_table_name << " CASCADE;";
  querystr << "ALTER TABLE " << hazy_intermediate_table_name << "_copy" << " RENAME TO " << hazy_intermediate_table_name << ";";
  querystr << "CREATE INDEX index1_" << hazy_intermediate_table_name << " ON " << hazy_intermediate_table_name << "(eps, id) WITH (FILLFACTOR=100);";
  querystr << "CREATE INDEX index2_" << hazy_intermediate_table_name << " ON " << hazy_intermediate_table_name << " USING hash(id);";
  
  LOGGING_ONLY(Timer t1(true););
  
  int retVal = this->db_conn->execute_statement_msg(querystr.str().c_str(), "[UPDATE INTERNAL TABLE]");
  checkQueryReturnValue(retVal, querystr.str());
  
  LOGGING_ONLY(std::cout << "sort: time for recreating the table with two indexes: " << t1.stop() << std::endl;);
#ifdef __LOGGING
  std::ostringstream detailed_count_query;
  detailed_count_query.str("");
  detailed_count_query << "SELECT COUNT(*) FROM " << hazy_intermediate_table_name << " WHERE (eps > 0) <> (dotprdct_cached_without_waste" << unique_id_for_ps << "(feature_vector) > 0);";
  int count_val = 0;
  this->db_conn->execute_query_msg_int(detailed_count_query.str().c_str(), "[LOGGING sort info]", count_val);
  std::cout << "\t Change Label=" << count_val << std::endl;
#endif
  
}

/**
 * This function updates a disk region
 * The hybrid calls it (via a friend method) : *** This is factored this way for optimization.
 * The input is a model and X = [lower_bound, upper_bound]  NB: The region is inclusive.
 * for each tuple t s.t. t.eps in X, we reclassify them with the model.
 * \param[in] hm hazy model
 * \param[in] lower_bound low water
 * \param[in] upper_bound high water
 * \param[out] waste_time
 */
template<class T>
void
Ondisk_Storage_Manager<T>::
update_disk_region(struct hazy_model &hm, double lower_bound, double upper_bound, double &waste_time) {
  std::ostringstream ostr1, ostr2, ostr3;
  
  double waste_ratio = 0.0;
  std::string w_string, bias_string;
  model_to_dbstring(hm._model, w_string, bias_string);
  DEBUG_ONLY(std::cout << "[OD:update_disk_region:debug] model is: " << hm._model << std::endl << std::endl;);
  
  ostr1 << "EXECUTE model_initialization" << unique_id_for_ps << ";";
  if(!hm.test_and_set_model_in_db()) { ostr1 << "EXECUTE model_caching" << unique_id_for_ps << "(" << w_string << ", " << bias_string << ");"; }
  LOGGING_ONLY(std::cout <<"[OD:update_disk_region]" << std::endl;);
  
  LOGGING_ONLY(Timer t1(true););
  int retVal = this->db_conn->execute_statement_msg(ostr1.str().c_str(), "PROBLEM in insert query in update method\n");
  checkQueryReturnValue(retVal, ostr1.str());
  LOGGING_ONLY(std::cout << "\t time for drop view, initialize waste count and cache model:" << t1.stop() << std::endl;);
  
  ostr2 << "EXECUTE PREPARED_CREATE_VIEW" << unique_id_for_ps << "(" << lower_bound << ", " << upper_bound << ");";
  Timer waste_timer(true);
  DEBUG_ONLY(std::cout << "\t\t [lower_bound,upper_bound]=" << lower_bound << ", " << upper_bound << std::endl;);
  retVal = this->db_conn->execute_statement_msg(ostr2.str().c_str(), "[UPDATE]");
  checkQueryReturnValue(retVal, ostr2.str());
  double total_time = waste_timer.stop();
  LOGGING_ONLY(std::cout << "\t time for prepared statement: " << total_time << std::endl;);
  
  ostr3 << "EXECUTE retrieve_ratio" << unique_id_for_ps << ";";
  retVal = this->db_conn->execute_query_msg_double(ostr3.str().c_str(), "[update]", waste_ratio);
  checkQueryReturnValue(retVal, ostr3.str());
  waste_time = total_time * waste_ratio;
  LOGGING_ONLY(std::cout << "\t Waste Ratio is " << waste_ratio << " waste_time " << waste_time << std::endl;);
  
  // **** START LOGGING ****
#ifdef __LOGGING    
  std::ostringstream detailed_count_query;
  detailed_count_query << "SELECT COUNT(*) FROM " << hazy_intermediate_table_name << " WHERE eps BETWEEN " << lower_bound << " AND " << upper_bound << ";";
  int tuplesBetween = 0;
  retVal = this->db_conn->execute_query_msg_int(detailed_count_query.str().c_str(), "[update]", tuplesBetween);
  checkQueryReturnValue(retVal, detailed_count_query.str());
  std::cout << "\t tuples between lower_bound and upper bound: " << tuplesBetween << std::endl;
  
  detailed_count_query.str("");
  detailed_count_query << "SELECT COUNT(*) FROM " << hazy_intermediate_table_name << " WHERE (eps > 0) <> (dotprdct_cached_without_waste" << unique_id_for_ps << "(feature_vector) > 0);";
  int nChangeLabels = 0;
  retVal = this->db_conn->execute_query_msg_int(detailed_count_query.str().c_str(), "[update]", nChangeLabels);
  checkQueryReturnValue(retVal, detailed_count_query.str());
  std::cout << "\t tuples changed labels: " << nChangeLabels << std::endl;
#endif
  // **** END LOGGING ****  
}

/**
 * Retrieves the class of the entity
 * \param[in] e entitiy id
 * \param[out] c class of the entity
 * \param[in] m hazy model that we use to classify the entity
 */
template<class T>
void 
Ondisk_Storage_Manager<T>::
getEntityClass(key e, sClass &c, struct hazy_model &hm) {
  std::ostringstream o; // will hold the SQL command
  std::string model_string, bias_string;
  
  switch (hm._strategy) {
  case hazy_model::LAZY_NAIVE:
    // o << "EXECUTE model_initialization;";
    if(!hm.test_and_set_model_in_db()) { 
      model_to_dbstring(hm._model, model_string, bias_string);
      LOGGING_ONLY(std::cout << "[Model is not Cached]" << std::endl;);
      o << "EXECUTE model_caching" << unique_id_for_ps << "(" << model_string << ", " << bias_string << ");"; 
    } else {
      LOGGING_ONLY(std::cout << "[Model is Cached]" << std::endl;);
    }
    o << "EXECUTE single_naive_lazy" << unique_id_for_ps << "(" << e << ");";
    break;
  case hazy_model::LAZY_HAZY: 
    //o << "EXECUTE model_initialization;";
    if(!hm.test_and_set_model_in_db()) { 
      model_to_dbstring(hm._model, model_string, bias_string);
      LOGGING_ONLY(std::cout << "[Model is not Cached]" << std::endl;);
      o << "EXECUTE model_caching" << unique_id_for_ps << "(" << model_string << ", " << bias_string << ");"; 
    } else {
      LOGGING_ONLY(std::cout << "[Model is Cached]" << std::endl;);
    }
    o << "EXECUTE single_hazy_lazy" << unique_id_for_ps << "(" << e << ", " << hm.low_water << ", " << hm.high_water << ");";
    break;
  case hazy_model::EAGER_NAIVE:
  case hazy_model::EAGER_HAZY: 
    o << "EXECUTE single_entity_eager" << unique_id_for_ps << "(" << e << ");"; 
    break;
  default:
    std::cout << "[getEntityClass] Non Lazy or Eager strategy???" << std::endl;
    assert(false);
  }
  
  VERBOSE_ONLY(std::cout << "[getEntityClass] " << o.str() << std::endl;);
  LOGGING_ONLY(Timer qtimer(true););
  int retVal = this->db_conn->execute_query_msg_int(o.str().c_str(), "[Ondisk_Storage_Manager::GetEntityClass:__FILE__: __LINE__", c);
  checkQueryReturnValue(retVal, o.str());  
  LOGGING_ONLY(std::cout << "label of " << e << " is: " << c << " time:" << qtimer.stop() << std::endl;);
}

/**
 * Retrieves # of elements that belong to a given class
 * \param[in] c class that we ask # of elements belong it
 * \param[out] nClass # of elements in the class
 * \param[in] m hazy model that we use to classify entities
 * \param[out] waste_time waste time calculated during the method execution
 */
template<class T>
void
Ondisk_Storage_Manager<T>::
getNumInClass(sClass c, int &nClass, struct hazy_model &hm, double &waste_time) {
  assert(c == 1);
  std::string model_string, bias_string;
  std::ostringstream o;
  
  waste_time = 0.0;
  switch (hm._strategy) {
  case hazy_model::LAZY_NAIVE: 
    o << "EXECUTE model_initialization" << unique_id_for_ps << ";";
    if(!hm.test_and_set_model_in_db()) {    
      model_to_dbstring(hm._model, model_string, bias_string);    
      o << "EXECUTE model_caching" << unique_id_for_ps << "(" << model_string << ", " << bias_string << ");"; 
    }
    o << "EXECUTE count_naive_lazy" << unique_id_for_ps << ";";
    break;
  case hazy_model::LAZY_HAZY:
    o << "EXECUTE model_initialization" << unique_id_for_ps << ";";
    if(!hm.test_and_set_model_in_db()) {   
      model_to_dbstring(hm._model, model_string, bias_string);    
      o << "EXECUTE model_caching" << unique_id_for_ps << "(" << model_string << ", " << bias_string << ");"; 
    }
    o << "EXECUTE count_hazy_lazy" << unique_id_for_ps << "(" << hm.low_water << "," << hm.high_water << ");";
    break;
  case hazy_model::EAGER_NAIVE:
    o << "EXECUTE count_naive_eager" << unique_id_for_ps << ";"; 
    break;
  case hazy_model::EAGER_HAZY:
    o << "EXECUTE count_hazy_eager" << unique_id_for_ps << "(" << hm.low_water << ", " << hm.high_water << ");";
    break;
  default:
    std::cout << "UNKNOWN STRATEGY!!" << std::endl;
    assert(false);    
    break;
  }
  // Only LAZY, HAZY needs to update waste
  VERBOSE_ONLY(std::cout << "[getNumInClass] " << o.str() << std::endl;);
  if( hm._strategy == hazy_model::LAZY_HAZY ) {
    Timer t(true);
    double waste_ratio = 0.0;
    int retVal = this->db_conn->execute_query_msg_int(o.str().c_str(), "[Ondisk_Storage_Manager::GetEntityClass:__FILE__: __LINE__", nClass);
    checkQueryReturnValue(retVal, o.str());  
    std::ostringstream o2;
    o2 << "EXECUTE retrieve_ratio" << unique_id_for_ps << ";";
    retVal = this->db_conn->execute_query_msg_double(o2.str().c_str(), "[Ondisk_Storage_Manager::GetEntityClass:__FILE__: __LINE__", waste_ratio);
    checkQueryReturnValue(retVal, o2.str());  
    waste_time = t.stop() * waste_ratio;
    LOGGING_ONLY(std::cout << "\tWaste Ratio: " << waste_ratio << " waste time: " << waste_time << std::endl;);
  } else {
    int retVal = this->db_conn->execute_query_msg_int(o.str().c_str(), "[Ondisk_Storage_Manager::GetEntityClass:__FILE__: __LINE__", nClass);
    checkQueryReturnValue(retVal, o.str());  
    waste_time = 0.0;
  }
  
  LOGGING_ONLY(std::cout << "num in class " << c << " is: " << nClass << std::endl;);
}

/**
 * Inserts the entity to the storage manager
 * \param[in] m hazy model which is used when the approach is hazy, because we need to calculate eps value
 * \param[in] id id of the entity
 * \param[in] f feature vector of the entity
 */
void
Ondisk_Storage_Manager_Sparse::
insertEntity(struct hazy_model &m, key id, SVector f) {
  std::string feature_vector_str;
  svector_to_dbstring(f, feature_vector_str);
  
  std::ostringstream ostr;
  ostr << "INSERT INTO " << this->entity_table_name << "(id,feature_vector) ";
  ostr << "VALUES (" << id << "," << feature_vector_str << ");";
  
  if(hazy_model::isHazy(_strategy) || hazy_model::isEager(_strategy)) {
    std::string model_string, bias_string;
    
    ostr << "EXECUTE model_initialization" << unique_id_for_ps << ";";
    if(!m.test_and_set_model_in_db()) { 
      model_to_dbstring(m._model, model_string, bias_string); 
      ostr << "EXECUTE model_caching" << unique_id_for_ps << "(" << model_string << ", " << bias_string << ");"; 
    }
    
    if(hazy_model::isHazy(_strategy))
      ostr << "INSERT INTO " << hazy_intermediate_table_name << "(id, eps, feature_vector) VALUES (" << id << ", " << "dotprdct_cached_without_waste" << unique_id_for_ps << "(" << feature_vector_str << ")::float8" << ", " << feature_vector_str << ");";
    if(_strategy == hazy_model::EAGER_HAZY)
      ostr << "INSERT INTO " << external_table_name << "(id, eps, class) VALUES (" << id << ", " << "dotprdct_cached_without_waste" << unique_id_for_ps << "(" << feature_vector_str << ")::float8" << ", (dotprdct_cached_without_waste" << unique_id_for_ps << "(" << feature_vector_str << ")::float8 > 0)::int);";
    else if(_strategy == hazy_model::EAGER_NAIVE)
      ostr << "INSERT INTO " << external_table_name << "(id, class) VALUES (" << id << ", " << "(dotprdct_cached_without_waste" << unique_id_for_ps << "(" << feature_vector_str << ") > 0)::int);";
  }
  int retVal = db_conn->execute_statement_msg(ostr.str().c_str(), "[Ondisk_Storage_Manager::InsertEntity:__FILE__: __LINE__");
  checkQueryReturnValue(retVal, ostr.str());
  LOGGING_ONLY(std::cout << "entity inserted, id: " << id << std::endl;);
}

/**
 * Inserts the entity to the storage manager
 * \param[in] m hazy model which is used when the approach is hazy, because we need to calculate eps value
 * \param[in] id id of the entity
 * \param[in] f feature vector of the entity
 */
void
Ondisk_Storage_Manager_Dense::insertEntity(struct hazy_model &m, key id, FVector f) {
  std::string feature_vector_str;
  fvector_to_dbstring(f, feature_vector_str);
  
  std::ostringstream ostr;
  ostr << "INSERT INTO " << this->entity_table_name << "(id,feature_vector) ";
  ostr << "VALUES (" << id << "," << feature_vector_str << ");";
  
  if(hazy_model::isHazy(_strategy) || hazy_model::isEager(_strategy)) {
    std::string model_string, bias_string;
    
    ostr << "EXECUTE model_initialization" << unique_id_for_ps << ";";
    if(!m.test_and_set_model_in_db()) { model_to_dbstring(m._model, model_string, bias_string); ostr << "EXECUTE model_caching" << unique_id_for_ps << "(" << model_string << ", " << bias_string << ");"; }
    
    if(hazy_model::isHazy(_strategy))
      ostr << "INSERT INTO " << hazy_intermediate_table_name << "(id, eps, feature_vector) VALUES (" << id << ", " << "dotprdct_cached_without_waste" << unique_id_for_ps << "(" << feature_vector_str << ")::float8" << ", " << feature_vector_str << ");";
    if(_strategy == hazy_model::EAGER_HAZY)
      ostr << "INSERT INTO " << external_table_name << "(id, eps, class) VALUES (" << id << ", " << "dotprdct_cached_without_waste" << unique_id_for_ps << "(" << feature_vector_str << ")::float8" << ", (dotprdct_cached_without_waste" << unique_id_for_ps << "(" << feature_vector_str << ")::float8 > 0)::int);";
    else if(_strategy == hazy_model::EAGER_NAIVE)
      ostr << "INSERT INTO " << external_table_name << "(id, class) VALUES (" << id << ", " << "(dotprdct_cached_without_waste" << unique_id_for_ps << "(" << feature_vector_str << ") > 0)::int);";
  }
  int retVal = db_conn->execute_statement_msg(ostr.str().c_str(), "[Ondisk_Storage_Manager::InsertEntity:__FILE__: __LINE__");
  checkQueryReturnValue(retVal, ostr.str());
  LOGGING_ONLY(std::cout << "entity inserted, id: " << id << std::endl;);
}
