/**
 * result_table.cxx
 */
#include <iostream>
#include <sstream>

#include "result_table.h"

using namespace std;
/**
 * Constructor
 * \param[in] entitySize # of entities
 */
template<class T>
ResultTable<T>::ResultTable(int ent_size) {
  this->entity_size = ent_size;
  table = new resultTuple[ent_size];
  pFeatures = new SVector[ent_size];
  packedIds = new int[ent_size];
  bCluster = true;
}
/**
 * Returns id of tuple at given index
 * \param[in] index index of entity in entity table
 * \return id of tuple
 */
template<class T>
int 
ResultTable<T>::getId(int index) {
  return table[index].id;
}
/**
 * Returns pointer to the id of tuple at given index
 * \param[in] index index of entity in entity table
 * \return pointer to the id of tuple
 */
template<class T>
int*
ResultTable<T>::getPointer(int index) {
  return &(packedIds[index]);
}

/**
 * Returns feature vector of tuple at given index
 * \param[in] index index of entity in entity table
 * \return feature vector of tuple
 */
template<class T>
const T*
ResultTable<T>::getFeatureArray(int index) {
  return table[index].featureArray;
}
/**
 * Returns eps of tuple at given index
 * \param[in] index index of entity in entity table
 * \return epsilon of tuple
 */
template<class T>
double
ResultTable<T>::getEps(int index) {
  return table[index].eps;
}
/**
 * Returns truth value of tuple at given index
 * \param[in] index index of entity in entity table
 * \return truth value of tuple
 */
template<class T>
int
ResultTable<T>::getTruthValue(int index) {
  return table[index].truthValue;
}

/**
 * Sets id of tuple at given index
 * \param[in] index index of entity in entity table
 * \param[in] id new id
 */
template<class T>
void
ResultTable<T>::setId(int index, int id) {
  packedIds[index] = id;
  table[index].id = packedIds[index];
}
/**
 * Sets feature vector component of tuple at given index
 * \param[in] index index of entity in entity table
 * \param[in] featureIndex index of feature vector 
 * \param[in] featureValue value of feature vector at given featureIndex
 */
template<class T>
void
ResultTable<T>::setFeatureVectorComponent(int index, int featureIndex, VFloat featureValue) {
  pFeatures[index].set(featureIndex, featureValue);
  table[index].featureArray = &(pFeatures[index]);
}

/**
 * Sets feature vector of tuple at given index from string in svm format
 * \param[in] index index of entity in entity table
 * \param[in] str feature vector in svm format
 */
template<class T>
void
ResultTable<T>::setFeatureVectorFromStr(int index, string ftr) {
  istringstream s;
  s.str(ftr + "\n");
  s >> pFeatures[index];
  table[index].featureArray = &(pFeatures[index]);
}

/**
 * Sets eps value of tuple at given index
 * \param[in] index index of entity in entity table
 * \param[in] new_eps new eps value
 */
template<class T>
void
ResultTable<T>::setEps(int index, double new_eps) {
  table[index].eps = new_eps;
}

/**
 * Sets truth value of tuple at given index
 * \param[in] index index of entity in entity table
 * \param[in] new_truth_value new truth value
 */
template<class T>
void
ResultTable<T>::setTruthValue(int index, int new_truth_value) {
  table[index].truthValue = new_truth_value;
}

/**
 * Compares two tuples based on eps and returns -1 if first one is smaller, 0 if they are equal and 1 otherwise
 * \param[in] a first tuple
 * \param[in] b second tuple
 * \return -1 if first eps < second eps, 0 if they are equal, 1 otherwise
 */
template<class T>
int
ResultTable<T>::resultTupleComp (const void *a, const void *b) {
  double val_a = ((resultTuple *) a)->eps;
  double val_b = ((resultTuple *) b)->eps;
  
  if(val_a < val_b)
    return -1;
  else if(val_a == val_b)
    return 0;
  else
    return 1;
}

/**
 * sorts entities based on epsilons
 */
template<class T>
void
ResultTable<T>::sort() {
  qsort(table, entity_size, sizeof(resultTuple), resultTupleComp);
  // For Clustering
  if(bCluster) {	  
    //Timing m;
    //m.startTimer();
    T * tmpFeatures  = new T[entity_size];
    for(int i = 0; i < entity_size; i++) {    
      tmpFeatures[i] = T(*(table[i].featureArray));
      table[i].featureArray = &tmpFeatures[i];
    }
    delete [] pFeatures;
    pFeatures = tmpFeatures;
    // m.stopTimer();
    // std::cout << " done in " << m.getElapsedTime () << endl;
  }
  for(int i = 0; i < entity_size; i ++)
    packedIds[i] = table[i].id;
}
