/**
 * \file storage_manager.h
 */
#ifndef STORAGE_MANAGER_H
#define STORAGE_MANAGER_H

#include "vectors.h"
#include "timer.h"
#include "hazy_model.h"
#include "postgres_util.h"


/**
 * This tuple parser assumes the query returns
 * attribute 0 is id
 * attribute 1 is vector 
 * order is important
 */
template<class T>
struct sParse_IDvec {
  int id;
  T f;
  static int parse_tuple(PGresult *res, int index, sParse_IDvec<T> &c) {
    c.id = atoi(PQgetvalue(res, index, 0));
    std::string f_vec = PQgetvalue(res, index, 1);
    int r = c.f.fromPSQL(f_vec);
    return r;
  }
};


/**
 * Storage manager is an abstract class that allows us to hide the details of 
 * mainmemory, ondisk, or hybrid.
 * Storage manager that interacts with either main memory entities or on-disk entities (or both for hybrid)
 */
typedef int key; // primary key change it
typedef int sClass;

/**
 * Commonly needed information to build the model is stored
 * in the storage sepc
 */
struct storage_spec {
  // This indicates whether the model needs to be rebuilt or is an existing model.
  bool bRebuild;
  // entity table name in the database
  std::string entity_table_name;
  
  /**
   * A storage manager is allowed to store the strategy, and can reject an update
   * that asks for a different strategy.
   */
  hazy_model::strategy strategy;

  int unique_id;
};

template <class T> 
class Storage_Manager {
 protected:
  std::string entity_table_name;
  Hazy_Database *db_conn;
  
 public:
  /**
   * Default constructor
   */
  Storage_Manager() {db_conn = NULL;}
  /**
   * The constructor that initializes Storage_Maneger object with the given database object
   * \param[in] db_conn database connection
   */
  Storage_Manager(Hazy_Database *db_conn) {this->db_conn = db_conn;}
  
  /**
   * Makes a resort to reorganize entities according to their epsilon values
   * \param[in] m hazy model that we use to re-calculate epsllon values
   */
  virtual void resort(struct hazy_model &m) = 0;
  
  /**
   * Makes an incremental update on the entities by using hazy model
   * \param[in] m hazy model that we use to update entities classes
   * \param[out] waste_time waste time calcaulted during the method execution
   */
  virtual void incrementalUpdate(struct hazy_model &m, double &waste_time) = 0;
  
  // ************** Load Methods
  // Insert tuple: This only populates the internal table.
  // THIS DOES NOT DO SORTING. If you build the table using this function
  // THEN YOU NEED TO CALL RESORT.
  /**
   * Inserts the entity to the storage manager
   * \param[in] m hazy model which is used when the approach is hazy, because we need to calculate eps value
   * \param[in] e id of the entity
   * \param[in] f feature vector of the entity
   */
  virtual void insertEntity(struct hazy_model &m, key e, T f) = 0; 
  
  // this is the incremental maintenance call on adding in a new feature.
  // virtual void addEntity(struct hazy_model &m, key e, T f) = 0;
  
  // Repeated inserts. Overload for bulk loading.
  /**
   * Loads entities from the given file
   * \param[in] m hazy model
   * \param[in] szFileName file name
   */
  virtual void loadFromFile(struct hazy_model &m, const char* szFileName); 
  // This fails if there is no db connection in the constructor...
  /**
   * Loads entities from the database
   * \param[in] m hazy model
   */
  virtual int loadFromDatabase(struct hazy_model &m);
  
  // **********************
  // Read Methods
  // Single Entity Read
  /**
   * Retrieves the class of the entity
   * \param[in] e entitiy id
   * \param[out] c class of the entity
   * \param[in] m hazy model that we use to classify the entity
   */
  virtual void getEntityClass(key e, sClass &c, struct hazy_model &m) = 0;
  // All Members Read
  /**
   * Retrieves # of elements that belong to a given class
   * \param[in] c class that we ask # of elements belong it
   * \param[out] nClass # of elements in the class
   * \param[in] m hazy model that we use to classify entities
   * \param[out] waste_time waste time calculated during the method execution
   */
  virtual void getNumInClass(sClass c, int &nClass, struct hazy_model &m, double& waste_time) = 0;
  
  /**
   * Destructor
   */
  virtual ~Storage_Manager();
};

#include "storage_manager.hxx"
#endif
