#ifndef INCREMENTAL_SGD_H
#define INCREMENTAL_SGD_H

#include "model.h"
#include "reservoir.h"
#include <cmath>

const int RESERVOIR_ITERATIONS = 3;

/**
 * This class presents generic IncrementalSGD and two kind of sgd that extends this class which are TrueSGD and ReservoirSGD
 */

// template parameter is the kind of updates that one takes.
template<class T>
class IncrementalSGD {
 protected:
  ///step size
  double lambda;
  
  ///model  
  struct model _m;
  /**
   * This takes one step with this example.
   * \param[in] y class of the example
   * \param[in] ex feature vector of the example
   */
  void oneStep(double y, T ex); 
 public:
  /**
   * Constructor that initializes IncrementalSgd
   * \param[in] dim dimension of the model
   * \param[in] l lambda
   */
  IncrementalSGD(int dim, double l);
  
  void resetModel();

  /*
   * We want the model to be copied in, so it's not passed by reference.
   * \param[in] m model to be copied
   */
  void setModel(model m) { _m = m; }
  /**
   * Returns model
   * \return model
   */
  struct model& getModel() { return _m; }
  
  /**
   * There are many algorithms that do not train when we an add example
   * the return value indicates whether or not the model has changed as a result
   * of this update
   * \param[in] y class of the given example
   * \param[in] ex feature vector of the given example
   * \return whether model changed or not
   */
  bool virtual addExample(int y, T ex) = 0;
  
  /**
   * Desctructor
   */
  virtual ~IncrementalSGD() { } 
  
  /**
   * Classifies given feature vector of the entity with the current model and returns class of the example
   * \param[in] f feature vector of the entity
   * \return class of the example
   */
  bool classifyExample(T f);
  
  /**
   * Classifies given feature vector of the entity with the given model and returns distance of the entity to the model
   * \param[in] m model to classify the entity
   * \param[in] x feature vector of the entity
   * \return distance of the entity to the model
   */
  static double classifyExample(struct model &m, T x) {
    return dot(m.w,x) * m.wscale + m.bias;
  }
};

/**
 * This class extends IncrementalSGD and implements TrueSGD. This means that at each iteration, model changes
 */
template<class T>
class TrueSGD : public IncrementalSGD<T> {
 public:
  /**
   * Constructor that initializes TrueSGD
   * \param[in] dim dimension of the model
   * \param[in] lambda lambda value
   */
 TrueSGD(int dim,double lambda) : IncrementalSGD<T>(dim,lambda) { } 
  /**
   * Uses the given example to train the model
   * \param[in] y class of the example
   * \param[in] ex feature vector of the example
   * \return whether model changed or not
   */
  bool addExample(int y, T ex);
};

/**
 * This class extends IncrementalSGD and implements ReservoirSGD. This means that during the training we use a reservoir with the size r.
 * First, we use given training examples to fill reservoir.
 * Then, at each iteration, we randomly replace an example with the given training example in the reservoir and then train the model with the all elements in the reservoir
 */
template<class T> 
class ReservoirSGD : public IncrementalSGD<T> {
  Reservoir<T> _r;
 public:
  /**
   * Constructor that initializes ReservoirSGD
   * \param[in] dim dimension of the model
   * \param[in] lambda lambda value
   * \param[in] _rsize reservoir size
   */
 ReservoirSGD(int dim, double lambda, int _rsize) : IncrementalSGD<T>(dim, lambda),_r(_rsize)  { } 
  /**
   * Returns whether reservoir is filled or not
   * \return whether reservoir is filled
   */
  bool isFull() { return _r.isFull(); } // useful as a fast path
  /**
   * Uses the given example to train the model
   * \param[in] y class of the example
   * \param[in] ex feature vector of the example
   * \return whether model changed or not
   */
  bool addExample(int y, T ex);
};

#include "incremental_sgd.hxx"

#endif
