/**
 * \file reservoir.h
 */
#ifndef _reservoir_h_
#define _reservoir_h_
#include "vectors.h"
#include "globals.h"

/**
 * Reservoir class to keep examples in reservoir for training
 */
template <class T>
class Reservoir {
 private:
  ///size of reservoir
  int reservoir_size;
  ///keeps how many examples are in reservoir
  int current_size;
  ///list of labels of examples in reservoir
  int *exampleClasses;
  ///list of feature vectors of examples in reservoir
  T *pFeatures;
  /**
   * Swaps labels of given two examples
   *
   * \param[in] i position of first example
   * \param[in] j position of second example
   */
  void swapClassesOfLabels(int i, int j);
  /**
   * Swaps feature vectors of given two examples
   *
   * \param[in] i position of first example
   * \param[in] j position of second example
   */
  void swapVectors(int i, int j);
  /**
   * Writes given example(label and feature vector) to the given index
   *
   * \param[in] index index of example where it will be written
   * \param[in] classOfExample label of example
   * \param[in] featureArray feature vector of example
   */
  void copyExample(int index, int classOfExample, T featureArray);
  /**
   * Shuffles examples in the reservoir
   */
  void shuffle();
  
  /**
   * Initializes the reservoir with the given reservoir size and seed for randomness
   * \param[in] res_size reservoir size
   * \param[in] seed random seed
   */
  void init(int res_size, long seed);	
 public:
  ///Constructor
  /**
   * Constructor that initializes the reservoir with the given reservoir size
   * \param[in] res_size reservoir size
   */
  Reservoir(int res_size);
  /**
   * Constructor that initializes the reservoir with the given reservoir size and seed for randomness
   * \param[in] res_size reservoir size
   * \param[in] seed random seed
   */
  Reservoir(int res_size, long seed);
  /**
   * Inserts new example to the reservoir
   * \param[in] newEx example string in svm format
   */
  void addExample(std::string newEx);
  
  /**
   * Adds example to the reservoir
   * \param[in] y class of the example
   * \param[in] ex feature vector of the example
   */
  void addExample(int y, T ex);
  /**
   * Returns label of example at given index
   *
   * \param[in] index index of example in the reservoir
   * \return label of example
   */	
  int getExampleClass(int index);
  /**
   * Returns feature vector of example at given index
   *
   * \param[in] index index of example in the reservoir
   * \return feature vector of example
   */
  const T getExampleFeatureVector(int index);
  /**
   * Returns how much of reservoir is filled
   * \return # of examples in reservoir
   */
  int getCurrentSize();
  /**
   * Returns whether the reservoir is full or not
   * \return reservoir is full or not
   */
  bool isFull() { return getCurrentSize() ==  reservoir_size; }
};
#include "reservoir.hxx"
#endif
