/**
 * \file reservoir.cxx
 */
#include <iostream>
#include <cmath>

// #include "reservoir.h"

/**
 * Initializes the reservoir with the given reservoir size and seed for randomness
 * \param[in] res_size reservoir size
 * \param[in] seed random seed
 */
template<class T>
void
Reservoir<T>::init(int size, long seed) {
  srand(seed);
  this->reservoir_size = size;
  this->current_size = 0;
  exampleClasses = new int[reservoir_size];
  pFeatures      = new T[reservoir_size];
}

///Constructor
/**
 * Constructor that initializes the reservoir with the given reservoir size
 * \param[in] res_size reservoir size
 */
template<class T>
Reservoir<T>::Reservoir(int size) {
  init(size, time(NULL));
}

/**
 * Constructor that initializes the reservoir with the given reservoir size and seed for randomness
 * \param[in] res_size reservoir size
 * \param[in] seed random seed
 */
template<class T>
Reservoir<T>::Reservoir(int size, long seed) {
  init(size, seed);
}

/**
 * Swaps labels of given two examples
 * \param[in] i position of first example
 * \param[in] j position of second example
 */
template<class T>
void
Reservoir<T>::swapClassesOfLabels(int i, int j) {
  int tempClass = exampleClasses[i];
  exampleClasses[i] = exampleClasses[j];
  exampleClasses[j] = tempClass;
}

/**
 * Swaps feature vectors of given two examples
 * \param[in] i position of first example
 * \param[in] j position of second example
 */
template<class T>
void
Reservoir<T>::swapVectors(int i, int j) {
  T tempVector = pFeatures[i];
  pFeatures[i] = pFeatures[j];
  pFeatures[j] = tempVector;
}

/**
 * Shuffles examples in the reservoir
 */
template <class T>
void
Reservoir<T>::shuffle() {
  for(int i = 0; i < reservoir_size; i ++) {
    int randVal = int(floor(drand48() * reservoir_size));
    this->swapClassesOfLabels(randVal, i);
    this->swapVectors(randVal, i);
  }
}
/**
 * Writes given example(label and feature vector) to the given index
 * \param[in] index index of example where it will be written
 * \param[in] classOfExample label of example
 * \param[in] featureArray feature vector of example
 */
template<class T>
void
Reservoir<T>::copyExample(int index, int classOfExample, T featureArray) {
  exampleClasses[index] = classOfExample; 
  pFeatures[index] = featureArray;
}

/**
 * Adds example to the reservoir
 * \param[in] y class of the example
 * \param[in] ex feature vector of the example
 */
template<class T>
void 
Reservoir<T>::addExample(int classOfExample, T featureArray)  {
  if(current_size < reservoir_size) {
    copyExample(current_size, classOfExample, featureArray);
    current_size ++;
    
    if(current_size == reservoir_size)
      shuffle();
  }
  else {
    int new_val = int(floor(drand48() * reservoir_size));
    copyExample(new_val, classOfExample, featureArray);
  }
}

/**
 * Inserts new example to the reservoir
 * \param[in] newEx example string in svm format
 */
template<class T>
void
Reservoir<T>::addExample(std::string newExample) {
  int classOfExample;
  T featureArray;
  std::istringstream f;
  f.str(newExample + "\n");	
  f >> classOfExample >> featureArray;       
  addExample(classOfExample, featureArray);
}

/**
 * Returns label of example at given index
 * \param[in] index index of example in the reservoir
 * \return label of example
 */	
template<class T>
int
Reservoir<T>::getExampleClass(int index) {
  return exampleClasses[index];
}

/**
 * Returns how much of reservoir is filled
 * \return # of examples in reservoir
 */
template<class T>
int
Reservoir<T>::getCurrentSize() {
  return current_size;
}

/**
 * Returns feature vector of example at given index
 * \param[in] index index of example in the reservoir
 * \return feature vector of example
 */
template<class T>
const T
Reservoir<T>::getExampleFeatureVector(int index) {
  return pFeatures[index];
}
