// This contains structures for serialization.

/**
 * Pair structure that is used for sparse vectors
 */
typedef struct Pair {
  int index;
  double value;
} Pair;

/**
 * This structure is used when SVector is being written as binary data during socket communication
 * Essentially, Pair and this structure are same
 */
struct serialized_svector_entry {
  int i;
  double v;
};

/**
 * This structure is used when FVector is being written as binary data during socket communication
 */
struct serialized_fvector_entry {
  double v;
};

#ifdef __POSTGRES_CALL
#include "postgres.h"

/**
 * Calculates total # of bytes that are used by SVector
 * \param[in] _size # of elements in SVector
 * \return size of the SVector in bytes
 */
size_t
calculate_svector_total_size(int _size) {
  size_t size_of_pair = sizeof(struct serialized_svector_entry);
  return sizeof(int) + size_of_pair * _size;
}

/**
 * Calculates total # of bytes that are used by FVector
 * \param[in] _size # of elements (dimension) in FVector
 * /return size of the FVector in bytes
 */
size_t
calculate_fvector_total_size(int _size) {
  size_t size_of_pair = sizeof(struct serialized_fvector_entry);
  return sizeof(int) + size_of_pair * _size;
}

/**
 * Writes given Fvector as binary data
 * \param[out] out data pointer that FVector will be written to
 * \param[in] max_size maximum size of the data pointer (out)
 * \param[in] _size # of elements in the FVector
 * \param[in] ddata FVector data
 * \return # of bytes that are written to out data
 */
size_t
write_fvector_bin_psql_double(void *out, size_t max_size, int _size, double *ddata) {
  int i               = 0;
  size_t total_size   = calculate_fvector_total_size(_size);
  if(total_size > max_size) {
    elog(INFO, "total size: %d, max size: %d", total_size, max_size);
    elog(ERROR, "We blew up the buffer in [write_fvector_bin_psql_double]\n");
    return 0;
  }
  ((int *)out)[0] = _size;
  struct serialized_fvector_entry *_data = (struct serialized_fvector_entry*) &(((int *)out)[1]);
  memcpy(_data, ddata, total_size);
  return total_size;
}

/**
 * Writes given Svector as binary data
 * \param[out] out data pointer that SVector will be written to
 * \param[in] max_size maximum size of the data pointer (out)
 * \param[in] _size # of elements in the SVector
 * \param[in] ddata SVector data
 * \return # of bytes that are written to out data
 */
size_t
write_svector_bin_psql_pair(void *out, size_t max_size, int _size, Pair *ddata)  {
  int i               = 0;
  size_t total_size   = calculate_svector_total_size(_size);
  if(total_size > max_size) {
    elog(INFO, "total size: %d, max size: %d", total_size, max_size);
    elog(ERROR, "We blew up the buffer in [write_fvector_bin_psql_pair]\n");
    return 0;
  }
  ((int *)out)[0] = _size;
  struct serialized_svector_entry *_data = (struct serialized_svector_entry*) &(((int *)out)[1]);
  
  memcpy(_data, ddata, total_size);
  return total_size;
}

#endif
