/**
 * \file postgres_util.cxx
 */

#include "postgres_util.h"

///Default constructor that takes only database name (needs to consider username and password, too!)
Hazy_Database::Hazy_Database(std::string dbname) {
  std::string conninfo;
  conninfo = "dbname = " + dbname;
  
  conn = PQconnectdb(conninfo.c_str());
  
  if (PQstatus(conn) == CONNECTION_BAD)
    {
      fprintf(stderr, "Connection to database '%s' failed.\n", dbname.c_str());
      fprintf(stderr, "%s", PQerrorMessage(conn));
      exit_nicely();
    }
}

/**
 * Executes given statement
 * \param[in] stmt query statement
 * \param[in] msg error msg to print in case of error
 *
 * \return success
 */
int 
Hazy_Database::execute_statement_msg(const char* stmt, const char* msg) {
  res = PQexec(conn, stmt);
  if (!res || (PQresultStatus(res) != PGRES_COMMAND_OK && PQresultStatus(res) != PGRES_TUPLES_OK))
    {   
      fprintf(stderr, "PROBLEM %s. PSQL= %s\n", msg, PQerrorMessage(conn));
      PQclear(res);
      exit_nicely();
      return -1;
    }
  PQclear(res);
  return 0;
}

/**
 * Executes given statement and retrieve text value
 * \param[in] query query statement
 * \param[in] msg error msg to print in case of error
 * \param[out] result string value to be retrieved
 *
 * \return success
 */
int 
Hazy_Database::execute_query_msg_text(const char* query, const char* msg, std::string &result) {
  res = PQexec(conn, query);
  if (!res || (PQresultStatus(res) != PGRES_COMMAND_OK && PQresultStatus(res) != PGRES_TUPLES_OK))
    {
      fprintf(stderr, "PROBLEM %s: %s\n", msg, PQerrorMessage(conn));
      PQclear(res);
      exit_nicely();
      return -1;
    }
  
  result = PQgetvalue(res, 0, 0);
  PQclear(res);
  return 0;
}

/**
 * Executes given statement and retrieve double value
 * \param[in] query query statement
 * \param[in] msg error msg to print in case of error
 * \param[out] result double value to be retrieved
 *
 * \return success
 */
int 
Hazy_Database::execute_query_msg_double(const char* query, const char* msg, double &result) {
  res = PQexec(conn, query);
  if (!res || (PQresultStatus(res) != PGRES_COMMAND_OK && PQresultStatus(res) != PGRES_TUPLES_OK))
    {
      fprintf(stderr, "PROBLEM %s: %s\n", msg, PQerrorMessage(conn));
      PQclear(res);
      exit_nicely();
      return -1;
    }
  
  result = atof(PQgetvalue(res, 0, 0));
  PQclear(res);
  return 0;
}

/**
 * Executes given statement and retrieve int value
 * \param[in] query query statement
 * \param[in] msg error msg to print in case of error
 * \param[out] result int value to be retrieved
 *
 * \return success
 */
int 
Hazy_Database::execute_query_msg_int(const char* query, const char* msg, int &result) {
  res = PQexec(conn, query);
  if (!res || (PQresultStatus(res) != PGRES_COMMAND_OK && PQresultStatus(res) != PGRES_TUPLES_OK))
    {
      fprintf(stderr, "PROBLEM %s: %s\n", msg, PQerrorMessage(conn));
      PQclear(res);
      exit_nicely();
      return -1;
    }
  
  result = atoi(PQgetvalue(res, 0, 0));
  PQclear(res);
  return 0;
}


/**
 * Copies given content of the stream to the given table
 * \param[in] s input stream that holds data
 * \param[in] table_name table name in the database that data will be copied
 * \param[in] delimiter delimiter that used to format data in the input stream
 * \return success
 */
int
Hazy_Database::copy_to_table(std::istream &s, std::string &table_name, std::string &delimiter) {
  std::string start_copy_query = "COPY " + table_name + " FROM STDIN DELIMITER '" + delimiter + "';";
  
  res = PQexec(conn, start_copy_query.c_str());
  if (!res || (PQresultStatus(res) != PGRES_COPY_IN))
    {
      fprintf(stderr, "PROBLEM %s: %s\n", start_copy_query.c_str(), PQerrorMessage(conn));
      PQclear(res);
      exit_nicely();
      return -1;
    }
  
  //send data
  std::string line;
  while(!s.eof() && getline(s, line)) {
    line = line + "\n";
    int ret_val = PQputCopyData(conn, line.c_str(), strlen(line.c_str()));
    if(ret_val != 1) {
      fprintf(stderr, "problem in copy to table method during copying");
      exit_nicely();
      return -1;
    }
  }
  
  int ret_val = PQputCopyEnd(conn, NULL);
  if(ret_val != 1) {
    fprintf(stderr, "problem in copy to table method when ending copying");
    exit_nicely();
    return -1;
  }
  
  res = PQgetResult(conn);
  if(!res || PQresultStatus(res) != PGRES_COMMAND_OK) {
    fprintf(stderr, "PROBLEM %s\n", PQerrorMessage(conn));
    PQclear(res);
    exit_nicely();
    return -1;
  }
  
  PQclear(res);
  return 0;
}

/**
 * Copies content of the database table to the output stream
 * \param[in] s output stream that the table content will be written to
 * \param[in] table_name name of the table in the database that has the data
 * \param[in] delimiter delimiter that will be used to format data in the output stream
 * \return success
 */
int
Hazy_Database::copy_from_table(std::ostream &s, std::string &table_name, std::string &delimiter) {
  std::string start_copy_query = "COPY " + table_name + " TO STDOUT DELIMITER '" + delimiter + "';";
  
  res = PQexec(conn, start_copy_query.c_str());
  if (!res || (PQresultStatus(res) != PGRES_COPY_OUT))
    {
      fprintf(stderr, "PROBLEM %s: %s\n", start_copy_query.c_str(), PQerrorMessage(conn));
      PQclear(res);
      exit_nicely();
      return -1;
    }
  
  int nFields = PQnfields(res);
  
  char **buff = (char **) malloc(sizeof (char *) * nFields);
  for(int i = 0; i < nFields; i ++)
    buff[i] = (char *) malloc(sizeof(char) * 100);
  
  while (PQgetCopyData(conn, buff, 0) != -1) {
    s << buff[0];
  }
  
  return 0;
}


// /** 
//  * This method closes database connection, when any exception occurs
//  */
// void
// Hazy_Database::exit_nicely()
// {
//   fprintf(stderr, "problem: %s", PQerrorMessage(conn));
//   PQfinish(conn);
// }

int
Hazy_Database::count_binary_query(const char *query, double v1, double v2, int &count_of_tuples) {
  char *paramValues[2];
  int     paramLengths[2];
  int     paramFormats[2];
  
  paramValues[0] = pack_double(v1);
  paramLengths[0] = sizeof(double);
  paramFormats[0] = 1;
  
  paramValues[1] = pack_double(v2);
  paramLengths[1] = sizeof(double);
  paramFormats[1] = 1;
  
  res = PQexecParams(conn,
		     query,
		     2,   /* one param */
		     NULL,  /* let the backend deduce param type */
		     paramValues,
		     paramLengths,
		     paramFormats,
		     0);    /* ask for binary results */
  
  if (PQresultStatus(res) != PGRES_TUPLES_OK) {
    fprintf(stderr, "SELECT failed: %s", PQerrorMessage(conn));
    PQclear(res);
    exit_nicely();
    return -1;
  }
  
  count_of_tuples = atoi(PQgetvalue(res, 0, 0));
  PQclear(res);
  
  return 0;
}

/**
 * Closes database connection
 */
void
Hazy_Database::closeConnection() {
  PQfinish(conn);
}

/**
 * Converts given SVector to database string
 * \param[in] s given svector
 * \param[out] feature_vector_str database string
 */
void
svector_to_dbstring(const SVector &s, std::string& feature_vector_str) {
  std::ostringstream feature_vector;
  
  std::ios::fmtflags oldflags = feature_vector.flags();
  std::streamsize oldprec = feature_vector.precision();
  feature_vector << std::scientific << std::setprecision(sizeof(VFloat)==4 ? 7 : 16);
  
  int n = s.npairs();
  /*	feature_vector << "ARRAY[";
	
	for(const SVector::Pair *p = s; p->i >= 0; p ++) {
	if(n == s.npairs())
	feature_vector << "'(" << p->i << ", " << p->v << ")'::pair" << ((n > 1) ? "," : "");
	else
	feature_vector << "'(" << p->i << ", " << p->v << ")'" << ((n > 1) ? "," : "");
	n--;
	}
	
	feature_vector << "]";*/
  feature_vector << "'{";
  for(const SVector::Pair *p = s; p->i >= 0; p ++) {
    feature_vector << "\"(" << p->i << "," << p->v << ")\"" << ((n > 1) ? "," : "");
    n--;
  }
  feature_vector << "}'::pair[]";
  
  feature_vector_str = std::string(feature_vector.str());
  
  feature_vector.precision(oldprec);
  feature_vector.flags(oldflags);
}

/**
 * Converts given FVector to database string
 * \param[in] v given fvector
 * \param[out] feature_values database string
 */
void
fvector_to_dbstring(const FVector &v, std::string &feature_values) {
  std::ostringstream values;
  
  std::ios::fmtflags oldflags = values.flags();
  std::streamsize oldprec = values.precision();
  values << std::scientific << std::setprecision(sizeof(VFloat)==4 ? 7 : 16);
  
  /*  values << "ARRAY[";
      for(int i = 0; i < v.size(); i++) {
      if(i == 0)
      values << v.get(i) << "::float8" << ((i < v.size() - 1) ? "," : "");
      else
      values << v.get(i) << ((i < v.size() - 1) ? "," : "");
      }
      values << "]";*/
  values << "'{";
  for(int i = 0; i < v.size(); i ++)
    values << v.get(i) << ((i < v.size() - 1) ? "," : "");
  
  values << "}'::float8[]";
  
  feature_values = std::string(values.str());
  
  values.precision(oldprec);
  values.flags(oldflags);
  
}

/**
 * Converts given model vector and bias to database strings after scaling the model vector
 * \param[in] m model that database strings will be constructed
 * \param[out] model_string string that is constructed from model vector after scaling
 * \param[out] bias_string string that is constructed from bias
 */
void 
model_to_dbstring(const struct model &m, std::string &model_string, std::string &bias_string) {
  FVector v = m.w;
  std::ostringstream oss_bias;
  
  std::ios::fmtflags oldflags = oss_bias.flags();
  std::streamsize oldprec = oss_bias.precision();
  oss_bias << std::scientific << std::setprecision(sizeof(double)==4 ? 7 : 16);
  
  
  oss_bias << m.bias;
  v.scale(m.wscale);
  fvector_to_dbstring(v, model_string);
  bias_string = std::string(oss_bias.str());
  
  oss_bias.precision(oldprec);
  oss_bias.flags(oldflags);
}

/**
 * Checks that query is executed successfully
 * \param[in] retVal success that returned from execution of the query
 * \param[in] query the query which is executed
 */
void
checkQueryReturnValue(int retVal, std::string query) {
  if(retVal == -1) {
    fprintf(stderr, "error on query: %s", query.c_str());
    exit(1);
  }
}	

// 
// Type unpacking routines

void memcpy_advance(char* &data, char *dest, int len) {
  memcpy(data, dest, len);
  data += len;
}

int
unpack_int32(char *data) {
  return ntohl(*((uint32_t*) data));
}


double
unpack_double(char* data) {
  union
  {
    double d;
    unsigned int n[2];
  } swap;
  memcpy(&(swap.n), data, sizeof(unsigned int)*2);
  unsigned int tmp = ntohl(swap.n[0]);
  swap.n[0]        = ntohl(swap.n[1]);
  swap.n[1]        = tmp;
  return swap.d;
}

char *pack_double(double val) {
  union
  {
    double d;
    int64 i;
  } swap;

  swap.d = val;

  char *data = (char *) malloc(sizeof(unsigned int) * 2);

  uint32    n32;

  /* High order half first, since we're doing MSB-first */
  n32 = (uint32) (swap.i >> 32);
  n32 = htonl(n32);
  memcpy(&(data[0]), (char *) &n32, 4);

  /* Now the low order half */
  n32 = (uint32) (swap.i);
  n32 = htonl(n32);  

  memcpy(&(data[4]), (char *) &n32, 4);

  return data;
}
