/**
 * \file postgres_util.h
 */
#ifndef __postgres_util_h
#define __postgres_util_h
#include "globals.h"
#include "libpq-fe.h"
#include "model.h"
#include <iomanip>
#include <iostream>
#include <vector>
#include <map>

/* for ntohl/htonl-- binary encoding in psql */
#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef HAVE_INT64
typedef long int int64;
#endif

typedef unsigned int uint32;


/**
 * This class provides database connection, executes update and select queries
 */
class Hazy_Database {
 private:
  ///Connection object
  PGconn *conn;
  ///Result object
  PGresult *res;
  /**
   * When query gives error, then it closes database connection
   */
  void exit_nicely()
  {
    fprintf(stderr, "problem: %s", PQerrorMessage(conn));
    PQfinish(conn);
  }
  
 public:
  ///Default constructor that takes only database name (needs to consider username and password, too!)
  Hazy_Database(std::string dbname);
  /**
   * Executes given statement
   * \param[in] stmt query statement
   * \param[in] msg error msg to print in case of error
   *
   * \return success
   */
  int execute_statement_msg(const char* stmt, const char *msg);
  /**
   * Executes given statement and retrieve string value
   * \param[in] query query statement
   * \param[in] msg error msg to print in case of error
   * \param[out] result string value to be retrieved
   *
   * \return success
   */
  int execute_query_msg_text(const char* query, const char* msg, std::string &result);

  /**
   * Executes given statement and retrieve double value
   * \param[in] query query statement
   * \param[in] msg error msg to print in case of error
   * \param[out] result double value to be retrieved
   *
   * \return success
   */
  int execute_query_msg_double(const char* query, const char* msg, double &result);
  /**
   * Executes given statement and retrieve int value
   * \param[in] query query statement
   * \param[in] msg error msg to print in case of error
   * \param[out] result int value to be retrieved
   *
   * \return success
   */
  int execute_query_msg_int(const char* query, const char* msg, int &result);
  
  /**
   * Copies given content of the stream to the given table
   * \param[in] s input stream that holds data
   * \param[in] table_name table name in the database that data will be copied
   * \param[in] delimiter delimiter that used to format data in the input stream
   * \return success
   */
  int copy_to_table(std::istream &s, std::string &table_name, std::string &delimiter);
  /**
   * Copies content of the database table to the output stream
   * \param[in] s output stream that the table content will be written to
   * \param[in] table_name name of the table in the database that has the data
   * \param[in] delimiter delimiter that will be used to format data in the output stream
   * \return success
   */
  int copy_from_table(std::ostream &s, std::string &table_name, std::string &delimiter);
  
  /**
   * Parses the postgres tuples and stores them in a vector
   * \param[in] szQuery query to retrieve tuples from postgres
   * \param[in] tuple_parser function pointer to parse the tuples
   * \param[out] _out the vector that parsed tuples will be stored
   * \return success
   */
  template<class T>
    int postgresTupleParser(const char* szQuery, int (*tuple_parser)(PGresult *r, int index, T &out), std::vector<T> &_out)
    {
      res = PQexec(conn, szQuery);
      if (!res || (PQresultStatus(res) != PGRES_COMMAND_OK && PQresultStatus(res) != PGRES_TUPLES_OK))
	{
	  fprintf(stderr, "PROBLEM %s: %s\n", "query error in postgresTupleParser", PQerrorMessage(conn));
	  PQclear(res);
	  this->exit_nicely();
	  return -1;
	}
      int _n = PQntuples(res);
      for(int i = 0; i < _n; i ++) {
	T _t;
	if(tuple_parser(res, i, _t) < 0) {     
	  fprintf(stderr, "tuple could not be parsed in postgresTupleParser method\n");
	  return -1;
	}
	//    int id = atoi(PQgetvalue(res, i, 0));
	//string f_vec = PQgetvalue(res, i, 1);
	//double eps_val = atof(PQgetvalue(res, i, 2));
	_out.push_back(_t);
      }
      return _n;
    }

  template<class T>
    int
    postgresTupleParserBinary(const char* szQuery, int (*tuple_parser)(PGresult *r, int index, T &out), std::vector<T> &_out) {
    res = PQexecParams(conn,
		       szQuery,
		       0,
		       NULL,
		       NULL,
		       NULL,
		       NULL,
		       1);

    if (!res || (PQresultStatus(res) != PGRES_COMMAND_OK && PQresultStatus(res) != PGRES_TUPLES_OK))
      {
	fprintf(stderr, "PROBLEM %s: %s\n", "query error in postgresTupleParser", PQerrorMessage(conn));
	PQclear(res);
	exit_nicely();
	return -1;
      }
    int _n = PQntuples(res);
    for(int i = 0; i < _n; i ++) {
      T _t;
      if(tuple_parser(res, i, _t) < 0) {
	fprintf(stderr, "tuple could not be parsed in postgresTupleParser method\n");
	return -1;
      }
      _out.push_back(_t);
    }
    PQclear(res);
    return _n;
  }
  
  /**
   * Parses the postgres tuples in hybrid approach and stores them in a map
   * \param[in] szQuery query to retrieve tuples from postgres
   * \param[in] tuple_parser function pointer to parse the tuples
   * \param[out] _out the map that parsed tuples will be stored
   * \return success
   */
  template<class K, class T>
    int postgresMapParser(const char* szQuery, int (*tuple_parser)(PGresult *r, int index, K &k, T &v), std::map<K,T> &_out)
  {
    res = PQexec(conn, szQuery);
    if (!res || (PQresultStatus(res) != PGRES_COMMAND_OK && PQresultStatus(res) != PGRES_TUPLES_OK))
      {
	fprintf(stderr, "PROBLEM %s: %s\n", "query error in postgresTupleParser", PQerrorMessage(conn));
	PQclear(res);
	exit_nicely();
	return -1;
      }
    int _n = PQntuples(res);
    for(int i = 0; i < _n; i ++) {
      K _k;
      T _t;
      if(tuple_parser(res, i, _k, _t) < 0) {     
	fprintf(stderr, "tuple could not be parsed in postgresTupleParser method\n");
	return -1;
      }
      _out[_k] = _t;
    }
    return _n;
  }

  int count_binary_query(const char *query, double v1, double v2, int &count_of_tuples);

  /**
   * Closes database connection
   */
  void closeConnection();
};


// Helper functions
/**
 * Converts given SVector to database string
 * \param[in] s given svector
 * \param[out] feature_vector_str database string
 */
void svector_to_dbstring(const SVector& s, std::string& feature_vector_str);
/**
 * Converts given FVector to database string
 * \param[in] v given fvector
 * \param[out] feature_vector_str database string
 */
void fvector_to_dbstring(const FVector& v, std::string &feature_values);
/**
 * Converts given model vector and bias to database strings after scaling the model vector
 * \param[in] m model that database strings will be constructed
 * \param[out] model_string string that is constructed from model vector after scaling
 * \param[out] bias_string string that is constructed from bias
 */
void model_to_dbstring(const struct model &m, std::string &model_string, std::string &bias_string);
/**
 * Checks that query is executed successfully
 * \param[in] retVal success that returned from execution of the query
 * \param[in] query the query which is executed
 */
void checkQueryReturnValue(int retVal, std::string query);

int unpack_int32(char *data);
uint64_t unpack_int64(char* data);
double unpack_double(char* data);
char* pack_double(double val);


#endif
